=== added file '.bzrignore'
--- .bzrignore	1970-01-01 00:00:00 +0000
+++ .bzrignore	2009-01-19 10:34:45 +0000
@@ -0,0 +1,10 @@
+*.pyc
+*.pyo
+.*.swp
+/jail
+/bin/timount/timount
+/bin/trampoline/conf.h
+/bin/trampoline/trampoline
+/ivle/conf/conf.py
+/ivle/conf/jailconf.py
+/www/php/phpBB3/config.php

=== modified file 'bin/ivle-enrol'
--- bin/ivle-enrol	2009-01-13 01:36:15 +0000
+++ bin/ivle-enrol	2009-02-02 04:00:25 +0000
@@ -25,13 +25,39 @@
 import os
 import sys
 
-import ivle.db
+from ivle.database import get_store, Subject, Semester, Offering, User, \
+                          AlreadyEnrolledError
+
+def die(error):
+    print >> sys.stderr, '%s: %s' % (os.path.basename(sys.argv[0]), error)
+    sys.exit(1)
 
 if len(sys.argv) != 5:
-    print >> sys.stderr, "usage: %s <login> <subjectcode> <semester> <year>" \
+    print >> sys.stderr, "usage: %s <login> <subjectcode> <year> <semester>" \
                          % os.path.basename(sys.argv[0])
-    sys.exit(1)
-
-if not ivle.db.DB().add_enrolment(*sys.argv[1:]):
-    print >> sys.stderr, "enrolment failed!"
-    sys.exit(1)
+    sys.exit()
+
+store = get_store()
+
+user = User.get_by_login(store, sys.argv[1])
+if user is None:
+    die("user not found")
+
+offering = store.find(Offering,
+                    Subject.code == unicode(sys.argv[2]),
+                    Semester.year == unicode(sys.argv[3]),
+                    Semester.semester == unicode(sys.argv[4]),
+                    Offering.subject_id == Subject.id,
+                    Offering.semester_id == Semester.id).one()
+if not offering:
+    die("subject offering not found")
+
+if not offering.semester.active:
+    die("cannot enrol - semester inactive")
+    
+try:
+    offering.enrol(user)
+except AlreadyEnrolledError:
+    die("already enrolled")
+
+store.commit()

=== modified file 'bin/ivle-enrolallusers'
--- bin/ivle-enrolallusers	2009-01-13 01:36:15 +0000
+++ bin/ivle-enrolallusers	2009-02-02 04:00:25 +0000
@@ -34,7 +34,7 @@
 import optparse
 import logging
 
-import ivle.db
+import ivle.database
 import ivle.pulldown_subj
 
 p = optparse.OptionParser()
@@ -54,18 +54,11 @@
     print >> sys.stderr, "%s must be run as root" % sys.argv[0]
     sys.exit(1)
 
-try:
-    db = ivle.db.DB()
-    if options.user is None:
-        users = db.get_users()
-    else:
-        users = [db.get_user(options.user)]
-    def repack(user):
-        return (user.login, user.unixid)
-    uids = dict(map(repack,users))
-except Exception, message:
-    logging.error(str(message))
-    sys.exit(1)
+store = ivle.database.get_store()
+if options.user is None:
+    users = store.find(ivle.database.User).order_by(ivle.database.User.login)
+else:
+    users = [ivle.database.User.get_by_login(store, options.user)]
 
 if options.user is None:
     logging.info("enrolment started")
@@ -76,16 +69,18 @@
     logging.error("Year must be numeric")
     sys.exit(1)
 
-users.sort(key=lambda user: user.login)
+options.year = unicode(options.year)
+
 for user in users:
     try:
         # Get all subjects this user is enrolled in, and add them to the DB if
         # they match one of our local subject codes
-        res = ivle.pulldown_subj.enrol_user(user.login, options.year)
+        res = ivle.pulldown_subj.enrol_user(store, user, options.year)
         logging.info("Enrolled user %s in %d subject%s." % (user.login, res,
                         '' if res == 1 else 's'))
     except Exception, message:
         logging.warning(str(message))
         continue
     
+store.commit()
 logging.info("enrolment completed successfully")

=== modified file 'bin/ivle-listusers'
--- bin/ivle-listusers	2009-01-13 01:36:15 +0000
+++ bin/ivle-listusers	2009-02-02 04:00:25 +0000
@@ -27,7 +27,7 @@
 import os
 import getopt
 
-import ivle.db
+import ivle.database
 
 # Options processing
 def usage():
@@ -56,15 +56,10 @@
     print "Must run listusers.py as root."
     sys.exit()
 
-try:
-    db = ivle.db.DB()
-    list = db.get_users()
-except Exception, message:
-    print "Error: " + str(message)
-    sys.exit(1)
+store = ivle.database.get_store()
+users = store.find(ivle.database.User).order_by(ivle.database.User.login)
 
-list.sort(key=lambda user: user.login)
-for user in list:
+for user in users:
     if justnames:
         print user.login
     else:

=== modified file 'bin/ivle-makeuser'
--- bin/ivle-makeuser	2009-01-13 01:36:15 +0000
+++ bin/ivle-makeuser	2009-02-02 04:00:25 +0000
@@ -31,7 +31,8 @@
 import os
 import getopt
 
-import ivle.makeuser
+from ivle.database import get_store, User
+from ivle.pulldown_subj import enrol_user
 
 # Requireds and optionals will be used to display the usage message
 # AND do argument processing
@@ -66,22 +67,27 @@
 # Get the dictionary of fields from opts and args
 user = {}
 for i in range(0, len(requireds)):
-    user[requireds[i]] = args[i]
+    user[requireds[i]] = unicode(args[i])
 for short, long, _ in optionals:
     try:
-        user[long] = opts['-' + short]
+        user[long] = unicode(opts['-' + short])
     except KeyError:
         try:
-            user[long] = opts['--' + long]
+            user[long] = unicode(opts['--' + long])
         except KeyError:
             pass
 login = user['login']
 if 'nick' not in user:
     user['nick'] = user['fullname']
 
+store = get_store()
+
 try:
     # Make the user's database entry
-    ivle.makeuser.make_user_db(**user)
+    userobj = User(**user)
+    store.add(userobj)
+    enrol_user(store, userobj)
+    store.commit()
 except Exception, message:
     print "Error: " + str(message)
     sys.exit(1)

=== modified file 'bin/ivle-marks'
--- bin/ivle-marks	2009-01-13 01:36:15 +0000
+++ bin/ivle-marks	2009-02-02 04:00:25 +0000
@@ -27,10 +27,10 @@
 import os
 import re
 import csv
-import time
 from xml.dom import minidom
 
-import ivle.db
+import ivle.database
+import ivle.worksheet
 import ivle.conf
 
 if os.getuid() != 0:
@@ -63,7 +63,7 @@
     (This is not marks, it's other user data).
     """
     last_login = (None if user.last_login is None else
-                    time.strftime("%d/%m/%y", user.last_login))
+                    user.last_login.strftime("%d/%m/%y"))
     return [user.studentid, user.login, user.fullname, last_login]
 userdata_header = ["Student ID", "Login", "Full name", "Last login"]
 
@@ -110,7 +110,7 @@
     """
     return worksheets + ["Total %", "Mark"]
 
-def get_marks_user(subject, worksheets, user):
+def get_marks_user(subject, worksheet_names, user):
     """
     Given a subject, a list of strings (the assessable worksheets), and a user
     object, returns the user's percentage for each worksheet, overall, and
@@ -127,18 +127,15 @@
     problems_done = 0
     problems_total = 0
 
-    for worksheet in worksheets:
-        try:
-            # We simply ignore optional exercises here
-            mand_done, mand_total, _, _ = (
-                db.calculate_score_worksheet(user.login, subject,
-                    worksheet))
-            worksheet_pcts.append(float(mand_done) / mand_total)
-            problems_done += mand_done
-            problems_total += mand_total
-        except ivle.db.DBException:
-            # Worksheet is probably not in database yet
-            pass
+    for worksheet_name in worksheet_names:
+        worksheet = ivle.database.Worksheet.get_by_name(store,
+            subject, worksheet_name)
+        # We simply ignore optional exercises here
+        mand_done, mand_total, _, _ = (
+            ivle.worksheet.calculate_score(store, user, worksheet))
+        worksheet_pcts.append(float(mand_done) / mand_total)
+        problems_done += mand_done
+        problems_total += mand_total
     problems_pct = float(problems_done) / problems_total
     problems_pct_int = (100 * problems_done) / problems_total
     # XXX Marks calculation (should be abstracted out of here!)
@@ -153,11 +150,9 @@
     csvfile.writerow(userdata + marksdata)
 
 try:
-    # Get the list of assessable worksheets from the subject.xml file,
-    # and the list of all users from the DB.
+    # Get the list of assessable worksheets from the subject.xml file.
     worksheets = get_assessable_worksheets(subject)
-    db = ivle.db.DB()
-    list = db.get_users()
+    store = ivle.database.get_store()
 except Exception, message:
     print >>sys.stderr, "Error: " + str(message)
     sys.exit(1)
@@ -166,7 +161,6 @@
 csvfile = csv.writer(sys.stdout)
 csvfile.writerow(userdata_header + get_marks_header(worksheets))
 
-list.sort(key=lambda user: user.login)
-for user in list:
+for user in store.find(ivle.database.User).order_by(ivle.database.User.login):
     writeuser(subject, worksheets, user, csvfile)
 

=== modified file 'bin/ivle-mountallusers'
--- bin/ivle-mountallusers	2009-01-13 01:36:15 +0000
+++ bin/ivle-mountallusers	2009-02-02 04:00:25 +0000
@@ -29,7 +29,7 @@
 import optparse
 import logging
 
-import ivle.db
+import ivle.database
 import ivle.conf
 
 p = optparse.OptionParser()
@@ -49,21 +49,12 @@
                           os.path.basename(sys.argv[0])
     sys.exit(1)
 
-try:
-    db = ivle.db.DB()
-    list = db.get_users()
-    res = db.get_all('login', ['login', 'unixid'])
-    def repack(flds):
-        return (flds['login'], flds['unixid'])
-    uids = dict(map(repack,res))
-except Exception, message:
-    logging.error(str(message))
-    sys.exit(1)
+store = ivle.database.get_store()
+users = store.find(ivle.database.User).order_by(ivle.database.User.login)
 
 logging.info("mass aufs mount started")
 
-list.sort(key=lambda user: user.login)
-for user in list:
+for user in users:
     login = user.login
     # This is where we'll mount to...
     destdir = os.path.join(ivle.conf.jail_base, login)

=== modified file 'bin/ivle-remakeuser'
--- bin/ivle-remakeuser	2009-01-13 01:36:15 +0000
+++ bin/ivle-remakeuser	2009-02-02 04:00:25 +0000
@@ -29,7 +29,7 @@
 import optparse
 import logging
 
-import ivle.db
+import ivle.database
 import ivle.makeuser
 
 p = optparse.OptionParser()
@@ -48,42 +48,25 @@
                           os.path.basename(sys.argv[0])
     sys.exit(1)
 
-try:
-    db = ivle.db.DB()
-    if options.all:
-        list = db.get_users()
-    else:
-        if len(arguments) == 0:
-            print >> sys.stderr, "must be run with -a or a username"
-            sys.exit(1)
-        list = [db.get_user(arguments[0])]
-    res = db.get_all('login', ['login', 'unixid'])
-    def repack(flds):
-        return (flds['login'], flds['unixid'])
-    uids = dict(map(repack,res))
-except Exception, message:
-    logging.error(str(message))
-    sys.exit(1)
+store = ivle.database.get_store()
+
+if options.all:
+    users = store.find(ivle.database.User).order_by(ivle.database.User.login)
+else:
+    if len(arguments) == 0:
+        print >> sys.stderr, "must be run with -a or a username"
+        sys.exit(1)
+    users = [ivle.database.User.get_by_login(store, arguments[0])]
 
 logging.info("rebuild started")
 
-list.sort(key=lambda user: user.login)
-for user in list:
-    login = user.login
-
+for user in users:
     try:
-        # Resolve the user's login into a UID
-        try:
-            uid = uids[login]
-        except KeyError:
-            raise Exception("user %s does not have a unixid in the database"
-                % login)
-        # Remake the user's jail
-        ivle.makeuser.make_jail(login, uid)
+        ivle.makeuser.make_jail(user)
     except Exception, message:
         logging.warning(str(message))
         continue
 
-    logging.debug("recreated user %s's jail." % login)
+    logging.debug("recreated user %s's jail." % user.login)
     
 logging.info("rebuild completed successfully")

=== modified file 'bin/ivle-showenrolment'
--- bin/ivle-showenrolment	2009-01-13 01:36:15 +0000
+++ bin/ivle-showenrolment	2009-02-02 04:00:25 +0000
@@ -25,26 +25,28 @@
 import os
 import sys
 
-import ivle.db
+import ivle.database
 
 if len(sys.argv) != 2:
     print >> sys.stderr, "usage: %s <login>" % os.path.basename(sys.argv[0])
     sys.exit(1)
 
-db = ivle.db.DB()
-try:
-    user = db.get_user(sys.argv[1])
-except ivle.db.DBException:
-    print >> sys.stderr, "cannot retrieve user - probably doesn't exist"
+store = ivle.database.get_store()
+user = ivle.database.User.get_by_login(store, sys.argv[1])
+
+if not user:
+    print>>sys.stderr, "%s: user does not exist"%os.path.basename(sys.argv[0])
     sys.exit(1)
-enrols = db.get_enrolment(sys.argv[1])
-db.close()
 
-if len(enrols) > 0:
-    print 'Showing enrolment for %s (%s):' % (sys.argv[1], user.fullname)
+if user.enrolments.count() > 0:
+    print 'IVLE enrolment for %s (%s):' % (user.login, user.fullname)
     print '    Code       Name Semester Year'
     print '-------- ---------- -------- ----'
-    for enrol in enrols:
-        print '%(subj_code)8s %(subj_short_name)10s %(semester)8s %(year)4s' % enrol
+    for e in user.enrolments:
+        print '%8s %10s %8s %4s' % (
+                e.offering.subject.code, e.offering.subject.short_name,
+                e.offering.semester.semester, e.offering.semester.year
+                )
 else:
-    print '%s (%s) is not enrolled in any offerings' % (sys.argv[1], user.fullname)
+    print '%s (%s) is not enrolled in any IVLE offerings' % (
+                        user.login, user.fullname)

=== modified file 'doc/setup/install_proc.txt'
--- doc/setup/install_proc.txt	2009-01-13 01:36:15 +0000
+++ doc/setup/install_proc.txt	2009-01-14 22:42:10 +0000
@@ -23,7 +23,8 @@
                      apache2 libapache2-mod-python libapache2-svn       \
                      python2.5-dev python-svn python-webpy python-cjson \
                      postgresql python-pygresql php5 php5-pgsql         \
-                     python-docutils python-epydoc python-ldap
+                     python-docutils python-epydoc python-ldap          \
+                     python-storm python-psycopg2
 
 ##########################################################################
 # Configure postgres

=== modified file 'ivle/auth/__init__.py'
--- ivle/auth/__init__.py	2009-01-13 01:36:15 +0000
+++ ivle/auth/__init__.py	2009-01-14 23:23:12 +0000
@@ -0,0 +1,21 @@
+# IVLE - Informatics Virtual Learning Environment
+# Copyright (C) 2007-2008 The University of Melbourne
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+
+class AuthError(Exception):
+    def __init__(self, message="Invalid username or password."):
+        self.message = message
+        self.args = (message,)

=== modified file 'ivle/auth/authenticate.py'
--- ivle/auth/authenticate.py	2009-01-13 01:36:15 +0000
+++ ivle/auth/authenticate.py	2009-01-15 01:06:33 +0000
@@ -25,8 +25,8 @@
 # Has a plugin interface for authentication modules.
 # An authentication module is a Python module with an "auth" function,
 # which accepts 3 positional arguments.
-# plugin_module.auth(dbconn, login, password, user)
-# dbconn is an open connection to the IVLE database.
+# plugin_module.auth(store, login, password, user)
+# store is an open store connected to the IVLE database.
 # login and password are required strings, password is cleartext.
 # user is a User object or None.
 # If it's a User object, it must return the same object if it returns a user.
@@ -43,18 +43,20 @@
 import sys
 import os
 
-from autherror import AuthError
+from ivle.auth import AuthError
 import ivle.conf
-import ivle.db
-import ivle.user
+import ivle.database
 
-def authenticate(login, password):
+def authenticate(store, login, password):
     """Determines whether a particular login/password combination is
-    valid. The password is in cleartext.
+    valid for the given database. The password is in cleartext.
 
     Returns a User object containing the user's details on success.
     Raises an AuthError containing an appropriate error message on failure.
 
+    'store' is expected to be a storm.store.Store connected to the IVLE
+    database to which we should authenticate.
+
     The User returned is guaranteed to be in the IVLE database.
     This could be from reading or writing to the DB. If authenticate can't
     find the user in the DB, it may get user data from some other source
@@ -68,47 +70,37 @@
     # Spawn a DB object just for making this call.
     # (This should not spawn a DB connection on each page reload, only when
     # there is no session object to begin with).
-    dbconn = ivle.db.DB()
-
-    try:
-        user = dbconn.get_user(login)
-    except ivle.db.DBException:
-        # If our attempt to get the named user from the db fails,
-        # then set user to None.
-        # We may still auth (eg. by pulling details from elsewhere and writing
-        # to DB).
-        user = None
-    try:
-        for modname, m in auth_modules:
-            # May raise an AuthError - allow to propagate
-            auth_result = m(dbconn, login, password, user)
-            if auth_result is None:
-                # Can't auth with this module; try another
-                pass
-            elif auth_result == False:
-                return None
-            elif isinstance(auth_result, ivle.user.User):
-                if user is not None and auth_result is not user:
-                    # If user is not None, then it must return the same user
-                    raise AuthError("Internal error: "
-                        "Bad authentication module %s (changed user)"
-                        % repr(modname))
-                elif user is None:
-                    # We just got ourselves some user details from an external
-                    # source. Put them in the DB.
-                    dbconn.create_user(auth_result)
-                    pass
-                return auth_result
-            else:
+
+    user = ivle.database.User.get_by_login(store, login)
+
+    for modname, m in auth_modules:
+        # May raise an AuthError - allow to propagate
+        auth_result = m(store, login, password, user)
+        if auth_result is None:
+            # Can't auth with this module; try another
+            pass
+        elif auth_result == False:
+            return None
+        elif isinstance(auth_result, ivle.database.User):
+            if user is not None and auth_result is not user:
+                # If user is not None, then it must return the same user
                 raise AuthError("Internal error: "
-                    "Bad authentication module %s (bad return type)"
+                    "Bad authentication module %s (changed user)"
                     % repr(modname))
-        # No auths checked out; fail.
-        raise AuthError()
-    finally:
-        dbconn.close()
+            elif user is None:
+                # We just got ourselves some user details from an external
+                # source. Put them in the DB.
+                store.add(auth_result)
+                pass
+            return auth_result
+        else:
+            raise AuthError("Internal error: "
+                "Bad authentication module %s (bad return type)"
+                % repr(modname))
+    # No auths checked out; fail.
+    raise AuthError()
 
-def simple_db_auth(dbconn, login, password, user):
+def simple_db_auth(store, login, password, user):
     """
     A plugin auth function, as described above.
     This one just authenticates against the local database.
@@ -120,7 +112,11 @@
         # The login doesn't exist. Therefore return None so we can try other
         # means of authentication.
         return None
-    auth_result = dbconn.user_authenticate(login, password)
+
+    # They should always match, but it's best to be sure!
+    assert(user.login == login)
+
+    auth_result = user.authenticate(password)
     # auth_result is either True, False (fail) or None (try another)
     if auth_result is None:
         return None

=== removed file 'ivle/auth/autherror.py'
--- ivle/auth/autherror.py	2009-01-13 01:36:15 +0000
+++ ivle/auth/autherror.py	1970-01-01 00:00:00 +0000
@@ -1,25 +0,0 @@
-# IVLE - Informatics Virtual Learning Environment
-# Copyright (C) 2007-2008 The University of Melbourne
-#
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
-
-# Module: auth error
-# Author: Matt Giuca
-# Date:   21/2/2008
-
-class AuthError(Exception):
-    def __init__(self, message="Invalid username or password."):
-        self.message = message
-        self.args = (message,)

=== modified file 'ivle/auth/guest_auth.py'
--- ivle/auth/guest_auth.py	2009-01-13 01:36:15 +0000
+++ ivle/auth/guest_auth.py	2009-01-15 01:06:33 +0000
@@ -30,14 +30,14 @@
 # This module is NOT active implicitly - it must be specified in the
 # "auth_modules" config string.
 
-from autherror import AuthError
-from ivle.user import User
+from ivle.auth import AuthError
+from ivle.database import User
 from ivle.caps import Role
 
 # XXX: What to put here
 GUEST_UID = 4000
 
-def auth(dbconn, login, password, user):
+def auth(store, login, password, user):
     """
     A plugin auth function, as described above.
     This one authenticates against a "guest"/"guest" account.
@@ -51,5 +51,5 @@
         return user
 
     # Create a guest user
-    return User(login="guest", fullname="Guest Account", nick="Guest",
-        role=Role.ANYONE, state="no_agreement", unixid=GUEST_UID)
+    return User(login=u"guest", fullname=u"Guest Account", nick=u"Guest",
+        role=Role.ANYONE, state=u"no_agreement", unixid=GUEST_UID)

=== modified file 'ivle/auth/ldap_auth.py'
--- ivle/auth/ldap_auth.py	2009-01-13 01:36:15 +0000
+++ ivle/auth/ldap_auth.py	2009-01-15 01:06:33 +0000
@@ -27,10 +27,10 @@
 
 import ldap
 
-from autherror import AuthError
+from ivle.auth import AuthError
 import ivle.conf
 
-def auth(dbconn, login, password, user):
+def auth(store, login, password, user):
     """
     A plugin auth function, as described above.
     This one authenticates against an LDAP server.

=== modified file 'ivle/caps.py'
--- ivle/caps.py	2009-01-13 01:36:15 +0000
+++ ivle/caps.py	2009-01-14 07:17:29 +0000
@@ -50,7 +50,7 @@
         # Internally, stored as an int
         if isinstance(role, Role):
             self._role = role._role
-        elif isinstance(role, str):
+        elif isinstance(role, basestring):
             try:
                 self._role = Role._roles_to_int[role.lower()]
             except KeyError:

=== modified file 'ivle/conf/apps.py'
--- ivle/conf/apps.py	2009-01-13 01:36:15 +0000
+++ ivle/conf/apps.py	2009-01-14 23:21:43 +0000
@@ -163,6 +163,11 @@
                     requireauth = True,
                     hashelp = False)
 
+app_logout = App(dir = "logout",
+                    name = "Logout",
+                    requireauth = True,
+                    hashelp = False)
+
 # Mapping URL names to apps
 
 app_url = {
@@ -184,6 +189,7 @@
     "svnlog" : app_svnlog,
     "subjects" : app_subjects,
     "home" : app_home,
+    "logout" : app_logout,
 }
 if enable_debuginfo:
     app_url["debuginfo"] = app_debuginfo

=== modified file 'ivle/database.py'
--- ivle/database.py	2009-01-14 06:52:36 +0000
+++ ivle/database.py	2009-01-20 02:49:22 +0000
@@ -24,9 +24,30 @@
 It also provides miscellaneous utility functions for database interaction.
 """
 
-from storm.locals import create_database, Store
+import md5
+import datetime
+
+from storm.locals import create_database, Store, Int, Unicode, DateTime, \
+                         Reference, ReferenceSet, Bool, Storm, Desc
 
 import ivle.conf
+import ivle.caps
+
+__all__ = ['get_store',
+            'User',
+            'Subject', 'Semester', 'Offering', 'Enrolment',
+            'ProjectSet', 'Project', 'ProjectGroup', 'ProjectGroupMembership',
+            'Exercise', 'Worksheet', 'WorksheetExercise',
+            'ExerciseSave', 'ExerciseAttempt',
+            'AlreadyEnrolledError'
+        ]
+
+def _kwarg_init(self, **kwargs):
+    for k,v in kwargs.items():
+        if k.startswith('_') or not hasattr(self.__class__, k):
+            raise TypeError("%s got an unexpected keyword argument '%s'"
+                % (self.__class__.__name__, k))
+        setattr(self, k, v)
 
 def get_conn_string():
     """
@@ -42,3 +63,451 @@
     instance connected to the configured IVLE database.
     """
     return Store(create_database(get_conn_string()))
+
+# USERS #
+
+class User(Storm):
+    """
+    Represents an IVLE user.
+    """
+    __storm_table__ = "login"
+
+    id = Int(primary=True, name="loginid")
+    login = Unicode()
+    passhash = Unicode()
+    state = Unicode()
+    rolenm = Unicode()
+    unixid = Int()
+    nick = Unicode()
+    pass_exp = DateTime()
+    acct_exp = DateTime()
+    last_login = DateTime()
+    svn_pass = Unicode()
+    email = Unicode()
+    fullname = Unicode()
+    studentid = Unicode()
+    settings = Unicode()
+
+    def _get_role(self):
+        if self.rolenm is None:
+            return None
+        return ivle.caps.Role(self.rolenm)
+    def _set_role(self, value):
+        if not isinstance(value, ivle.caps.Role):
+            raise TypeError("role must be an ivle.caps.Role")
+        self.rolenm = unicode(value)
+    role = property(_get_role, _set_role)
+
+    __init__ = _kwarg_init
+
+    def __repr__(self):
+        return "<%s '%s'>" % (type(self).__name__, self.login)
+
+    def authenticate(self, password):
+        """Validate a given password against this user.
+
+        Returns True if the given password matches the password hash for this
+        User, False if it doesn't match, and None if there is no hash for the
+        user.
+        """
+        if self.passhash is None:
+            return None
+        return self.hash_password(password) == self.passhash
+
+    def hasCap(self, capability):
+        """Given a capability (which is a Role object), returns True if this
+        User has that capability, False otherwise.
+        """
+        return self.role.hasCap(capability)
+
+    @property
+    def password_expired(self):
+        fieldval = self.pass_exp
+        return fieldval is not None and datetime.datetime.now() > fieldval
+
+    @property
+    def account_expired(self):
+        fieldval = self.acct_exp
+        return fieldval is not None and datetime.datetime.now() > fieldval
+
+    def _get_enrolments(self, justactive):
+        return Store.of(self).find(Enrolment,
+            Enrolment.user_id == self.id,
+            (Enrolment.active == True) if justactive else True,
+            Enrolment.offering_id == Offering.id,
+            Offering.semester_id == Semester.id,
+            Offering.subject_id == Subject.id).order_by(
+                Desc(Semester.year),
+                Desc(Semester.semester),
+                Desc(Subject.code)
+            )
+
+    def _set_password(self, password):
+        if password is None:
+            self.passhash = None
+        else:
+            self.passhash = unicode(User.hash_password(password))
+    password = property(fset=_set_password)
+
+    @property
+    def subjects(self):
+        return Store.of(self).find(Subject,
+            Enrolment.user_id == self.id,
+            Enrolment.active == True,
+            Offering.id == Enrolment.offering_id,
+            Subject.id == Offering.subject_id).config(distinct=True)
+
+    # TODO: Invitations should be listed too?
+    def get_groups(self, offering=None):
+        preds = [
+            ProjectGroupMembership.user_id == self.id,
+            ProjectGroup.id == ProjectGroupMembership.project_group_id,
+        ]
+        if offering:
+            preds.extend([
+                ProjectSet.offering_id == offering.id,
+                ProjectGroup.project_set_id == ProjectSet.id,
+            ])
+        return Store.of(self).find(ProjectGroup, *preds)
+
+    @property
+    def groups(self):
+        return self.get_groups()
+
+    @property
+    def active_enrolments(self):
+        '''A sanely ordered list of the user's active enrolments.'''
+        return self._get_enrolments(True)
+
+    @property
+    def enrolments(self):
+        '''A sanely ordered list of all of the user's enrolments.'''
+        return self._get_enrolments(False) 
+
+    @staticmethod
+    def hash_password(password):
+        return md5.md5(password).hexdigest()
+
+    @classmethod
+    def get_by_login(cls, store, login):
+        """
+        Get the User from the db associated with a given store and
+        login.
+        """
+        return store.find(cls, cls.login == unicode(login)).one()
+
+# SUBJECTS AND ENROLMENTS #
+
+class Subject(Storm):
+    __storm_table__ = "subject"
+
+    id = Int(primary=True, name="subjectid")
+    code = Unicode(name="subj_code")
+    name = Unicode(name="subj_name")
+    short_name = Unicode(name="subj_short_name")
+    url = Unicode()
+
+    offerings = ReferenceSet(id, 'Offering.subject_id')
+
+    __init__ = _kwarg_init
+
+    def __repr__(self):
+        return "<%s '%s'>" % (type(self).__name__, self.short_name)
+
+class Semester(Storm):
+    __storm_table__ = "semester"
+
+    id = Int(primary=True, name="semesterid")
+    year = Unicode()
+    semester = Unicode()
+    active = Bool()
+
+    offerings = ReferenceSet(id, 'Offering.semester_id')
+
+    __init__ = _kwarg_init
+
+    def __repr__(self):
+        return "<%s %s/%s>" % (type(self).__name__, self.year, self.semester)
+
+class Offering(Storm):
+    __storm_table__ = "offering"
+
+    id = Int(primary=True, name="offeringid")
+    subject_id = Int(name="subject")
+    subject = Reference(subject_id, Subject.id)
+    semester_id = Int(name="semesterid")
+    semester = Reference(semester_id, Semester.id)
+    groups_student_permissions = Unicode()
+
+    enrolments = ReferenceSet(id, 'Enrolment.offering_id')
+    members = ReferenceSet(id,
+                           'Enrolment.offering_id',
+                           'Enrolment.user_id',
+                           'User.id')
+    project_sets = ReferenceSet(id, 'ProjectSet.offering_id')
+
+    __init__ = _kwarg_init
+
+    def __repr__(self):
+        return "<%s %r in %r>" % (type(self).__name__, self.subject,
+                                  self.semester)
+
+    def enrol(self, user):
+        '''Enrol a user in this offering.'''
+        # We'll get a horrible database constraint violation error if we try
+        # to add a second enrolment.
+        if Store.of(self).find(Enrolment,
+                               Enrolment.user_id == user.id,
+                               Enrolment.offering_id == self.id).count() == 1:
+            raise AlreadyEnrolledError()
+
+        e = Enrolment(user=user, offering=self, active=True)
+        self.enrolments.add(e)
+
+class Enrolment(Storm):
+    __storm_table__ = "enrolment"
+    __storm_primary__ = "user_id", "offering_id"
+
+    user_id = Int(name="loginid")
+    user = Reference(user_id, User.id)
+    offering_id = Int(name="offeringid")
+    offering = Reference(offering_id, Offering.id)
+    notes = Unicode()
+    active = Bool()
+
+    @property
+    def groups(self):
+        return Store.of(self).find(ProjectGroup,
+                ProjectSet.offering_id == self.offering.id,
+                ProjectGroup.project_set_id == ProjectSet.id,
+                ProjectGroupMembership.project_group_id == ProjectGroup.id,
+                ProjectGroupMembership.user_id == self.user.id)
+
+    __init__ = _kwarg_init
+
+    def __repr__(self):
+        return "<%s %r in %r>" % (type(self).__name__, self.user,
+                                  self.offering)
+
+class AlreadyEnrolledError(Exception):
+    pass
+
+# PROJECTS #
+
+class ProjectSet(Storm):
+    __storm_table__ = "project_set"
+
+    id = Int(name="projectsetid", primary=True)
+    offering_id = Int(name="offeringid")
+    offering = Reference(offering_id, Offering.id)
+    max_students_per_group = Int()
+
+    projects = ReferenceSet(id, 'Project.project_set_id')
+    project_groups = ReferenceSet(id, 'ProjectGroup.project_set_id')
+
+    __init__ = _kwarg_init
+
+    def __repr__(self):
+        return "<%s %d in %r>" % (type(self).__name__, self.id,
+                                  self.offering)
+
+class Project(Storm):
+    __storm_table__ = "project"
+
+    id = Int(name="projectid", primary=True)
+    synopsis = Unicode()
+    url = Unicode()
+    project_set_id = Int(name="projectsetid")
+    project_set = Reference(project_set_id, ProjectSet.id)
+    deadline = DateTime()
+
+    __init__ = _kwarg_init
+
+    def __repr__(self):
+        return "<%s '%s' in %r>" % (type(self).__name__, self.synopsis,
+                                  self.project_set.offering)
+
+class ProjectGroup(Storm):
+    __storm_table__ = "project_group"
+
+    id = Int(name="groupid", primary=True)
+    name = Unicode(name="groupnm")
+    project_set_id = Int(name="projectsetid")
+    project_set = Reference(project_set_id, ProjectSet.id)
+    nick = Unicode()
+    created_by_id = Int(name="createdby")
+    created_by = Reference(created_by_id, User.id)
+    epoch = DateTime()
+
+    members = ReferenceSet(id,
+                           "ProjectGroupMembership.project_group_id",
+                           "ProjectGroupMembership.user_id",
+                           "User.id")
+
+    __init__ = _kwarg_init
+
+    def __repr__(self):
+        return "<%s %s in %r>" % (type(self).__name__, self.name,
+                                  self.project_set.offering)
+
+class ProjectGroupMembership(Storm):
+    __storm_table__ = "group_member"
+    __storm_primary__ = "user_id", "project_group_id"
+
+    user_id = Int(name="loginid")
+    user = Reference(user_id, User.id)
+    project_group_id = Int(name="groupid")
+    project_group = Reference(project_group_id, ProjectGroup.id)
+
+    __init__ = _kwarg_init
+
+    def __repr__(self):
+        return "<%s %r in %r>" % (type(self).__name__, self.user,
+                                  self.project_group)
+
+# WORKSHEETS AND EXERCISES #
+
+class Exercise(Storm):
+    # Note: Table "problem" is called "Exercise" in the Object layer, since
+    # it's called that everywhere else.
+    __storm_table__ = "problem"
+
+    id = Int(primary=True, name="problemid")
+    name = Unicode(name="identifier")
+    spec = Unicode()
+
+    worksheets = ReferenceSet(id,
+        'WorksheetExercise.exercise_id',
+        'WorksheetExercise.worksheet_id',
+        'Worksheet.id'
+    )
+
+    __init__ = _kwarg_init
+
+    def __repr__(self):
+        return "<%s %s>" % (type(self).__name__, self.name)
+
+    @classmethod
+    def get_by_name(cls, store, name):
+        """
+        Get the Exercise from the db associated with a given store and name.
+        If the exercise is not in the database, creates it and inserts it
+        automatically.
+        """
+        ex = store.find(cls, cls.name == unicode(name)).one()
+        if ex is not None:
+            return ex
+        ex = Exercise(name=unicode(name))
+        store.add(ex)
+        store.commit()
+        return ex
+
+class Worksheet(Storm):
+    __storm_table__ = "worksheet"
+
+    id = Int(primary=True, name="worksheetid")
+    # XXX subject is not linked to a Subject object. This is a property of
+    # the database, and will be refactored.
+    subject = Unicode()
+    name = Unicode(name="identifier")
+    assessable = Bool()
+    mtime = DateTime()
+
+    exercises = ReferenceSet(id,
+        'WorksheetExercise.worksheet_id',
+        'WorksheetExercise.exercise_id',
+        Exercise.id)
+    # Use worksheet_exercises to get access to the WorksheetExercise objects
+    # binding worksheets to exercises. This is required to access the
+    # "optional" field.
+    worksheet_exercises = ReferenceSet(id,
+        'WorksheetExercise.worksheet_id')
+
+    __init__ = _kwarg_init
+
+    def __repr__(self):
+        return "<%s %s>" % (type(self).__name__, self.name)
+
+    # XXX Refactor this - make it an instance method of Subject rather than a
+    # class method of Worksheet. Can't do that now because Subject isn't
+    # linked referentially to the Worksheet.
+    @classmethod
+    def get_by_name(cls, store, subjectname, worksheetname):
+        """
+        Get the Worksheet from the db associated with a given store, subject
+        name and worksheet name.
+        """
+        return store.find(cls, cls.subject == unicode(subjectname),
+            cls.name == unicode(worksheetname)).one()
+
+    def remove_all_exercises(self, store):
+        """
+        Remove all exercises from this worksheet.
+        This does not delete the exercises themselves. It just removes them
+        from the worksheet.
+        """
+        store.find(WorksheetExercise,
+            WorksheetExercise.worksheet == self).remove()
+
+class WorksheetExercise(Storm):
+    __storm_table__ = "worksheet_problem"
+    __storm_primary__ = "worksheet_id", "exercise_id"
+
+    worksheet_id = Int(name="worksheetid")
+    worksheet = Reference(worksheet_id, Worksheet.id)
+    exercise_id = Int(name="problemid")
+    exercise = Reference(exercise_id, Exercise.id)
+    optional = Bool()
+
+    __init__ = _kwarg_init
+
+    def __repr__(self):
+        return "<%s %s in %s>" % (type(self).__name__, self.exercise.name,
+                                  self.worksheet.name)
+
+class ExerciseSave(Storm):
+    """
+    Represents a potential solution to an exercise that a user has submitted
+    to the server for storage.
+    A basic ExerciseSave is just the current saved text for this exercise for
+    this user (doesn't count towards their attempts).
+    ExerciseSave may be extended with additional semantics (such as
+    ExerciseAttempt).
+    """
+    __storm_table__ = "problem_save"
+    __storm_primary__ = "exercise_id", "user_id", "date"
+
+    exercise_id = Int(name="problemid")
+    exercise = Reference(exercise_id, Exercise.id)
+    user_id = Int(name="loginid")
+    user = Reference(user_id, User.id)
+    date = DateTime()
+    text = Unicode()
+
+    __init__ = _kwarg_init
+
+    def __repr__(self):
+        return "<%s %s by %s at %s>" % (type(self).__name__,
+            self.exercise.name, self.user.login, self.date.strftime("%c"))
+
+class ExerciseAttempt(ExerciseSave):
+    """
+    An ExerciseAttempt is a special case of an ExerciseSave. Like an
+    ExerciseSave, it constitutes exercise solution data that the user has
+    submitted to the server for storage.
+    In addition, it contains additional information about the submission.
+    complete - True if this submission was successful, rendering this exercise
+        complete for this user.
+    active - True if this submission is "active" (usually true). Submissions
+        may be de-activated by privileged users for special reasons, and then
+        they won't count (either as a penalty or success), but will still be
+        stored.
+    """
+    __storm_table__ = "problem_attempt"
+    __storm_primary__ = "exercise_id", "user_id", "date"
+
+    # The "text" field is the same but has a different name in the DB table
+    # for some reason.
+    text = Unicode(name="attempt")
+    complete = Bool()
+    active = Bool()

=== removed file 'ivle/db.py'
--- ivle/db.py	2009-01-13 01:36:15 +0000
+++ ivle/db.py	1970-01-01 00:00:00 +0000
@@ -1,1143 +0,0 @@
-# IVLE - Informatics Virtual Learning Environment
-# Copyright (C) 2007-2008 The University of Melbourne
-#
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
-
-# Module: Database
-# Author: Matt Giuca
-# Date:   15/2/2008
-
-# Code to talk to the PostgreSQL database.
-# (This is the Data Access Layer).
-# All DB code should be in this module to ensure portability if we want to
-# change the DB implementation.
-# This means no SQL strings should be outside of this module. Add functions
-# here to perform the activities needed, and place the SQL code for those
-# activities within.
-
-# CAUTION to editors of this module.
-# All string inputs must be sanitized by calling _escape before being
-# formatted into an SQL query string.
-
-import pg
-import md5
-import copy
-import time
-
-import ivle.conf
-from ivle import (caps, user)
-
-TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M:%S'
-
-def _escape(val):
-    """Wrapper around pg.escape_string. Prepares the Python value for use in
-    SQL. Returns a string, which may be safely placed verbatim into an SQL
-    query.
-    Handles the following types:
-    * str: Escapes the string, and also quotes it.
-    * int/long/float: Just converts to an unquoted string.
-    * bool: Returns as "TRUE" or "FALSE", unquoted.
-    * NoneType: Returns "NULL", unquoted.
-    * common.caps.Role: Returns the role as a quoted, lowercase string.
-    * time.struct_time: Returns the time as a quoted string for insertion into
-        a TIMESTAMP column.
-    Raises a DBException if val has an unsupported type.
-    """
-    # "E'" is postgres's way of making "escape" strings.
-    # Such strings allow backslashes to escape things. Since escape_string
-    # converts a single backslash into two backslashes, it needs to be fed
-    # into E mode.
-    # Ref: http://www.postgresql.org/docs/8.2/static/sql-syntax-lexical.html
-    # WARNING: PostgreSQL-specific code
-    if val is None:
-        return "NULL"
-    elif isinstance(val, str) or isinstance(val, unicode):
-        return "E'" + pg.escape_string(val) + "'"
-    elif isinstance(val, bool):
-        return "TRUE" if val else "FALSE"
-    elif isinstance(val, int) or isinstance(val, long) \
-        or isinstance(val, float):
-        return str(val)
-    elif isinstance(val, caps.Role):
-        return _escape(str(val))
-    elif isinstance(val, time.struct_time):
-        return _escape(time.strftime(TIMESTAMP_FORMAT, val))
-    else:
-        raise DBException("Attempt to insert an unsupported type "
-            "into the database (%s)" % repr(type(val)))
-
-def _parse_boolean(val):
-    """
-    Accepts a boolean as output from the DB (either the string 't' or 'f').
-    Returns a boolean value True or False.
-    Also accepts other values which mean True or False in PostgreSQL.
-    If none match, raises a DBException.
-    """
-    # On a personal note, what sort of a language allows 7 different values
-    # to denote each of True and False?? (A: SQL)
-    if isinstance(val, bool):
-        return val
-    elif val == 't':
-        return True
-    elif val == 'f':
-        return False
-    elif val == 'true' or val == 'y' or val == 'yes' or val == '1' \
-        or val == 1:
-        return True
-    elif val == 'false' or val == 'n' or val == 'no' or val == '0' \
-        or val == 0:
-        return False
-    else:
-        raise DBException("Invalid boolean value returned from DB")
-
-def _passhash(password):
-    return md5.md5(password).hexdigest()
-
-class DBException(Exception):
-    """A DBException is for bad conditions in the database or bad input to
-    these methods. If Postgres throws an exception it does not get rebadged.
-    This is only for additional exceptions."""
-    pass
-
-class DB:
-    """An IVLE database object. This object provides an interface to
-    interacting with the IVLE database without using any external SQL.
-
-    Most methods of this class have an optional dry argument. If true, they
-    will return the SQL query string and NOT actually execute it. (For
-    debugging purposes).
-
-    Methods may throw db.DBException, or any of the pg exceptions as well.
-    (In general, be prepared to catch exceptions!)
-    """
-    def __init__(self):
-        """Connects to the database and creates a DB object.
-        Takes no parameters - gets all the DB info from the configuration."""
-        self.open = False
-        self.db = pg.connect(host=ivle.conf.db_host, port=ivle.conf.db_port,
-                         dbname=ivle.conf.db_dbname,
-                         user=ivle.conf.db_user, passwd=ivle.conf.db_password)
-        self.open = True
-
-    def __del__(self):
-        if self.open:
-            self.db.close()
-
-    # GENERIC DB FUNCTIONS #
-
-    @staticmethod
-    def check_dict(dict, tablefields, disallowed=frozenset([]), must=False):
-        """Checks that a dict does not contain keys that are not fields
-        of the specified table.
-        dict: A mapping from string keys to values; the keys are checked to
-            see that they correspond to login table fields.
-        tablefields: Collection of strings for field names in the table.
-            Only these fields will be allowed.
-        disallowed: Optional collection of strings for field names that are
-            not allowed.
-        must: If True, the dict MUST contain all fields in tablefields.
-            If False, it may contain any subset of the fields.
-        Returns True if the dict is valid, False otherwise.
-        """
-        allowed = frozenset(tablefields) - frozenset(disallowed)
-        dictkeys = frozenset(dict.keys())
-        if must:
-            return allowed == dictkeys
-        else:
-            return allowed.issuperset(dictkeys)
-
-    def insert(self, dict, tablename, tablefields, disallowed=frozenset([]),
-        dry=False):
-        """Inserts a new row in a table, using data from a supplied
-        dictionary (which will be checked by check_dict).
-        dict: Dictionary mapping column names to values. The values may be
-            any of the following types:
-            str, int, long, float, NoneType.
-        tablename: String, name of the table to insert into. Will NOT be
-            escaped - must be a valid identifier.
-        tablefields, disallowed: see check_dict.
-        dry: Returns the SQL query as a string, and does not execute it.
-        Raises a DBException if the dictionary contains invalid fields.
-        """
-        if not DB.check_dict(dict, tablefields, disallowed):
-            extras = set(dict.keys()) - tablefields
-            raise DBException("Supplied dictionary contains invalid fields. (%s)" % (repr(extras)))
-        # Build two lists concurrently: field names and values, as SQL strings
-        fieldnames = []
-        values = []
-        for k,v in dict.items():
-            fieldnames.append(k)
-            values.append(_escape(v))
-        if len(fieldnames) == 0: return
-        fieldnames = ', '.join(fieldnames)
-        values = ', '.join(values)
-        query = ("INSERT INTO %s (%s) VALUES (%s);"
-            % (tablename, fieldnames, values))
-        if dry: return query
-        self.db.query(query)
-
-    def return_insert(self, dict, tablename, tablefields, returning,
-        disallowed=frozenset([]), dry=False):
-        """Inserts a new row in a table, using data from a supplied
-        dictionary (which will be checked by check_dict) and returns certain 
-        fields as a dict.
-        dict: Dictionary mapping column names to values. The values may be
-            any of the following types:
-            str, int, long, float, NoneType.
-        tablename: String, name of the table to insert into. Will NOT be
-            escaped - must be a valid identifier.
-        returning: List of fields to return, not escaped
-        tablefields, disallowed: see check_dict.
-        dry: Returns the SQL query as a string, and does not execute it.
-        Raises a DBException if the dictionary contains invalid fields.
-        """
-        if not DB.check_dict(dict, tablefields, disallowed):
-            extras = set(dict.keys()) - tablefields
-            raise DBException("Supplied dictionary contains invalid fields. (%s)" % (repr(extras)))
-        # Build two lists concurrently: field names and values, as SQL strings
-        fieldnames = []
-        values = []
-        for k,v in dict.items():
-            fieldnames.append(k)
-            values.append(_escape(v))
-        if len(fieldnames) == 0: return
-        fieldnames = ', '.join(fieldnames)
-        values = ', '.join(values)
-        returns = ', '.join(returning)
-        query = ("INSERT INTO %s (%s) VALUES (%s) RETURNING (%s);"
-            % (tablename, fieldnames, values, returns))
-        if dry: return query
-        return self.db.query(query)
-
-
-    def update(self, primarydict, updatedict, tablename, tablefields,
-        primary_keys, disallowed_update=frozenset([]), dry=False):
-        """Updates a row in a table, matching against primarydict to find the
-        row, and using the data in updatedict (which will be checked by
-        check_dict).
-        primarydict: Dict mapping column names to values. The keys should be
-            the table's primary key. Only rows which match this dict's values
-            will be updated.
-        updatedict: Dict mapping column names to values. The columns will be
-            updated with the given values for the matched rows.
-        tablename, tablefields, disallowed_update: See insert.
-        primary_keys: Collection of strings which together form the primary
-            key for this table. primarydict must contain all of these as keys,
-            and only these keys.
-        """
-        if (not (DB.check_dict(primarydict, primary_keys, must=True)
-            and DB.check_dict(updatedict, tablefields, disallowed_update))):
-            raise DBException("Supplied dictionary contains invalid or missing fields (1).")
-        # Make a list of SQL fragments of the form "field = 'new value'"
-        # These fragments are ALREADY-ESCAPED
-        setlist = []
-        for k,v in updatedict.items():
-            setlist.append("%s = %s" % (k, _escape(v)))
-        wherelist = []
-        for k,v in primarydict.items():
-            wherelist.append("%s = %s" % (k, _escape(v)))
-        if len(setlist) == 0 or len(wherelist) == 0:
-            return
-        # Join the fragments into a comma-separated string
-        setstring = ', '.join(setlist)
-        wherestring = ' AND '.join(wherelist)
-        # Build the whole query as an UPDATE statement
-        query = ("UPDATE %s SET %s WHERE %s;"
-            % (tablename, setstring, wherestring))
-        if dry: return query
-        self.db.query(query)
-
-    def delete(self, primarydict, tablename, primary_keys, dry=False):
-        """Deletes a row in the table, matching against primarydict to find
-        the row.
-        primarydict, tablename, primary_keys: See update.
-        """
-        if not DB.check_dict(primarydict, primary_keys, must=True):
-            raise DBException("Supplied dictionary contains invalid or missing fields (2).")
-        wherelist = []
-        for k,v in primarydict.items():
-            wherelist.append("%s = %s" % (k, _escape(v)))
-        if len(wherelist) == 0:
-            return
-        wherestring = ' AND '.join(wherelist)
-        query = ("DELETE FROM %s WHERE %s;" % (tablename, wherestring))
-        if dry: return query
-        self.db.query(query)
-
-    def get_single(self, primarydict, tablename, getfields, primary_keys,
-        error_notfound="No rows found", dry=False):
-        """Retrieves a single row from a table, returning it as a dictionary
-        mapping field names to values. Matches against primarydict to find the
-        row.
-        primarydict, tablename, primary_keys: See update/delete.
-        getfields: Collection of strings; the field names which will be
-            returned as keys in the dictionary.
-        error_notfound: Error message if 0 rows match.
-        Raises a DBException if 0 rows match, with error_notfound as the msg.
-        Raises an AssertError if >1 rows match (this should not happen if
-            primary_keys is indeed the primary key).
-        """
-        if not DB.check_dict(primarydict, primary_keys, must=True):
-            raise DBException("Supplied dictionary contains invalid or missing fields (3).")
-        wherelist = []
-        for k,v in primarydict.items():
-            wherelist.append("%s = %s" % (k, _escape(v)))
-        if len(getfields) == 0 or len(wherelist) == 0:
-            return
-        # Join the fragments into a comma-separated string
-        getstring = ', '.join(getfields)
-        wherestring = ' AND '.join(wherelist)
-        # Build the whole query as an SELECT statement
-        query = ("SELECT %s FROM %s WHERE %s;"
-            % (getstring, tablename, wherestring))
-        if dry: return query
-        result = self.db.query(query)
-        # Expecting exactly one
-        if result.ntuples() != 1:
-            # It should not be possible for ntuples to be greater than 1
-            assert (result.ntuples() < 1)
-            raise DBException(error_notfound)
-        # Return as a dictionary
-        return result.dictresult()[0]
-
-    def get_all(self, tablename, getfields, dry=False):
-        """Retrieves all rows from a table, returning it as a list of
-        dictionaries mapping field names to values.
-        tablename, getfields: See get_single.
-        """
-        if len(getfields) == 0:
-            return
-        getstring = ', '.join(getfields)
-        query = ("SELECT %s FROM %s;" % (getstring, tablename))
-        if dry: return query
-        return self.db.query(query).dictresult()
-
-    def start_transaction(self, dry=False):
-        """Starts a DB transaction.
-        Will not commit any changes until self.commit() is called.
-        """
-        query = "START TRANSACTION;"
-        if dry: return query
-        self.db.query(query)
-
-    def commit(self, dry=False):
-        """Commits (ends) a DB transaction.
-        Commits all changes since the call to start_transaction.
-        """
-        query = "COMMIT;"
-        if dry: return query
-        self.db.query(query)
-
-    def rollback(self, dry=False):
-        """Rolls back (ends) a DB transaction, undoing all changes since the
-        call to start_transaction.
-        """
-        query = "ROLLBACK;"
-        if dry: return query
-        self.db.query(query)
-
-    # USER MANAGEMENT FUNCTIONS #
-
-    login_primary = frozenset(["login"])
-    login_fields_list = [
-        "login", "passhash", "state", "unixid", "email", "nick", "fullname",
-        "rolenm", "studentid", "acct_exp", "pass_exp", "last_login", "svn_pass"
-    ]
-    login_fields = frozenset(login_fields_list)
-
-    def create_user(self, user_obj=None, dry=False, **kwargs):
-        """Creates a user login entry in the database.
-        Two ways to call this - passing a user object, or passing
-        all fields as separate arguments.
-
-        Either pass a "user_obj" as the first argument (in which case other
-        fields will be ignored), or pass all fields as arguments.
-
-        All user fields are to be passed as args. The argument names
-        are the field names of the "login" table of the DB schema.
-        However, instead of supplying a "passhash", you must supply a
-        "password" argument, which will be hashed internally.
-        Also "state" must not given explicitly; it is implicitly set to
-        "no_agreement".
-        Raises an exception if the user already exists, or the dict contains
-        invalid keys or is missing required keys.
-        """
-        if 'passhash' in kwargs:
-            raise DBException("Supplied arguments include passhash (invalid) (1).")
-        # Make a copy of the dict. Change password to passhash (hashing it),
-        # and set 'state' to "no_agreement".
-        if user_obj is None:
-            # Use the kwargs
-            fields = copy.copy(kwargs)
-        else:
-            # Use the user object
-            fields = dict(user_obj)
-        if 'password' in fields:
-            fields['passhash'] = _passhash(fields['password'])
-            del fields['password']
-        if 'role' in fields:
-            # Convert role to rolenm
-            fields['rolenm'] = str(user_obj.role)
-            del fields['role']
-        if user_obj is None:
-            fields['state'] = "no_agreement"
-            # else, we'll trust the user, but it SHOULD be "no_agreement"
-            # (We can't change it because then the user object would not
-            # reflect the DB).
-        if 'local_password' in fields:
-            del fields['local_password']
-        # Execute the query.
-        return self.insert(fields, "login", self.login_fields, dry=dry)
-
-    def update_user(self, login, dry=False, **kwargs):
-        """Updates fields of a particular user. login is the name of the user
-        to update. The dict contains the fields which will be modified, and
-        their new values. If any value is omitted from the dict, it does not
-        get modified. login and studentid may not be modified.
-        Passhash may be modified by supplying a "password" field, in
-        cleartext, not a hashed password.
-
-        Note that no checking is done. It is expected this function is called
-        by a trusted source. In particular, it allows the password to be
-        changed without knowing the old password. The caller should check
-        that the user knows the existing password before calling this function
-        with a new one.
-        """
-        if 'passhash' in kwargs:
-            raise DBException("Supplied arguments include passhash (invalid) (2).")
-        if "password" in kwargs:
-            kwargs = copy.copy(kwargs)
-            kwargs['passhash'] = _passhash(kwargs['password'])
-            del kwargs['password']
-        return self.update({"login": login}, kwargs, "login",
-            self.login_fields, self.login_primary, ["login", "studentid"],
-            dry=dry)
-
-    def get_user(self, login, dry=False):
-        """Given a login, returns a User object containing details looked up
-        in the DB.
-
-        Raises a DBException if the login is not found in the DB.
-        """
-        userdict = self.get_single({"login": login}, "login",
-            self.login_fields, self.login_primary,
-            error_notfound="get_user: No user with that login name", dry=dry)
-        if dry:
-            return userdict     # Query string
-        # Package into a User object
-        return user.User(**userdict)
-
-    def get_users(self, dry=False):
-        """Returns a list of all users in the DB, as User objects.
-        """
-        userdicts = self.get_all("login", self.login_fields, dry=dry)
-        if dry:
-            return userdicts    # Query string
-        # Package into User objects
-        return [user.User(**userdict) for userdict in userdicts]
-
-    def get_user_loginid(self, login, dry=False):
-        """Given a login, returns the integer loginid for this user.
-
-        Raises a DBException if the login is not found in the DB.
-        """
-        userdict = self.get_single({"login": login}, "login",
-            ['loginid'], self.login_primary,
-            error_notfound="get_user_loginid: No user with that login name",
-            dry=dry)
-        if dry:
-            return userdict     # Query string
-        return userdict['loginid']
-
-    def user_authenticate(self, login, password, dry=False):
-        """Performs a password authentication on a user. Returns True if
-        "passhash" is the correct passhash for the given login, False
-        if the passhash does not match the password in the DB,
-        and None if the passhash in the DB is NULL.
-        Also returns False if the login does not exist (so if you want to
-        differentiate these cases, use get_user and catch an exception).
-        """
-        query = ("SELECT passhash FROM login WHERE login = %s;"
-            % _escape(login))
-        if dry: return query
-        result = self.db.query(query)
-        if result.ntuples() == 1:
-            # Valid username. Check password.
-            passhash = result.getresult()[0][0]
-            if passhash is None:
-                return None
-            return _passhash(password) == passhash
-        else:
-            return False
-
-    # PROBLEM AND PROBLEM ATTEMPT FUNCTIONS #
-
-    def get_problem_problemid(self, exercisename, dry=False):
-        """Given an exercise name, returns the associated problemID.
-        If the exercise name is NOT in the database, it inserts it and returns
-        the new problemID. Hence this may mutate the DB, but is idempotent.
-        """
-        try:
-            d = self.get_single({"identifier": exercisename}, "problem",
-                ['problemid'], frozenset(["identifier"]),
-                dry=dry)
-            if dry:
-                return d        # Query string
-        except DBException:
-            if dry:
-                # Shouldn't try again, must have failed for some other reason
-                raise
-            # if we failed to get a problemid, it was probably because
-            # the exercise wasn't in the db. So lets insert it!
-            #
-            # The insert can fail if someone else simultaneously does
-            # the insert, so if the insert fails, we ignore the problem. 
-            try:
-                self.insert({'identifier': exercisename}, "problem",
-                        frozenset(['identifier']))
-            except Exception, e:
-                pass
-
-            # Assuming the insert succeeded, we should be able to get the
-            # problemid now.
-            d = self.get_single({"identifier": exercisename}, "problem",
-                ['problemid'], frozenset(["identifier"]))
-
-        return d['problemid']
-
-    def insert_problem_attempt(self, login, exercisename, date, complete,
-        attempt, dry=False):
-        """Inserts the details of a problem attempt into the database.
-        exercisename: Name of the exercise. (identifier field of problem
-            table). If this exercise does not exist, also creates a new row in
-            the problem table for this exercise name.
-        login: Name of the user submitting the attempt. (login field of the
-            login table).
-        date: struct_time, the date this attempt was made.
-        complete: bool. Whether the test passed or not.
-        attempt: Text of the attempt.
-
-        Note: Even if dry, will still physically call get_problem_problemid,
-        which may mutate the DB, and get_user_loginid, which may fail.
-        """
-        problemid = self.get_problem_problemid(exercisename)
-        loginid = self.get_user_loginid(login)  # May raise a DBException
-
-        return self.insert({
-                'problemid': problemid,
-                'loginid': loginid,
-                'date': date,
-                'complete': complete,
-                'attempt': attempt,
-            }, 'problem_attempt',
-            frozenset(['problemid','loginid','date','complete','attempt']),
-            dry=dry)
-
-    def write_problem_save(self, login, exercisename, date, text, dry=False):
-        """Writes text to the problem_save table (for when the user saves an
-        exercise). Creates a new row, or overwrites an existing one if the
-        user has already saved that problem.
-        (Unlike problem_attempt, does not keep historical records).
-        """
-        problemid = self.get_problem_problemid(exercisename)
-        loginid = self.get_user_loginid(login)  # May raise a DBException
-
-        try:
-            return self.insert({
-                    'problemid': problemid,
-                    'loginid': loginid,
-                    'date': date,
-                    'text': text,
-                }, 'problem_save',
-                frozenset(['problemid','loginid','date','text']),
-                dry=dry)
-        except pg.ProgrammingError:
-            # May have failed because this problemid/loginid row already
-            # exists (they have a unique key constraint).
-            # Do an update instead.
-            if dry:
-                # Shouldn't try again, must have failed for some other reason
-                raise
-            self.update({
-                    'problemid': problemid,
-                    'loginid': loginid,
-                },
-                {
-                    'date': date,
-                    'text': text,
-                }, "problem_save",
-                frozenset(['date', 'text']),
-                frozenset(['problemid', 'loginid']))
-
-    def get_problem_stored_text(self, login, exercisename, dry=False):
-        """Given a login name and exercise name, returns the text of the
-        last saved/submitted attempt for this question.
-        Returns None if the user has not saved or made an attempt on this
-        problem.
-        (If the user has both saved and submitted, it returns whichever was
-        made last).
-
-        Note: Even if dry, will still physically call get_problem_problemid,
-        which may mutate the DB, and get_user_loginid, which may fail.
-        """
-        problemid = self.get_problem_problemid(exercisename)
-        loginid = self.get_user_loginid(login)  # May raise a DBException
-        # This very complex query finds all submissions made by this user for
-        # this problem, as well as the save made by this user for this
-        # problem, and returns the text of the newest one.
-        # (Whichever is newer out of the save or the submit).
-        query = """SELECT text FROM
-    (
-        (SELECT * FROM problem_save WHERE loginid = %d AND problemid = %d)
-    UNION
-        (SELECT problemid, loginid, date, text FROM problem_attempt
-         AS problem_attempt (problemid, loginid, date, text)
-         WHERE loginid = %d AND problemid = %d AND active)
-    )
-    AS _
-    ORDER BY date DESC
-    LIMIT 1;""" % (loginid, problemid, loginid, problemid)
-        if dry: return query
-        result = self.db.query(query)
-        if result.ntuples() == 1:
-            # The user has made at least 1 attempt. Return the newest.
-            return result.getresult()[0][0]
-        else:
-            return None
-
-    def get_problem_attempts(self, login, exercisename, allow_inactive=True,
-                             dry=False):
-        """Given a login name and exercise name, returns a list of dicts, one
-        for each attempt made for that exercise.
-        Dicts are {'date': 'formatted_time', 'complete': bool}.
-        Ordered with the newest first.
-        
-        Note: By default, returns de-activated problem attempts (unlike
-        get_problem_stored_text).
-        If allow_inactive is False, will not return disabled attempts.
-
-        Note: Even if dry, will still physically call get_problem_problemid,
-        which may mutate the DB, and get_user_loginid, which may fail.
-        """
-        problemid = self.get_problem_problemid(exercisename)
-        loginid = self.get_user_loginid(login)  # May raise a DBException
-        andactive = '' if allow_inactive else ' AND active'
-        query = """SELECT date, complete FROM problem_attempt
-    WHERE loginid = %d AND problemid = %d%s
-    ORDER BY date DESC;""" % (loginid, problemid, andactive)
-        if dry: return query
-        result = self.db.query(query).getresult()
-        # Make into dicts (could use dictresult, but want to convert values)
-        return [{'date': date, 'complete': _parse_boolean(complete)}
-                for date, complete in result]
-
-    def get_problem_attempt(self, login, exercisename, as_of,
-        allow_inactive=True, dry=False):
-        """Given a login name, exercise name, and struct_time, returns the
-        text of the submitted attempt for this question as of that date.
-        Returns None if the user had not made an attempt on this problem at
-        that date.
-        
-        Note: By default, returns de-activated problem attempts (unlike
-        get_problem_stored_text).
-        If allow_inactive is False, will not return disabled attempts.
-
-        Note: Even if dry, will still physically call get_problem_problemid,
-        which may mutate the DB, and get_user_loginid, which may fail.
-        """
-        problemid = self.get_problem_problemid(exercisename)
-        loginid = self.get_user_loginid(login)  # May raise a DBException
-        # Very similar to query in get_problem_stored_text, but without
-        # looking in problem_save, and restricting to a certain date.
-        andactive = '' if allow_inactive else ' AND active'
-        query = """SELECT attempt FROM problem_attempt
-    WHERE loginid = %d AND problemid = %d%s AND date <= %s
-    ORDER BY date DESC
-    LIMIT 1;""" % (loginid, problemid, andactive, _escape(as_of))
-        if dry: return query
-        result = self.db.query(query)
-        if result.ntuples() == 1:
-            # The user has made at least 1 attempt. Return the newest.
-            return result.getresult()[0][0]
-        else:
-            return None
-
-    def get_problem_status(self, login, exercisename, dry=False):
-        """Given a login name and exercise name, returns information about the
-        user's performance on that problem.
-        Returns a tuple of:
-            - A boolean, whether they have successfully passed this exercise.
-            - An int, the number of attempts they have made up to and
-              including the first successful attempt (or the total number of
-              attempts, if not yet successful).
-        Note: exercisename may be an int, in which case it will be directly
-        used as the problemid.
-        """
-        if isinstance(exercisename, int):
-            problemid = exercisename
-        else:
-            problemid = self.get_problem_problemid(exercisename)
-        loginid = self.get_user_loginid(login)  # May raise a DBException
-
-        # ASSUME that it is completed, get the total number of attempts up to
-        # and including the first successful attempt.
-        # (Get the date of the first successful attempt. Then count the number
-        # of attempts made <= that date).
-        # Will return an empty table if the problem has never been
-        # successfully completed.
-        query = """SELECT COUNT(*) FROM problem_attempt
-    WHERE loginid = %d AND problemid = %d AND active AND date <=
-        (SELECT date FROM problem_attempt
-            WHERE loginid = %d AND problemid = %d AND complete AND active
-            ORDER BY date ASC
-            LIMIT 1);""" % (loginid, problemid, loginid, problemid)
-        if dry: return query
-        result = self.db.query(query)
-        count = int(result.getresult()[0][0])
-        if count > 0:
-            # The user has made at least 1 successful attempt.
-            # Return True for success, and the number of attempts up to and
-            # including the successful one.
-            return (True, count)
-        else:
-            # Returned 0 rows - this indicates that the problem has not been
-            # completed.
-            # Return the total number of attempts, and False for success.
-            query = """SELECT COUNT(*) FROM problem_attempt
-    WHERE loginid = %d AND problemid = %d AND active;""" % (loginid, problemid)
-            result = self.db.query(query)
-            count = int(result.getresult()[0][0])
-            return (False, count)
-
-    # WORKSHEET/PROBLEM ASSOCIATION AND MARKS CALCULATION
-
-    def get_worksheet_mtime(self, subject, worksheet, dry=False):
-        """
-        For a given subject/worksheet name, gets the time the worksheet was
-        last updated in the DB, if any.
-        This can be used to check if there is a newer version on disk.
-        Returns the timestamp as a time.struct_time, or None if the worksheet
-        is not found or has no stored mtime.
-        """
-        try:
-            r = self.get_single(
-                {"subject": subject, "identifier": worksheet},
-                "worksheet", ["mtime"], ["subject", "identifier"],
-                dry=dry)
-        except DBException:
-            # Assume the worksheet is not in the DB
-            return None
-        if dry:
-            return r
-        if r["mtime"] is None:
-            return None
-        return time.strptime(r["mtime"], TIMESTAMP_FORMAT)
-
-    def create_worksheet(self, subject, worksheet, problems=None,
-        assessable=None):
-        """
-        Inserts or updates rows in the worksheet and worksheet_problems
-        tables, to create a worksheet in the database.
-        This atomically performs all operations. If the worksheet is already
-        in the DB, removes it and all its associated problems and rebuilds.
-        Sets the timestamp to the current time.
-
-        problems is a collection of pairs. The first element of the pair is
-        the problem identifier ("identifier" column of the problem table). The
-        second element is an optional boolean, "optional". This can be omitted
-        (so it's a 1-tuple), and then it will default to False.
-
-        Problems and assessable are optional, and if omitted, will not change
-        the existing data. If the worksheet does not yet exist, and assessable
-        is omitted, it defaults to False.
-
-        Note: As with get_problem_problemid, if a problem name is not in the
-        DB, it will be added to the problem table.
-        """
-        self.start_transaction()
-        try:
-            # Use the current time as the "mtime" field
-            mtime = time.localtime()
-            try:
-                # Get the worksheetid
-                r = self.get_single(
-                    {"subject": subject, "identifier": worksheet},
-                    "worksheet", ["worksheetid"], ["subject", "identifier"])
-                worksheetid = r["worksheetid"]
-
-                # Delete any problems which might exist, if problems is
-                # supplied. If it isn't, keep the existing ones.
-                if problems is not None:
-                    query = ("DELETE FROM worksheet_problem "
-                        "WHERE worksheetid = %d;" % worksheetid)
-                    self.db.query(query)
-                # Update the row with the new details
-                if assessable is None:
-                    query = ("UPDATE worksheet "
-                        "SET mtime = %s WHERE worksheetid = %d;"
-                        % (_escape(mtime), worksheetid))
-                else:
-                    query = ("UPDATE worksheet "
-                        "SET assessable = %s, mtime = %s "
-                        "WHERE worksheetid = %d;"
-                        % (_escape(assessable), _escape(mtime), worksheetid))
-                self.db.query(query)
-            except DBException:
-                # Assume the worksheet is not in the DB
-                # If assessable is not supplied, default to False.
-                if assessable is None:
-                    assessable = False
-                # Create the worksheet row
-                query = ("INSERT INTO worksheet "
-                    "(subject, identifier, assessable, mtime) "
-                    "VALUES (%s, %s, %s, %s);"""
-                    % (_escape(subject), _escape(worksheet),
-                    _escape(assessable), _escape(mtime)))
-                self.db.query(query)
-                # Now get the worksheetid again - should succeed
-                r = self.get_single(
-                    {"subject": subject, "identifier": worksheet},
-                    "worksheet", ["worksheetid"], ["subject", "identifier"])
-                worksheetid = r["worksheetid"]
-
-            # Now insert each problem into the worksheet_problem table
-            if problems is not None:
-                for problem in problems:
-                    if isinstance(problem, tuple):
-                        prob_identifier = problem[0]
-                        try:
-                            optional = problem[1]
-                        except IndexError:
-                            optional = False
-                    else:
-                        prob_identifier = problem
-                        optional = False
-                    problemid = self.get_problem_problemid(prob_identifier)
-                    query = ("INSERT INTO worksheet_problem "
-                        "(worksheetid, problemid, optional) "
-                        "VALUES (%d, %d, %s);"
-                        % (worksheetid, problemid, _escape(optional)))
-                    self.db.query(query)
-
-            self.commit()
-        except:
-            self.rollback()
-            raise
-
-    def set_worksheet_assessable(self, subject, worksheet, assessable,
-        dry=False):
-        """
-        Sets the "assessable" field of a worksheet without updating the mtime.
-
-        IMPORTANT: This will NOT update the mtime. This is designed to allow
-        updates which did not come from the worksheet XML file. It would be
-        bad to update the mtime without consulting the XML file because then
-        it would appear the database is up to date, when it isn't.
-
-        Therefore, call this method if you are getting "assessable"
-        information from outside the worksheet XML file (eg. from the subject
-        XML file).
-
-        Unlike create_worksheet, raises a DBException if the worksheet is not
-        in the database.
-        """
-        return self.update({"subject": subject, "identifier": worksheet},
-            {"assessable": assessable}, "worksheet", ["assessable"],
-            ["subject", "identifier"], dry=dry)
-
-    def worksheet_is_assessable(self, subject, worksheet, dry=False):
-        r = self.get_single(
-            {"subject": subject, "identifier": worksheet},
-            "worksheet", ["assessable"], ["subject", "identifier"], dry=dry)
-        return _parse_boolean(r["assessable"])
-
-    def calculate_score_worksheet(self, login, subject, worksheet):
-        """
-        Calculates the score for a user on a given worksheet.
-        Returns a 4-tuple of ints, consisting of:
-        (No. mandatory exercises completed,
-         Total no. mandatory exercises,
-         No. optional exercises completed,
-         Total no. optional exercises)
-        """
-        self.start_transaction()
-        try:
-            mand_done = 0
-            mand_total = 0
-            opt_done = 0
-            opt_total = 0
-            # Get a list of problems and optionality for all problems in the
-            # worksheet
-            query = ("""SELECT problemid, optional FROM worksheet_problem
-    WHERE worksheetid = (SELECT worksheetid FROM worksheet
-                         WHERE subject = %s and identifier = %s);"""
-                    % (_escape(subject), _escape(worksheet)))
-            result = self.db.query(query)
-            # Now get the student's pass/fail for each problem in this worksheet
-            for problemid, optional in result.getresult():
-                done, _ = self.get_problem_status(login, problemid)
-                # done is a bool, whether this student has completed that
-                # problem
-                if _parse_boolean(optional):
-                    opt_total += 1
-                    if done: opt_done += 1
-                else:
-                    mand_total += 1
-                    if done: mand_done += 1
-            self.commit()
-        except:
-            self.rollback()
-            raise
-        return mand_done, mand_total, opt_done, opt_total
-
-    # ENROLMENT INFORMATION
-
-    def add_enrolment(self, login, subj_code, semester, year=None, dry=False):
-        """
-        Enrol a student in the given offering of a subject.
-        Returns True on success, False on failure (which usually means either
-        the student is already enrolled in the subject, the student was not
-        found, or no offering existed with the given details).
-        The return value can usually be ignored.
-        """
-        subj_code = str(subj_code)
-        semester = str(semester)
-        if year is None:
-            year = str(time.gmtime().tm_year)
-        else:
-            year = str(year)
-        query = """\
-INSERT INTO enrolment (loginid, offeringid)
-    VALUES (
-        (SELECT loginid FROM login WHERE login=%s),
-        (SELECT offeringid
-            FROM offering, subject, semester
-                WHERE subject.subjectid = offering.subject
-                AND semester.semesterid = offering.semesterid
-                AND subj_code=%s AND semester=%s AND year=%s)
-        );""" % (_escape(login), _escape(subj_code), _escape(semester),
-                 _escape(year))
-        if dry:
-            return query
-        try:
-            result = self.db.query(query)
-        except pg.ProgrammingError:
-            return False
-        return True
-
-    # SUBJECTS AND ENROLEMENT
-
-    def get_subjects(self, dry=False):
-        """
-        Get all subjects in IVLE.
-        Returns a list of dicts (all values strings), with the keys:
-        subj_code, subj_name, subj_short_name, url
-        """
-        return self.get_all("subject",
-            ("subjectid", "subj_code", "subj_name", "subj_short_name", "url"),
-            dry)
-
-    def get_offering_semesters(self, subjectid, dry=False):
-        """
-        Get the semester information for a subject as well as providing 
-        information about if the subject is active and which semester it is in.
-        """
-        query = """\
-SELECT offeringid, subj_name, year, semester, active
-FROM semester, offering, subject
-WHERE offering.semesterid = semester.semesterid AND
-    offering.subject = subject.subjectid AND
-    offering.subject = %d;"""%subjectid
-        if dry:
-            return query
-        results = self.db.query(query).dictresult()
-        # Parse boolean varibles
-        for result in results:
-            result['active'] = _parse_boolean(result['active'])
-        return results
-
-    def get_offering_members(self, offeringid, dry=False):
-        """
-        Gets the logins of all the people enroled in an offering
-        """
-        query = """\
-SELECT login.login AS login, login.fullname AS fullname
-FROM login, enrolment
-WHERE login.loginid = enrolment.loginid AND
-    enrolment.offeringid = %d
-    ORDER BY login.login;"""%offeringid
-        if dry:
-            return query
-        return self.db.query(query).dictresult()
-
-
-    def get_enrolment(self, login, dry=False):
-        """
-        Get all offerings (in IVLE) the student is enrolled in.
-        Returns a list of dicts (all values strings), with the keys:
-        offeringid, subj_code, subj_name, subj_short_name, year, semester, url
-        """
-        query = """\
-SELECT offering.offeringid, subj_code, subj_name, subj_short_name,
-       semester.year, semester.semester, subject.url
-FROM login, enrolment, offering, subject, semester
-WHERE enrolment.offeringid=offering.offeringid
-  AND login.loginid=enrolment.loginid
-  AND offering.subject=subject.subjectid
-  AND semester.semesterid=offering.semesterid
-  AND enrolment.active
-  AND login=%s;""" % _escape(login)
-        if dry:
-            return query
-        return self.db.query(query).dictresult()
-
-    def get_enrolment_groups(self, login, offeringid, dry=False):
-        """
-        Get all groups the user is member of in the given offering.
-        Returns a list of dicts (all values strings), with the keys:
-        name, nick
-        """
-        query = """\
-SELECT project_group.groupnm as name, project_group.nick as nick
-FROM project_set, project_group, group_member, login
-WHERE login.login=%s
-  AND project_set.offeringid=%s
-  AND group_member.loginid=login.loginid
-  AND project_group.groupid=group_member.groupid
-  AND project_group.projectsetid=project_set.projectsetid
-""" % (_escape(login), _escape(offeringid))
-        if dry:
-            return query
-        return self.db.query(query).dictresult()
-
-    def get_subjects_status(self, login, dry=False):
-        """
-        Get all subjects in IVLE, split into lists of enrolled and unenrolled
-        subjects.
-        Returns a tuple of lists (enrolled, unenrolled) of dicts
-        (all values strings) with the keys:
-        subj_code, subj_name, subj_short_name, url
-        """
-        enrolments = self.get_enrolment(login)
-        all_subjects = self.get_subjects()
-
-        enrolled_set = set(x['subj_code'] for x in enrolments)
- 
-        enrolled_subjects = [x for x in all_subjects
-                             if x['subj_code'] in enrolled_set]
-        unenrolled_subjects = [x for x in all_subjects
-                               if x['subj_code'] not in enrolled_set]
-        enrolled_subjects.sort(key=lambda x: x['subj_code'])
-        unenrolled_subjects.sort(key=lambda x: x['subj_code'])
-        return (enrolled_subjects, unenrolled_subjects)
-
-
-    # PROJECT GROUPS
-    def get_groups_by_user(self, login, offeringid=None, dry=False):
-        """
-        Get all project groups the student is in, corresponding to a
-        particular subject offering (or all offerings, if omitted).
-        Returns a list of tuples:
-        (int groupid, str groupnm, str group_nick, bool is_member).
-        (Note: If is_member is false, it means they have just been invited to
-        this group, not a member).
-        """
-        if offeringid is None:
-            and_offering = ""
-        else:
-            and_projectset_table = ", project_set"
-            and_offering = """
-AND project_group.projectsetid = project_set.projectsetid
-AND project_set.offeringid = %s""" % _escape(offeringid)
-        # Union both the groups this user is a member of, and the groups this
-        # user is invited to.
-        query = """\
-    SELECT project_group.groupid, groupnm, project_group.nick, True
-    FROM project_group, group_member, login %(and_projectset_table)s
-    WHERE project_group.groupid = group_member.groupid
-      AND group_member.loginid = login.loginid
-      AND login = %(login)s
-      %(and_offering)s
-UNION
-    SELECT project_group.groupid, groupnm, project_group.nick, False
-    FROM project_group, group_invitation, login %(and_projectset_table)s
-    WHERE project_group.groupid = group_invitation.groupid
-      AND group_invitation.loginid = login.loginid
-      AND login = %(login)s
-      %(and_offering)s
-;""" % {"login": _escape(login), "and_offering": and_offering,
-        "and_projectset_table": and_projectset_table}
-        if dry:
-            return query
-        # Convert 't' -> True, 'f' -> False
-        return [(groupid, groupnm, nick, ismember == 't')
-                for groupid, groupnm, nick, ismember
-                in self.db.query(query).getresult()]
-
-    def get_offering_info(self, projectsetid, dry=False):
-        """Takes information from projectset and returns useful information 
-        about the subject and semester. Returns as a dictionary.
-        """
-        query = """\
-SELECT subjectid, subj_code, subj_name, subj_short_name, url, year, semester, 
-active
-FROM subject, offering, semester, project_set
-WHERE offering.subject = subject.subjectid AND
-    offering.semesterid = semester.semesterid AND
-    project_set.offeringid = offering.offeringid AND
-    project_set.projectsetid = %d;"""%projectsetid
-        if dry:
-            return query
-        return self.db.query(query).dictresult()[0]
-
-    def get_projectgroup_members(self, groupid, dry=False):
-        """Returns the logins of all students in a project group
-        """
-        query = """\
-SELECT login.login as login, login.fullname as fullname
-FROM login, group_member
-WHERE login.loginid = group_member.loginid AND
-    group_member.groupid = %d
-ORDER BY login.login;"""%groupid
-        if dry:
-            return query
-        return self.db.query(query).dictresult()
-
-    def get_projectsets_by_offering(self, offeringid, dry=False):
-        """Returns all the projectsets in a particular offering"""
-        query = """\
-SELECT projectsetid, max_students_per_group
-FROM project_set
-WHERE project_set.offeringid = %d;"""%offeringid
-        if dry:
-            return query
-        return self.db.query(query).dictresult()
-
-    def get_groups_by_projectset(self, projectsetid, dry=False):
-        """Returns all the groups that are in a particular projectset"""
-        query = """\
-SELECT groupid, groupnm, nick, createdby, epoch
-FROM project_group
-WHERE project_group.projectsetid = %d;"""%projectsetid
-        if dry:
-            return query
-        return self.db.query(query).dictresult()
-
-    def close(self):
-        """Close the DB connection. Do not call any other functions after
-        this. (The behaviour of doing so is undefined).
-        """
-        self.db.close()
-        self.open = False

=== modified file 'ivle/dispatch/__init__.py'
--- ivle/dispatch/__init__.py	2009-01-13 01:36:15 +0000
+++ ivle/dispatch/__init__.py	2009-02-03 23:30:53 +0000
@@ -1,5 +1,5 @@
-# IVLE
-# Copyright (C) 2007-2008 The University of Melbourne
+# IVLE - Informatics Virtual Learning Environment
+# Copyright (C) 2007-2009 The University of Melbourne
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -15,16 +15,16 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-# Module: dispatch
-# Author: Matt Giuca
-# Date: 11/12/2007
-
-# This is a mod_python handler program. The correct way to call it is to have
-# Apache send all requests to be handled by the module 'dispatch'.
-
-# Top-level handler. Handles all requests to all pages in IVLE.
-# Handles authentication (not authorization).
-# Then passes the request along to the appropriate ivle app.
+# Author: Matt Giuca, Will Grant
+
+"""
+This is a mod_python handler program. The correct way to call it is to have
+Apache send all requests to be handled by the module 'dispatch'.
+
+Top-level handler. Handles all requests to all pages in IVLE.
+Handles authentication (not authorization).
+Then passes the request along to the appropriate ivle app.
+"""
 
 import sys
 import os
@@ -37,19 +37,39 @@
 import time
 
 import mod_python
-from mod_python import apache, Cookie
+import routes
 
 from ivle import util
 import ivle.conf
 import ivle.conf.apps
 import apps
-import login
 import html
-from request import Request
 import plugins.console # XXX: Relies on www/ being in the Python path.
 
-# List of cookies that IVLE uses (to be removed at logout)
-ivle_cookies = ["ivleforumcookie", "clipboard"]
+# XXX List of plugins, which will eventually be read in from conf
+# Elements are (module, classname) pairs.
+plugins_HACK = [
+    ('ivle.webapp.admin.user', 'Plugin'),
+]
+
+def get_routes_mapper():
+    """
+    Build a Mapper object for doing URL matching using 'routes', based on the
+    plugins config.
+    """
+    m = routes.Mapper()
+    for plugin_path, classname in plugins_HACK:
+        # Load the plugin module from somewhere in the Python path
+        # (Note that plugin_path is a fully-qualified Python module name).
+        plugin = getattr(__import__(plugin_path, fromlist=[classname]),
+            classname)
+        # Establish a URL pattern for each element of plugin.urls
+        for url in plugin.urls:
+            routex = url[0]
+            view_class = url[1]
+            kwargs_dict = url[2] if len(url) >= 3 else {}
+            m.connect(routex, view=view_class, **kwargs_dict)
+    return m
 
 def handler(req):
     """Handles a request which may be to anywhere in the site except media.
@@ -57,6 +77,7 @@
 
     req: An Apache request object.
     """
+    from ivle.dispatch.request import Request
     # Make the request object into an IVLE request which can be passed to apps
     apachereq = req
     try:
@@ -66,7 +87,7 @@
         # yet.
         handle_unknown_exception(apachereq, *sys.exc_info())
         # Tell Apache not to generate its own errors as well
-        return apache.OK
+        return mod_python.apache.OK
 
     # Run the main handler, and catch all exceptions
     try:
@@ -78,7 +99,7 @@
     except Exception:
         handle_unknown_exception(req, *sys.exc_info())
         # Tell Apache not to generate its own errors as well
-        return apache.OK
+        return mod_python.apache.OK
 
 def handler_(req, apachereq):
     """
@@ -86,19 +107,40 @@
     just used to catch exceptions.
     Takes both an IVLE request and an Apache req.
     """
+    from ivle.dispatch import login
+    from ivle.dispatch.request import Request
+
     # Hack? Try and get the user login early just in case we throw an error
     # (most likely 404) to stop us seeing not logged in even when we are.
     if not req.publicmode:
         req.user = login.get_user_details(req)
 
+    ### BEGIN New plugins framework ###
+    # XXX This should be done ONCE per Python process, not per request.
+    # (Wait till WSGI)
+    # XXX No authentication is done here
+    mapper = get_routes_mapper()
+    matchdict = mapper.match(req.uri)
+    if matchdict is not None:
+        viewcls = matchdict['view']
+        # Get the remaining arguments, less 'view', 'action' and 'controller'
+        # (The latter two seem to be built-in, and we don't want them).
+        kwargs = matchdict.copy()
+        del kwargs['view']
+        del kwargs['action']
+        del kwargs['controller']
+        # Instantiate the view, which should be a BaseView class
+        view = viewcls(req, **kwargs)
+        # Render the output
+        view.render(req)
+        req.store.commit()
+        return req.OK
+    ### END New plugins framework ###
+
     # Check req.app to see if it is valid. 404 if not.
     if req.app is not None and req.app not in ivle.conf.apps.app_url:
-        # Maybe it is a special app!
-        if req.app == 'logout':
-            logout(req)
-        else:
-            req.throw_error(Request.HTTP_NOT_FOUND,
-                "There is no application called %s." % repr(req.app))
+        req.throw_error(Request.HTTP_NOT_FOUND,
+            "There is no application called %s." % repr(req.app))
 
     # Special handling for public mode - only allow the public app, call it
     # and get out.
@@ -170,19 +212,6 @@
     # Use req.throw_error to do that.
     return req.OK
 
-def logout(req):
-    """Log out the current user (if any) by destroying the session state.
-    Then redirect to the top-level IVLE page."""
-    session = req.get_session()
-    session.invalidate()
-    session.delete()
-    # Invalidates all IVLE cookies
-    all_cookies = Cookie.get_cookies(req)
-    for cookie in all_cookies:
-        if cookie in ivle_cookies:
-            req.add_cookie(Cookie.Cookie(cookie,'',expires=1,path='/'))
-    req.throw_redirect(util.make_path('')) 
-
 def handle_unknown_exception(req, exc_type, exc_value, exc_traceback):
     """
     Given an exception that has just been thrown from IVLE, print its details
@@ -200,10 +229,10 @@
     # For some reason, some versions of mod_python have "_server" instead of
     # "main_server". So we check for both.
     try:
-        admin_email = apache.main_server.server_admin
+        admin_email = mod_python.apache.main_server.server_admin
     except AttributeError:
         try:
-            admin_email = apache._server.server_admin
+            admin_email = mod_python.apache._server.server_admin
         except AttributeError:
             admin_email = ""
     try:
@@ -211,11 +240,19 @@
         req.status = httpcode
     except AttributeError:
         httpcode = None
-        req.status = apache.HTTP_INTERNAL_SERVER_ERROR
+        req.status = mod_python.apache.HTTP_INTERNAL_SERVER_ERROR
+    try:
+        publicmode = req.publicmode
+    except AttributeError:
+        publicmode = True
     try:
         login = req.user.login
     except AttributeError:
         login = None
+    try:
+        role = req.user.role
+    except AttributeError:
+        role = None
 
     # Log File
     try:
@@ -311,31 +348,37 @@
 
         # Logging
         logging.error('%s\n%s'%(str(msg), tb))
-
+        # Error messages are only displayed is the user is NOT a student,
+        # or if there has been a problem logging the error message
+        show_errors = (not publicmode) and ((login and \
+                            str(role) != "student") or logfail)
         req.write("""<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"                 
 	"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">                                      
 <html xmlns="http://www.w3.org/1999/xhtml">
 <head><title>IVLE Internal Server Error</title></head>
 <body>
 <h1>IVLE Internal Server Error""")
-        if (codename is not None
-            and httpcode != apache.HTTP_INTERNAL_SERVER_ERROR):
-            req.write(": %s" % cgi.escape(codename))
+        if (show_errors):
+            if (codename is not None
+                        and httpcode != mod_python.apache.HTTP_INTERNAL_SERVER_ERROR):
+                req.write(": %s" % cgi.escape(codename))
+        
         req.write("""</h1>
 <p>An error has occured which is the fault of the IVLE developers or
-administration.</p>
+administration. The developers have been notified.</p>
 """)
-        if msg is not None:
-            req.write("<p>%s</p>\n" % cgi.escape(msg))
-        if httpcode is not None:
-            req.write("<p>(HTTP error code %d)</p>\n" % httpcode)
-        req.write("""
-<p>Please report this to <a href="mailto:%s">%s</a> (the system
-administrator). Include the following information:</p>
-""" % (cgi.escape(admin_email), cgi.escape(admin_email)))
+        if (show_errors):
+            if msg is not None:
+                req.write("<p>%s</p>\n" % cgi.escape(msg))
+            if httpcode is not None:
+                req.write("<p>(HTTP error code %d)</p>\n" % httpcode)
+            req.write("""
+    <p>Please report this to <a href="mailto:%s">%s</a> (the system
+    administrator). Include the following information:</p>
+    """ % (cgi.escape(admin_email), cgi.escape(admin_email)))
 
-        req.write("<pre>\n%s\n</pre>\n"%cgi.escape(tb))
-        if logfail:
-            req.write("<p>Warning: Could not open Error Log: '%s'</p>\n"
-                %cgi.escape(logfile))
+            req.write("<pre>\n%s\n</pre>\n"%cgi.escape(tb))
+            if logfail:
+                req.write("<p>Warning: Could not open Error Log: '%s'</p>\n"
+                    %cgi.escape(logfile))
         req.write("</body></html>")

=== modified file 'ivle/dispatch/html.py'
--- ivle/dispatch/html.py	2009-01-13 01:36:15 +0000
+++ ivle/dispatch/html.py	2009-01-14 10:10:12 +0000
@@ -54,16 +54,16 @@
     # Write inline JavaScript which gives the client code access to certain
     # server-side variables.
     if req.user:
-        username = repr(req.user.login)
+        username = '"%s"' % req.user.login
     else:
         username = "null"
     if req.write_javascript_settings:
         req.write("""  <script type="text/javascript">
-    root_dir = %s;
-    public_host = %s;
+    root_dir = "%s";
+    public_host = "%s";
     username = %s;
   </script>
-""" % (repr(ivle.conf.root_dir), repr(ivle.conf.public_host), username))
+""" % (ivle.conf.root_dir, ivle.conf.public_host, username))
     iconurl = get_icon_url(req.app, small=True)
     if iconurl:
         req.write("""  <link rel="shortcut icon" href="%s" />

=== modified file 'ivle/dispatch/login.py'
--- ivle/dispatch/login.py	2009-01-13 01:36:15 +0000
+++ ivle/dispatch/login.py	2009-01-15 01:18:00 +0000
@@ -21,13 +21,14 @@
 
 # Provides services for checking logins and presenting the login page.
 import os
-import time
+import datetime
 
 from mod_python import Session
 
 import ivle.conf
-from ivle import (util, db, caps, forumutil)
-from ivle.auth import authenticate, autherror
+from ivle import (util, caps, forumutil)
+from ivle.auth import authenticate, AuthError
+import ivle.database
 
 def login(req):
     """Determines whether the user is logged in or not (looking at sessions),
@@ -46,19 +47,19 @@
     """
     # Get the user details from the session, if already logged in
     # (None means not logged in yet)
-    login_details = get_user_details(req)
+    user = get_user_details(req)
 
     # Check the session to see if someone is logged in. If so, go with it.
     # No security is required here. You must have already been authenticated
     # in order to get a 'login_name' variable in the session.
-    if login_details is not None and login_details.state == "enabled":
+    if user is not None and user.state == "enabled":
         # Only allow users to authenticate if their account is ENABLED
-        return login_details
+        return user
 
     badlogin = None
 
     # Check if there is any postdata containing login information
-    if login_details is None and req.method == 'POST':
+    if user is None and req.method == 'POST':
         fields = req.get_fieldstorage()
         username = fields.getfirst('user')
         password = fields.getfirst('pass')
@@ -70,27 +71,25 @@
                 badlogin = "No password supplied."
             else:
                 try:
-                    login_details = \
-                        authenticate.authenticate(username.value, password.value)
-                # NOTE: Can't catch AuthError, since each module throws a
-                # different identity of AuthError.
-                except Exception, msg:
+                    user = authenticate.authenticate(req.store,
+                                username.value, password.value)
+                except AuthError, msg:
                     badlogin = msg
-                if login_details is None:
+                if user is None:
                     # Must have got an error. Do not authenticate.
                     pass
-                elif login_details.pass_expired():
+                elif user.password_expired:
                     badlogin = "Your password has expired."
-                elif login_details.acct_expired():
+                elif user.account_expired:
                     badlogin = "Your account has expired."
                 else:
                     # Success - Set the session and redirect to avoid POSTDATA
                     session = req.get_session()
-                    session['user'] = login_details
+                    session['login'] = user.login
                     session.save()
-                    db.DB().update_user(username.value,
-                                        last_login = time.localtime())
-                    req.add_cookie(forumutil.make_forum_cookie(login_details))
+                    user.last_login = datetime.datetime.now()
+                    req.store.commit()
+                    req.add_cookie(forumutil.make_forum_cookie(user))
                     req.throw_redirect(req.uri)
 
     # Present the HTML login page
@@ -99,30 +98,31 @@
     req.write_html_head_foot = True
 
     # User is not logged in or their account is not enabled.
-    if login_details is not None:
+    if user is not None:
         # Only possible if no errors occured thus far
-        if login_details.state == "no_agreement":
+        if user.state == "no_agreement":
             # User has authenticated but has not accepted the TOS.
             # Present them with the TOS page.
             # First set their username for display at the top, but make sure
             # the apps tabs are not displayed
-            req.user = login_details
+            req.user = user
             # IMPORTANT NOTE FOR HACKERS: You can't simply disable this check
             # if you are not planning to display a TOS page - the TOS
             # acceptance process actually calls usermgt to create the user
             # jails and related stuff.
-            present_tos(req, login_details.fullname)
+            present_tos(req, user.fullname)
             return None
-        elif login_details.state == "disabled":
+        elif user.state == "disabled":
             # User has authenticated but their account is disabled
             badlogin = "Your account has been disabled."
-        elif login_details.state == "pending":
+        elif user.state == "pending":
             # FIXME: this isn't quite the right answer, but it
             # should be more robust in the short term.
             session = req.get_session()
             session.invalidate()
             session.delete()
-            db.DB().update_user(login_details.login, state='no_agreement')
+            user.state = u'no_agreement'
+            req.store.commit()
             req.throw_redirect(req.uri)
 
     # Write the HTML for the login page
@@ -159,13 +159,14 @@
     session = req.get_session()
 
     # Check the session to see if someone is logged in. If so, go with it.
-    # No security is required here. You must have already been authenticated
-    # in order to get a 'login_name' variable in the session.
     try:
-        return session['user']
+        login = session['login']
     except KeyError:
         return None
 
+    # Get the full User object from the db associated with this login
+    return ivle.database.User.get_by_login(req.store, login)
+
 def present_tos(req, fullname):
     """Present the Terms of Service screen to the user (who has just logged in
     for the first time and needs to accept these before being admitted into

=== modified file 'ivle/dispatch/request.py'
--- ivle/dispatch/request.py	2009-01-14 06:52:36 +0000
+++ ivle/dispatch/request.py	2009-01-19 08:14:25 +0000
@@ -267,6 +267,23 @@
             # This includes binary strings.
             self.apache_req.write(string, flush)
 
+    def logout(self):
+        """Log out the current user by destroying the session state.
+        Then redirect to the top-level IVLE page."""
+        # List of cookies that IVLE uses (to be removed at logout)
+        ivle_cookies = ["ivleforumcookie", "clipboard"]
+        
+        if hasattr(self, 'session'):
+            self.session.invalidate()
+            self.session.delete()
+            # Invalidates all IVLE cookies
+            all_cookies = Cookie.get_cookies(self)
+            for cookie in all_cookies:
+                if cookie in ivle_cookies:
+                    self.add_cookie(Cookie.Cookie(cookie,'',expires=1,path='/'))
+        self.throw_redirect(ivle.util.make_path('')) 
+
+
     def flush(self):
         """Flushes the output buffer."""
         self.apache_req.flush()
@@ -303,7 +320,9 @@
         httpcode: An HTTP response status code. Pass a constant from the
         Request class.
         """
-        mod_python.util.redirect(self.apache_req, location)
+        # Note: location may be a unicode, but it MUST only have ASCII
+        # characters (non-ascii characters should be URL-encoded).
+        mod_python.util.redirect(self.apache_req, location.encode("ascii"))
 
     def add_cookie(self, cookie, value=None, **attributes):
         """Inserts a cookie into this request object's headers."""

=== modified file 'ivle/fileservice_lib/__init__.py'
--- ivle/fileservice_lib/__init__.py	2009-01-13 01:36:15 +0000
+++ ivle/fileservice_lib/__init__.py	2009-01-22 02:14:14 +0000
@@ -122,8 +122,7 @@
         try:
             action.handle_action(req, act, fields)
         except action.ActionError, message:
-            req.headers_out['X-IVLE-Action-Error'] = \
-                urllib.quote(str(message))
+            req.headers_out['X-IVLE-Action-Error'] = urllib.quote(str(message))
 
     return_type = fields.getfirst('return')
     listing.handle_return(req, return_type == "contents")

=== modified file 'ivle/fileservice_lib/action.py'
--- ivle/fileservice_lib/action.py	2009-01-13 01:36:15 +0000
+++ ivle/fileservice_lib/action.py	2009-01-22 04:02:36 +0000
@@ -135,8 +135,10 @@
 import pysvn
 
 from ivle import (util, studpath, zip)
+from ivle.fileservice_lib.exceptions import WillNotOverwrite
 import ivle.conf
 
+
 def get_login(_realm, existing_login, _may_save):
     """Callback function used by pysvn for authentication.
     realm, existing_login, _may_save: The 3 arguments passed by pysvn to
@@ -149,7 +151,8 @@
     # If we're being asked again, then it means the credentials failed for
     # some reason and we should just fail. (This is not desirable, but it's
     # better than being asked an infinite number of times).
-    return (existing_login != "", ivle.conf.login, ivle.conf.svn_pass, True)
+    return (existing_login != "", str(ivle.conf.login),
+                                  str(ivle.conf.svn_pass), True)
 
 # Make a Subversion client object
 svnclient = pysvn.Client()
@@ -263,6 +266,35 @@
     except shutil.Error:
         raise ActionError("Could not move the file specified")
 
+def svn_movefile(req, frompath, topath, copy=False):
+    """Performs an svn move, resolving filenames, checking for any errors,
+    and throwing ActionErrors if necessary. Can also be used to do a copy
+    operation instead.
+
+    frompath and topath are straight paths from the client. Will be checked.
+    """
+    if frompath is None or topath is None:
+        raise ActionError("Required field missing")
+    frompath = actionpath_to_local(req, frompath)
+    topath = actionpath_to_local(req, topath)
+    if not os.path.exists(frompath):
+        raise ActionError("The source file does not exist")
+    if os.path.exists(topath):
+        if frompath == topath:
+            raise ActionError("Source and destination are the same")
+        raise ActionError("A file already exists with that name")
+
+    try:
+        if copy:
+            svnclient.copy(frompath, topath)
+        else:
+            svnclient.move(frompath, topath)
+    except OSError:
+        raise ActionError("Could not move the file specified")
+    except pysvn.ClientError:
+        raise ActionError("Could not move the file specified")  
+
+
 ### ACTIONS ###
 
 def action_delete(req, fields):
@@ -370,7 +402,6 @@
 
     Reads fields: 'path', 'data' (file upload, multiple), 'unpack'
     """
-
     # Important: Data is "None" if the file submitted is empty.
     path = fields.getfirst('path')
     data = fields['data']
@@ -389,6 +420,13 @@
     for datum in data:
         # Each of the uploaded files
         filepath = os.path.join(path, datum.filename)
+        (_, _, filepath_local) = studpath.url_to_jailpaths(filepath)
+        if os.path.isdir(filepath_local):
+            raise ActionError("A directory already exists "
+                    + "with that name")
+        else:
+            if os.path.exists(filepath_local):
+                raise ActionError("A file already exists with that name")
         filedata = datum.file
 
         if unpack and datum.filename.lower().endswith(".zip"):
@@ -403,6 +441,8 @@
                 zip.unzip(abspath, filedata)
             except (OSError, IOError):
                 goterror = True
+            except WillNotOverwrite, e:
+                raise ActionError("File '" + e.filename + "' already exists.")
         else:
             # Not a zip file
             (_, _, filepath_local) = studpath.url_to_jailpaths(filepath)
@@ -444,12 +484,7 @@
     files = fields.getlist('file')
     if dst is None or src is None or mode is None:
         raise ActionError("Required field missing")
-    if mode == "copy":
-        copy = True
-    elif mode == "move":
-        copy = False
-    else:
-        raise ActionError("Invalid mode (must be 'copy' or 'move')")
+
     dst_local = actionpath_to_local(req, dst)
     if not os.path.isdir(dst_local):
         raise ActionError("dst is not a directory")
@@ -464,7 +499,16 @@
         # The destination is found by taking just the basename of the file
         topath = os.path.join(dst, os.path.basename(file))
         try:
-            movefile(req, frompath, topath, copy)
+            if mode == "copy":
+                movefile(req, frompath, topath, True)
+            elif mode == "move":
+                movefile(req, frompath, topath, False)
+            elif mode == "svncopy":
+                svn_movefile(req, frompath, topath, True)
+            elif mode == "svnmove":
+                svn_movefile(req, frompath, topath, False)
+            else:
+                raise ActionError("Invalid mode (must be '(svn)copy' or '(svn)move')")
         except ActionError, message:
             # Store the error for later; we want to copy as many as possible
             if errormsg is None:
@@ -709,6 +753,7 @@
             raise util.IVLEError(404, 'The specified repository path does not exist')
         else:
             raise ActionError(str(e[0]))
+            
 
 # Table of all action functions #
 # Each function has the interface f(req, fields).

=== added file 'ivle/fileservice_lib/exceptions.py'
--- ivle/fileservice_lib/exceptions.py	1970-01-01 00:00:00 +0000
+++ ivle/fileservice_lib/exceptions.py	2009-01-22 02:14:14 +0000
@@ -0,0 +1,27 @@
+# IVLE - Informatics Virtual Learning Environment
+# Copyright (C) 2007-2008 The University of Melbourne
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+
+# Module: Exceptions
+# Author: Nicholas Chadwick
+# Date: 22/01/2009
+
+class WillNotOverwrite(Exception):
+    "This exception is to be thrown whenever an action fails due to \
+        overwrite"
+    def __init__(self, filename):
+        super(Exception, self).__init__()
+        self.filename = filename

=== modified file 'ivle/fileservice_lib/listing.py'
--- ivle/fileservice_lib/listing.py	2009-01-13 01:36:15 +0000
+++ ivle/fileservice_lib/listing.py	2009-01-22 02:14:14 +0000
@@ -146,6 +146,7 @@
     # Currently goes to 403 Forbidden.
     urlpath = urlparse.urlparse(path)
     path = urlpath[2]
+    json = None
     if path is None:
         req.status = req.HTTP_FORBIDDEN
         req.headers_out['X-IVLE-Return-Error'] = 'Forbidden'
@@ -158,7 +159,11 @@
         # It's a directory. Return the directory listing.
         req.content_type = mime_dirlisting
         req.headers_out['X-IVLE-Return'] = 'Dir'
-        req.write(cjson.encode(get_dirlisting(req, svnclient, path)))
+        # TODO: Fix this dirty, dirty hack
+        newjson = get_dirlisting(req, svnclient, path)
+        if ("X-IVLE-Action-Error" in req.headers_out):
+            newjson["Error"] = req.headers_out["X-IVLE-Action-Error"]
+        req.write(cjson.encode(newjson))
     elif return_contents:
         # It's a file. Return the file contents.
         # First get the mime type of this file
@@ -237,6 +242,11 @@
         try:
             for filename in os.listdir(path):
                 listing[filename.decode('utf-8')] = file_to_fileinfo(path, filename)[1]
+                try:
+                    svnclient.status(os.path.join(path, filename), recurse = False)
+                    listing[filename.decode('utf-8')]['svnstatus'] = 'normal'
+                except:
+                    pass
         except OSError:
             # Non-directories will error - that's OK, we just want the "."
             pass

=== modified file 'ivle/interpret.py'
--- ivle/interpret.py	2009-01-13 01:36:15 +0000
+++ ivle/interpret.py	2009-01-20 03:20:57 +0000
@@ -22,7 +22,6 @@
 # Runs a student script in a safe execution environment.
 
 from ivle import studpath
-from ivle import db
 from ivle.util import IVLEError, IVLEJailError
 import ivle.conf
 
@@ -39,24 +38,6 @@
 
 CGI_BLOCK_SIZE = 65535
 
-uids = {}
-
-def get_uid(login):
-    """Get the unix uid corresponding to the given login name.
-       If it is not in the dictionary of uids, then consult the
-       database and retrieve an update of the user table."""
-    global uids
-    if login in uids:
-        return uids[login]
-
-    conn = db.DB()
-    res = conn.get_all('login', ['login', 'unixid'])
-    def repack(flds):
-        return (flds['login'], flds['unixid'])
-    uids = dict(map(repack,res))
-
-    return uids[login]
-
 def interpret_file(req, owner, jail_dir, filename, interpreter, gentle=True):
     """Serves a file by interpreting it using one of IVLE's builtin
     interpreters. All interpreters are intended to run in the user's jail. The
@@ -64,7 +45,7 @@
     to the individual interpreters to create the jail.
 
     req: An IVLE request object.
-    owner: Username of the user who owns the file being served.
+    owner: The user who owns the file being served.
     jail_dir: Absolute path to the user's jail.
     filename: Absolute filename within the user's jail.
     interpreter: A function object to call.
@@ -79,11 +60,9 @@
         filename_abs = os.path.join(os.sep, filename)
         filename_rel = filename
 
-    # Get the UID of the owner of the file
     # (Note: files are executed by their owners, not the logged in user.
     # This ensures users are responsible for their own programs and also
     # allows them to be executed by the public).
-    uid = get_uid(owner)
 
     # Split up req.path again, this time with respect to the jail
     (working_dir, _) = os.path.split(filename_abs)
@@ -94,7 +73,7 @@
     # (Note that paths "relative" to the jail actually begin with a '/' as
     # they are absolute in the jailspace)
 
-    return interpreter(uid, jail_dir, working_dir, filename_abs, req,
+    return interpreter(owner.unixid, jail_dir, working_dir, filename_abs, req,
                        gentle)
 
 class CGIFlags:

=== modified file 'ivle/makeuser.py'
--- ivle/makeuser.py	2009-01-13 01:36:15 +0000
+++ ivle/makeuser.py	2009-01-20 01:02:29 +0000
@@ -45,9 +45,10 @@
 import filecmp
 import logging
 import ivle.conf
-import ivle.db
 import ivle.pulldown_subj
 
+from ivle.database import ProjectGroup
+
 def chown_to_webserver(filename):
     """
     Chowns a file so the web server user owns it.
@@ -73,11 +74,10 @@
 
     chown_to_webserver(path)
 
-def rebuild_svn_config():
+def rebuild_svn_config(store):
     """Build the complete SVN configuration file.
     """
-    conn = ivle.db.DB()
-    users = conn.get_users()
+    users = store.find(ivle.database.User)
     groups = {}
     for u in users:
         role = str(u.role)
@@ -103,35 +103,30 @@
     os.rename(ivle.conf.svn_conf + ".new", ivle.conf.svn_conf)
     chown_to_webserver(ivle.conf.svn_conf)
 
-def rebuild_svn_group_config():
+def rebuild_svn_group_config(store):
     """Build the complete SVN configuration file for groups
     """
-    conn = ivle.db.DB()
-    groups = conn.get_all('project_group',
-        ['groupid', 'groupnm', 'projectsetid'])
     f = open(ivle.conf.svn_group_conf + ".new", "w")
     f.write("# IVLE SVN Group Repositories Configuration\n")
     f.write("# Auto-generated on %s\n" % time.asctime())
     f.write("\n")
-    for g in groups:
-        projectsetid = g['projectsetid']
-        offeringinfo = conn.get_offering_info(projectsetid)
-        subj_short_name = offeringinfo['subj_short_name']
-        year = offeringinfo['year']
-        semester = offeringinfo['semester']
-        reponame = "_".join([subj_short_name, year, semester, g['groupnm']])
+    for group in store.find(ProjectGroup):
+        offering = group.project_set.offering
+        reponame = "_".join([offering.subject.short_name,
+                             offering.semester.year,
+                             offering.semester.semester,
+                             group.name])
         f.write("[%s:/]\n"%reponame)
-        users = conn.get_projectgroup_members(g['groupid'])
-        for u in users:
-            f.write("%s = rw\n"%u['login'])
+        for user in group.members:
+            f.write("%s = rw\n" % user.login)
         f.write("\n")
     f.close()
     os.rename(ivle.conf.svn_group_conf + ".new", ivle.conf.svn_group_conf)
     chown_to_webserver(ivle.conf.svn_group_conf)
 
-def make_svn_auth(login, throw_on_error=True):
+def make_svn_auth(store, login, throw_on_error=True):
     """Setup svn authentication for the given user.
-       FIXME: create local.auth entry
+       Uses the given DB store object. Does not commit to the db.
     """
     passwd = md5.new(uuid.uuid4().bytes).digest().encode('hex')
     if os.path.exists(ivle.conf.svn_auth_ivle):
@@ -139,7 +134,8 @@
     else:
         create = "c"
 
-    ivle.db.DB().update_user(login, svn_pass=passwd)
+    user = ivle.database.User.get_by_login(store, login)
+    user.svn_pass = unicode(passwd)
 
     res = os.system("htpasswd -%smb %s %s %s" % (create,
                                               ivle.conf.svn_auth_ivle,
@@ -182,7 +178,7 @@
     return (to_add, to_remove)
 
 
-def make_jail(username, uid, force=True, svn_pass=None):
+def make_jail(user, force=True):
     """Creates a new user's jail space, in the jail directory as configured in
     conf.py.
 
@@ -193,17 +189,8 @@
 
     Chowns the user's directory within the jail to the given UID.
 
-    Note: This takes separate username and uid arguments. The UID need not
-    *necessarily* correspond to a Unix username at all, if all you are
-    planning to do is setuid to it. This allows the caller the freedom of
-    deciding the binding between username and uid, if any.
-
     force: If false, exception if jail already exists for this user.
     If true (default), overwrites it, but preserves home directory.
-
-    svn_pass: If provided this will be a string, the randomly-generated
-    Subversion password for this user (if you happen to already have it).
-    If not provided, it will be read from the database.
     """
     # MUST run as root or some of this may fail
     if os.getuid() != 0:
@@ -216,9 +203,9 @@
     elif not os.path.isdir(tempdir):
         os.unlink(tempdir)
         os.mkdir(tempdir)
-    userdir = os.path.join(ivle.conf.jail_src_base, username)
+    userdir = os.path.join(ivle.conf.jail_src_base, user.login)
     homedir = os.path.join(userdir, 'home')
-    userhomedir = os.path.join(homedir, username)   # Return value
+    userhomedir = os.path.join(homedir, user.login)   # Return value
 
     if os.path.exists(userdir):
         if not force:
@@ -239,30 +226,26 @@
         shutil.move(homebackup, homedir)
         # Change the ownership of all the files to the right unixid
         logging.debug("chown %s's home directory files to uid %d"
-            %(username, uid))
-        os.chown(userhomedir, uid, uid)
+            %(user.login, user.unixid))
+        os.chown(userhomedir, user.unixid, user.unixid)
         for root, dirs, files in os.walk(userhomedir):
             for fsobj in dirs + files:
-                os.chown(os.path.join(root, fsobj), uid, uid)
+                os.chown(os.path.join(root, fsobj), user.unixid, user.unixid)
     else:
         # No user jail exists
         # Set up the user's home directory
         os.makedirs(userhomedir)
         # Chown (and set the GID to the same as the UID).
-        os.chown(userhomedir, uid, uid)
+        os.chown(userhomedir, user.unixid, user.unixid)
         # Chmod to rwxr-xr-x (755)
         os.chmod(userhomedir, 0755)
 
-    # There are 2 special files which need to be generated specific to this
-    # user: ${python_site_packages}/lib/conf/conf.py and /etc/passwd.
-    # "__" username "__" users are exempt (special)
-    if not (username.startswith("__") and username.endswith("__")):
-        make_conf_py(username, userdir, ivle.conf.jail_system, svn_pass)
-        make_etc_passwd(username, userdir, ivle.conf.jail_system, uid)
+    make_conf_py(user.login, userdir, ivle.conf.jail_system, user.svn_pass)
+    make_etc_passwd(user.login, userdir, ivle.conf.jail_system, user.unixid)
 
     return userhomedir
 
-def make_conf_py(username, user_jail_dir, staging_dir, svn_pass=None):
+def make_conf_py(username, user_jail_dir, staging_dir, svn_pass):
     """
     Creates (overwriting any existing file, and creating directories) a
     file ${python_site_packages}/ivle/conf/conf.py in a given user's jail.
@@ -279,12 +262,6 @@
             ivle.conf.python_site_packages[1:], "ivle/conf/conf.py")
     os.makedirs(os.path.dirname(conf_path))
 
-    # If svn_pass isn't supplied, grab it from the DB
-    if svn_pass is None:
-        dbconn = ivle.db.DB()
-        svn_pass = dbconn.get_user(username).svn_pass
-        dbconn.close()
-
     # Read the contents of the template conf file
     try:
         template_conf_file = open(template_conf_path, "r")
@@ -327,40 +304,6 @@
                       % (username, unixid, unixid, username))
     passwd_file.close()
 
-def make_user_db(throw_on_error = True, **kwargs):
-    """Creates a user's entry in the database, filling in all the fields.
-    All arguments must be keyword args. They are the fields in the table.
-    However, instead of supplying a "passhash", you must supply a
-    "password" argument, which will be hashed internally.
-    Also do not supply a state. All users are created in the "no_agreement"
-    state.
-    Also pulls the user's subjects using the configured subject pulldown
-    module, and adds enrolments to the DB.
-    Throws an exception if the user already exists.
-    """
-    dbconn = ivle.db.DB()
-    dbconn.create_user(**kwargs)
-    dbconn.close()
-
-    if kwargs['password']:
-        if os.path.exists(ivle.conf.svn_auth_local):
-            create = ""
-        else:
-            create = "c"
-        res = os.system("htpasswd -%smb %s %s %s" % (create,
-                                                     ivle.conf.svn_auth_local,
-                                                     kwargs['login'],
-                                                     kwargs['password']))
-        if res != 0 and throw_on_error:
-            raise Exception("Unable to create local-auth for %s" % kwargs['login'])
-
-    # Make sure the file is owned by the web server
-    if create == "c":
-        chown_to_webserver(ivle.conf.svn_auth_local)
-
-    # Pulldown subjects and add enrolments
-    ivle.pulldown_subj.enrol_user(kwargs['login'])
-
 def mount_jail(login):
     # This is where we'll mount to...
     destdir = os.path.join(ivle.conf.jail_base, login)

=== modified file 'ivle/pulldown_subj/__init__.py'
--- ivle/pulldown_subj/__init__.py	2009-01-13 01:36:15 +0000
+++ ivle/pulldown_subj/__init__.py	2009-01-20 00:58:53 +0000
@@ -28,10 +28,12 @@
 
 import os
 import sys
+import datetime
 
 from subjecterror import SubjectError
 import ivle.conf
-import ivle.db
+
+from ivle.database import Subject, Offering, Semester, AlreadyEnrolledError
 
 def get_subjects(login):
     """
@@ -45,7 +47,7 @@
             return result
     return []
 
-def enrol_user(login, year=None):
+def enrol_user(store, user, year=None):
     """
     Looks up the student in whatever modules are available.
     The pulldown does not tell us what year to enrol the student for, so the
@@ -61,10 +63,27 @@
     Returns the number of subjects the user was enrolled in (not including
     subjects outside the system, or subjects already enrolled).
     """
+    if year is None:
+        year = unicode(datetime.datetime.now().year)
+
     count = 0
-    db = ivle.db.DB()
-    for subject, semester in get_subjects(login):
-        count += db.add_enrolment(login, subject, semester, year)
+    for subject, semester in get_subjects(user.login):
+        offering = store.find(Offering,
+                              Subject.code == subject,
+                              Offering.subject_id == Subject.id,
+                              Semester.year == year,
+                              Semester.semester == semester,
+                              Offering.semester_id == Semester.id).one()
+
+        # We can't find a matching offering, so we don't care about it.
+        if not offering:
+            continue
+
+        try:
+            offering.enrol(user)
+            count += 1
+        except AlreadyEnrolledError:
+            pass
     return count
 
 # Allow imports to get files from this directory.

=== modified file 'ivle/pulldown_subj/dummy_subj.py'
--- ivle/pulldown_subj/dummy_subj.py	2009-01-13 01:36:15 +0000
+++ ivle/pulldown_subj/dummy_subj.py	2009-01-19 23:29:19 +0000
@@ -33,9 +33,9 @@
     if ord(login[:1]) < ord('m'):
         # Return a list of subject/semester pairs
         return [
-            ('12345', '1'),
-            ('12345', '2'),
-            ('98765', '1'),
+            (u'12345', u'1'),
+            (u'12345', u'2'),
+            (u'98765', u'1'),
         ]
     else:
         # This demonstrates that you can return None, to indicate to the

=== added directory 'ivle/rpc'
=== added file 'ivle/rpc/__init__.py'
=== added file 'ivle/rpc/decorators.py'
--- ivle/rpc/decorators.py	1970-01-01 00:00:00 +0000
+++ ivle/rpc/decorators.py	2009-01-20 06:11:35 +0000
@@ -0,0 +1,32 @@
+'''Decorators useful for actions in the IVLE AJAX API.
+
+The first argument to any method wrapped by these needs to be a request.
+'''
+
+class require_method(object):
+    '''Require that the request has been made with the specified HTTP method.
+    '''
+    def __init__(self, method):
+        self.method = method
+
+    def __call__(self, func):
+        def method_or_die(req, *args, **kwargs):
+           if req.method != self.method:
+               req.throw_error(req.HTTP_METHOD_NOT_ALLOWED,
+               "Only %s requests can be made to this action." % self.method)
+           func(req, *args, **kwargs)
+        return method_or_die
+
+class require_cap(object):
+    '''Require that the logged in user has the specified capability.'''
+    def __init__(self, cap):
+        self.cap = cap
+
+    def __call__(self, func):
+        def cap_or_die(req, *args, **kwargs):
+           if not req.user or not req.user.hasCap(self.cap):
+               req.throw_error(req.HTTP_FORBIDDEN,
+               "You do not have permission to use this action.")
+           func(req, *args, **kwargs)
+        return cap_or_die
+

=== removed file 'ivle/user.py'
--- ivle/user.py	2009-01-13 01:36:15 +0000
+++ ivle/user.py	1970-01-01 00:00:00 +0000
@@ -1,114 +0,0 @@
-# IVLE - Informatics Virtual Learning Environment
-# Copyright (C) 2007-2008 The University of Melbourne
-#
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
-
-# Module: User
-# Author: Matt Giuca
-# Date:   19/2/2008
-
-# Provides a User class which stores the login details of a particular IVLE
-# user. Objects of this class are expected to be stored in the request
-# session.
-
-import ivle.db
-import time
-import ivle.caps
-
-# Similar to ivle.db.login_fields_list but contains a different set of fields.
-# The User object does not contain all the fields.
-user_fields_required = frozenset((
-    "login", "fullname", "role", "state", "unixid"
-))
-user_fields_list = (
-    "login", "state", "unixid", "email", "nick", "fullname",
-    "role", "studentid", "acct_exp", "pass_exp", "last_login",
-    "svn_pass", "local_password",
-)
-timestamp_fields = frozenset((
-    "acct_exp", "pass_exp", "last_login",
-))
-# Fields not included: passhash, last_login
-
-class UserException(Exception):
-    pass
-
-class User(object):
-    """
-    Stores the login details of a particular IVLE user.
-    Its fields correspond to (most of) the fields of the "login" table of the
-    IVLE db.
-    All fields are always present, but some may be None.
-    """
-    __slots__ = user_fields_list
-    def __init__(self, **kwargs):
-        # XXX Will ignore unknown fields instead of erroring
-        if "rolenm" in kwargs and "role" not in kwargs:
-            kwargs['role'] = ivle.caps.Role(kwargs['rolenm'])
-        if "passhash" in kwargs and "local_password" not in kwargs:
-            kwargs['local_password'] = kwargs['passhash'] is not None
-        for r in user_fields_list:
-            if r in kwargs:
-                self.__setattr__(r, kwargs[r])
-            elif r in user_fields_required:
-                # Required argument, not specified
-                raise TypeError("User: Required field %s missing" % repr(r))
-            else:
-                # Optional arguments
-                if r == "nick":
-                    try:
-                        self.nick = kwargs['fullname']
-                    except KeyError:
-                        raise TypeError("User: Required field "
-                            "'fullname' missing")
-                else:
-                    self.__setattr__(r, None)
-        for field in timestamp_fields:
-            # Convert all timestamp fields from strings to Timestamp objects
-            if hasattr(self, field):
-                val = self.__getattribute__(field)
-                if val is not None:
-                    val = time.strptime(val, ivle.db.TIMESTAMP_FORMAT)
-                    self.__setattr__(field, val)
-
-    def __repr__(self):
-        items = ["%s=%s" % (r, repr(self.__getattribute__(r)))
-            for r in user_fields_list]
-        return "User(" + ', '.join(items) + ")"
-    def __iter__(self):
-        """Iteration yielding field:value pairs.
-        (Allows the "dict" function to work on Users)
-        """
-        for r in user_fields_list:
-            yield (r, self.__getattribute__(r))
-
-    def hasCap(self, capability):
-        """Given a capability (which is a Role object), returns True if this
-        User has that capability, False otherwise.
-        """
-        return self.role.hasCap(capability)
-
-    def pass_expired(self):
-        """Determines whether the pass_exp field indicates that
-           login should be denied.
-        """
-        fieldval = self.pass_exp
-        return fieldval is not None and time.localtime() > fieldval
-    def acct_expired(self):
-        """Determines whether the acct_exp field indicates that
-           login should be denied.
-        """
-        fieldval = self.acct_exp
-        return fieldval is not None and time.localtime() > fieldval

=== modified file 'ivle/util.py'
--- ivle/util.py	2009-01-13 01:36:15 +0000
+++ ivle/util.py	2009-01-14 10:10:12 +0000
@@ -292,3 +292,14 @@
     elif count < expect:
         # Incomplete
         return count
+
+def object_to_dict(attrnames, obj):
+    """
+    Convert an object into a dictionary. This takes a shallow copy of the
+    object.
+    attrnames: Set (or iterable) of names of attributes to be copied into the
+        dictionary. (We don't auto-lookup, because this function needs to be
+        used on magical objects).
+    """
+    return dict((k, getattr(obj, k))
+        for k in attrnames if not k.startswith('_'))

=== added directory 'ivle/webapp'
=== added file 'ivle/webapp/__init__.py'
=== added directory 'ivle/webapp/admin'
=== added file 'ivle/webapp/admin/__init__.py'
=== added file 'ivle/webapp/admin/user.py'
--- ivle/webapp/admin/user.py	1970-01-01 00:00:00 +0000
+++ ivle/webapp/admin/user.py	2009-02-03 22:46:32 +0000
@@ -0,0 +1,76 @@
+# IVLE - Informatics Virtual Learning Environment
+# Copyright (C) 2007-2009 The University of Melbourne
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+
+# Author: Matt Giuca, Will Grant
+
+from ivle.webapp.base.views import JSONRESTView
+from ivle.webapp.base.plugins import BasePlugin
+import ivle.database
+import ivle.util
+
+# List of fields returned as part of the user JSON dictionary
+# (as returned by the get_user action)
+user_fields_list = (
+    "login", "state", "unixid", "email", "nick", "fullname",
+    "rolenm", "studentid", "acct_exp", "pass_exp", "last_login",
+    "svn_pass"
+)
+
+class UserRESTView(JSONRESTView):
+    """
+    A REST interface to the user object.
+    """
+    def __init__(self, req, login):
+        super(UserRESTView, self).__init__(self, req, login)
+        self.context = ivle.database.User.get_by_login(req.store, login)
+
+    def GET(self, req):
+        # XXX Check Caps
+        user = ivle.util.object_to_dict(user_fields_list, self.context)
+        # Convert time stamps to nice strings
+        for k in 'pass_exp', 'acct_exp', 'last_login':
+            if user[k] is not None:
+                user[k] = unicode(user[k])
+
+        user['local_password'] = self.context.passhash is not None
+        return user
+
+    def PATCH(self, req, data):
+        # XXX Check Caps
+        # XXX Admins can set extra fields
+        # Note: Cannot change password here (use change_password named op)
+
+        for f in user_fields_list:
+            try:
+                field = data[f]
+                if isinstance(field, str):
+                    field = unicode(field)
+                setattr(self.context, f, field)
+            except KeyError:
+                continue
+
+class Plugin(BasePlugin):
+    """
+    The Plugin class for the user plugin.
+    """
+    # Magic attribute: urls
+    # Sequence of pairs/triples of
+    # (regex str, handler class, kwargs dict)
+    # The kwargs dict is passed to the __init__ of the view object
+    urls = [
+        ('api/users/:login', UserRESTView)
+    ]

=== added directory 'ivle/webapp/base'
=== added file 'ivle/webapp/base/__init__.py'
=== added file 'ivle/webapp/base/plugins.py'
--- ivle/webapp/base/plugins.py	1970-01-01 00:00:00 +0000
+++ ivle/webapp/base/plugins.py	2009-02-03 13:02:36 +0000
@@ -0,0 +1,24 @@
+# IVLE - Informatics Virtual Learning Environment
+# Copyright (C) 2007-2009 The University of Melbourne
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+
+# Author: Matt Giuca, Will Grant
+
+class BasePlugin(object):
+    """
+    Base class for all Plugin classes.
+    """
+    pass

=== added file 'ivle/webapp/base/views.py'
--- ivle/webapp/base/views.py	1970-01-01 00:00:00 +0000
+++ ivle/webapp/base/views.py	2009-02-03 22:49:08 +0000
@@ -0,0 +1,65 @@
+# IVLE - Informatics Virtual Learning Environment
+# Copyright (C) 2007-2009 The University of Melbourne
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+
+# Author: Matt Giuca, Will Grant
+
+import cjson
+
+class BaseView(object):
+    """
+    Abstract base class for all view objects.
+    """
+    def __init__(self, req, **kwargs):
+        pass
+    def render(self, req):
+        pass
+
+class RESTView(BaseView):
+    """
+    A view which provides a RESTful interface. The content type is
+    unspecified (see JSONRESTView for a specific content type).
+    """
+    content_type = "application/octet-stream"
+
+    def __init__(self, req, *args, **kwargs):
+        pass
+
+    def render(self, req):
+        if req.method == 'GET':
+            outstr = self.GET(req)
+        # XXX PATCH hack
+        if req.method == 'PUT':
+            outstr = self.PATCH(req, req.read())
+        req.content_type = self.content_type
+        req.write(outstr)
+
+class JSONRESTView(RESTView):
+    """
+    A special case of RESTView which deals entirely in JSON.
+    """
+    content_type = "application/json"
+
+    def render(self, req):
+        if req.method == 'GET':
+            outjson = self.GET(req)
+        # XXX PATCH hack
+        if req.method == 'PUT':
+            outjson = self.PATCH(req, cjson.decode(req.read()))
+        req.content_type = self.content_type
+        if outjson is not None:
+            req.write(cjson.encode(outjson))
+            req.write("\n")

=== added file 'ivle/worksheet.py'
--- ivle/worksheet.py	1970-01-01 00:00:00 +0000
+++ ivle/worksheet.py	2009-01-20 05:03:55 +0000
@@ -0,0 +1,192 @@
+# IVLE - Informatics Virtual Learning Environment
+# Copyright (C) 2007-2009 The University of Melbourne
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+
+# Author: Matt Giuca
+
+"""
+Worksheet Utility Functions
+
+This module provides functions for tutorial and worksheet computations.
+"""
+
+from storm.locals import And, Asc, Desc
+import ivle.database
+
+__all__ = ['get_exercise_status', 'get_exercise_stored_text',
+           'get_exercise_attempts', 'get_exercise_attempt',
+          ]
+
+def get_exercise_status(store, user, exercise):
+    """Given a storm.store, User and Exercise, returns information about
+    the user's performance on that problem.
+    Returns a tuple of:
+        - A boolean, whether they have successfully passed this exercise.
+        - An int, the number of attempts they have made up to and
+          including the first successful attempt (or the total number of
+          attempts, if not yet successful).
+    """
+    ExerciseAttempt = ivle.database.ExerciseAttempt
+    # A Storm expression denoting all active attempts by this user for this
+    # exercise.
+    is_relevant = ((ExerciseAttempt.user_id == user.id) &
+                   (ExerciseAttempt.exercise_id == exercise.id) &
+                   (ExerciseAttempt.active == True))
+
+    # Get the first successful active attempt, or None if no success yet.
+    # (For this user, for this exercise).
+    first_success = store.find(ExerciseAttempt, is_relevant,
+            ExerciseAttempt.complete == True
+        ).order_by(Asc(ExerciseAttempt.date)).first()
+
+    if first_success is not None:
+        # Get the total number of active attempts up to and including the
+        # first successful attempt.
+        # (Subsequent attempts don't count, because the user had already
+        # succeeded by then).
+        num_attempts = store.find(ExerciseAttempt, is_relevant,
+                ExerciseAttempt.date <= first_success.date).count()
+    else:
+        # User has not yet succeeded.
+        # Get the total number of active attempts.
+        num_attempts = store.find(ExerciseAttempt, is_relevant).count()
+
+    return first_success is not None, num_attempts
+
+def get_exercise_stored_text(store, user, exercise):
+    """Given a storm.store, User and Exercise, returns an
+    ivle.database.ExerciseSave object for the last saved/submitted attempt for
+    this question (note that ExerciseAttempt is a subclass of ExerciseSave).
+    Returns None if the user has not saved or made an attempt on this
+    problem.
+    If the user has both saved and submitted, it returns whichever was
+    made last.
+    """
+    ExerciseSave = ivle.database.ExerciseSave
+    ExerciseAttempt = ivle.database.ExerciseAttempt
+
+    # Get the saved text, or None
+    saved = store.find(ExerciseSave,
+                ExerciseSave.user_id == user.id,
+                ExerciseSave.exercise_id == exercise.id).one()
+
+    # Get the most recent attempt, or None
+    attempt = store.find(ExerciseAttempt,
+            ExerciseAttempt.user_id == user.id,
+            ExerciseAttempt.exercise_id == exercise.id,
+            ExerciseAttempt.active == True,
+        ).order_by(Asc(ExerciseAttempt.date)).last()
+
+    # Pick the most recent of these two
+    if saved is not None:
+        if attempt is not None:
+            return saved if saved.date > attempt.date else attempt
+        else:
+            return saved
+    else:
+        if attempt is not None:
+            return attempt
+        else:
+            return None
+
+def _get_exercise_attempts(store, user, exercise, as_of=None,
+        allow_inactive=False):
+    """Same as get_exercise_attempts, but doesn't convert Storm's iterator
+    into a list."""
+    ExerciseAttempt = ivle.database.ExerciseAttempt
+
+    # Get the most recent attempt before as_of, or None
+    return store.find(ExerciseAttempt,
+            ExerciseAttempt.user_id == user.id,
+            ExerciseAttempt.exercise_id == exercise.id,
+            True if allow_inactive else ExerciseAttempt.active == True,
+            True if as_of is None else ExerciseAttempt.date <= as_of,
+        ).order_by(Desc(ExerciseAttempt.date))
+
+def get_exercise_attempts(store, user, exercise, as_of=None,
+        allow_inactive=False):
+    """Given a storm.store, User and Exercise, returns a list of
+    ivle.database.ExerciseAttempt objects, one for each attempt made for the
+    exercise, sorted from latest to earliest.
+
+    as_of: Optional datetime.datetime object. If supplied, only returns
+        attempts made before or at this time.
+    allow_inactive: If True, will return disabled attempts.
+    """
+    return list(_get_exercise_attempts(store, user, exercise, as_of,
+        allow_inactive))
+
+def get_exercise_attempt(store, user, exercise, as_of=None,
+        allow_inactive=False):
+    """Given a storm.store, User and Exercise, returns an
+    ivle.database.ExerciseAttempt object for the last submitted attempt for
+    this question.
+    Returns None if the user has not made an attempt on this
+    problem.
+
+    as_of: Optional datetime.datetime object. If supplied, only returns
+        attempts made before or at this time.
+    allow_inactive: If True, will return disabled attempts.
+    """
+    return _get_exercise_attempts(store, user, exercise, as_of,
+        allow_inactive).first()
+
+def save_exercise(store, user, exercise, text, date):
+    """Save an exercise for a user.
+
+    Given a store, User, Exercise and text and date, save the text to the
+    database. This will create the ExerciseSave if needed.
+    """
+    saved = store.find(ivle.database.ExerciseSave,
+                ivle.database.ExerciseSave.user_id == user.id,
+                ivle.database.ExerciseSave.exercise_id == exercise.id).one()
+    if saved is None:
+        saved = ivle.database.ExerciseSave(user=user, exercise=exercise)
+        store.add(saved)
+
+    saved.date = date
+    saved.text = text
+
+def calculate_score(store, user, worksheet):
+    """
+    Given a storm.store, User, Exercise and Worksheet, calculates a score for
+    the user on the given worksheet.
+    Returns a 4-tuple of ints, consisting of:
+    (No. mandatory exercises completed,
+     Total no. mandatory exercises,
+     No. optional exercises completed,
+     Total no. optional exercises)
+    """
+    mand_done = 0
+    mand_total = 0
+    opt_done = 0
+    opt_total = 0
+
+    # Get the student's pass/fail for each exercise in this worksheet
+    for worksheet_exercise in worksheet.worksheet_exercises:
+        exercise = worksheet_exercise.exercise
+        optional = worksheet_exercise.optional
+
+        done, _ = get_exercise_status(store, user, exercise)
+        # done is a bool, whether this student has completed that problem
+        if optional:
+            opt_total += 1
+            if done: opt_done += 1
+        else:
+            mand_total += 1
+            if done: mand_done += 1
+
+    return mand_done, mand_total, opt_done, opt_total

=== modified file 'ivle/zip.py'
--- ivle/zip.py	2009-01-13 01:36:15 +0000
+++ ivle/zip.py	2009-01-22 02:14:14 +0000
@@ -27,6 +27,7 @@
 import zipfile
 
 from ivle import studpath
+from ivle.fileservice_lib.exceptions import WillNotOverwrite
 
 def make_zip(basepath, paths, file):
     """Zips up a bunch of files on the student file space and writes it as
@@ -86,27 +87,32 @@
 
     zip.close()
 
-def unzip(path, file):
+def unzip(path, file, overwrite=False):
     """Unzips a zip file (or file-like object) into a path.
     Note: All files go directly into the path. To avoid having a "zip bomb"
     situation, the zip file should have a single directory in it with all the
     files.
     The path is an absolute path in the current filesystem
-    (if this code is executed inside the jail, then it's inside the jail,
-    if it's not then it's not).
+    (if this code is executed inside the jail, then it's inside the jail).
     """
     zip = zipfile.ZipFile(file, 'r')
     # First test the zip file
     if zip.testzip() is not None:
         raise OSError("ZIP: Bad zip file")
 
+    if (not overwrite):
+        for filename in zip.namelist():
+            localpath = os.path.join(path, filename)
+            if os.path.exists(localpath):
+                raise WillNotOverwrite(filename)
+
     for filename in zip.namelist():
         localpath = os.path.join(path, filename)
         # Create directory for filename
         (file_dir, _) = os.path.split(localpath)
         if not os.path.exists(file_dir):
             os.makedirs(file_dir)
-
+        
         if filename.endswith(os.sep):
             # Is a directory make the directory
             if not os.path.exists(localpath):

=== modified file 'services/usrmgt-server'
--- services/usrmgt-server	2009-01-13 01:36:15 +0000
+++ services/usrmgt-server	2009-01-15 03:02:36 +0000
@@ -5,7 +5,7 @@
 import logging
 
 import ivle.conf
-import ivle.db
+import ivle.database
 import ivle.chat
 import ivle.makeuser
 
@@ -27,7 +27,7 @@
 #   - Rebuild svn auth file
 #   - Rebuild passwd + push to nodes.
 
-def activate_user(props):
+def activate_user(store, props):
     """Create the on-disk stuff for the given user.
        Sets the state of the user in the db from pending to enabled.
        Expected properties:
@@ -40,57 +40,50 @@
 
     login = props['login']
 
-    db = ivle.db.DB()
-
-    try:
-
-        # FIXME: check we're pending
-
-        details = db.get_user(login)
-
-        # make svn config/auth
-
-        repopath = os.path.join(ivle.conf.svn_repo_path, 'users', login)
-        logging.debug("Creating user's Subversion repository")
-        ivle.makeuser.make_svn_repo(repopath, throw_on_error=True)
-
-        rebuild_svn_config(props)
-
-        logging.debug("Adding Subversion authentication")
-        passwd = ivle.makeuser.make_svn_auth(login, throw_on_error=True)
-        logging.debug("passwd: %s" % passwd)
-
-        logging.debug("Creating jail")
-        ivle.makeuser.make_jail(login, details.unixid, svn_pass=passwd)
-
-        logging.info("Enabling user")
-        db.update_user(login, state='enabled')
-
-        return {"response": "okay"}
-
-    finally:
-        db.close()
-
-def rebuild_svn_config(props):
+    # FIXME: check we're pending
+
+    # Get the full User object from the db associated with this
+    user = ivle.database.User.get_by_login(store, login)
+
+    # make svn config/auth
+    repopath = os.path.join(ivle.conf.svn_repo_path, 'users', login)
+    logging.debug("Creating user's Subversion repository")
+    ivle.makeuser.make_svn_repo(repopath, throw_on_error=True)
+
+    rebuild_svn_config(store, props)
+
+    logging.debug("Adding Subversion authentication")
+    passwd = ivle.makeuser.make_svn_auth(store, login,
+                                         throw_on_error=True)
+
+    logging.debug("Creating jail")
+    ivle.makeuser.make_jail(user)
+
+    logging.info("Enabling user")
+    user.state = u'enabled'
+
+    return {"response": "okay"}
+
+def rebuild_svn_config(store, props):
     """Rebuilds the svn config file
     Return value:
         response (okay, failure)
     """
     try:
-        ivle.makeuser.rebuild_svn_config()
+        ivle.makeuser.rebuild_svn_config(store)
     except Exception, e:
         logging.warning('Rebuild of Subversion authorization config failed!')
         return{'response': 'failure', 'msg': repr(e)}
 
     return {'response': 'okay'}
 
-def rebuild_svn_group_config(props):
+def rebuild_svn_group_config(store, props):
     """Rebuilds the svn group config file
     Return value:
         response (okay, failure)
     """
     try:
-        ivle.makeuser.rebuild_svn_group_config()
+        ivle.makeuser.rebuild_svn_group_config(store)
     except Exception, e:
         logging.warning(
             'Rebuild of Subversion group authorization config failed!')
@@ -98,7 +91,7 @@
 
     return {'response': 'okay'}
 
-def create_group_repository(props):
+def create_group_repository(store, props):
     """Creates on disk repository for the given group
     Expected properties:
         subj_short_name, year, semester, groupnm
@@ -141,8 +134,17 @@
 
 def dispatch(props):
     logging.debug(repr(props))
+
+    store = ivle.database.get_store()
     action = props.keys()[0]
-    return actions[action](props[action])
+    res = actions[action](store, props[action])
+
+    if res['response'] == 'okay':
+        store.commit()
+    else:
+        store.rollback()
+    store.close()
+    return res
 
 if __name__ == "__main__":
     pid = os.getpid()

=== modified file 'setup/configure.py'
--- setup/configure.py	2009-01-13 01:36:15 +0000
+++ setup/configure.py	2009-01-15 01:29:07 +0000
@@ -574,8 +574,4 @@
 # The location of the password file used to authenticate users
 # of the subversion repository from the ivle server.
 svn_auth_ivle = os.path.join(svn_path, 'ivle.auth')
-
-# The location of the password file used to authenticate local users
-# of the subversion repository.
-svn_auth_local = os.path.join(svn_path, 'local.auth')
 """)

=== modified file 'setup/install.py'
--- setup/install.py	2009-01-13 01:36:15 +0000
+++ setup/install.py	2009-01-19 08:14:25 +0000
@@ -170,7 +170,7 @@
         try:
             conf = open(ivle_revision_record_file, "w")
 
-            conf.write("""# SVN revision r%d
+            conf.write("""# SVN revision r%s
 # Source tree location: %s
 # Modified files:
 """ % (util.get_svn_revision(), os.getcwd()))

=== modified file 'subjects/sample/subject.xml'
--- subjects/sample/subject.xml	2008-01-25 06:20:32 +0000
+++ subjects/sample/subject.xml	2009-01-19 07:09:25 +0000
@@ -1,4 +1,4 @@
 <worksheets>
-    <worksheet id="ws1" name="Worksheet 1" />
+    <worksheet id="ws1" name="Worksheet 1" assessable="true" />
     <worksheet id="ws2" name="Worksheet 2" />
 </worksheets>

=== added file 'userdb/migrations/20090120-01.sql'
--- userdb/migrations/20090120-01.sql	1970-01-01 00:00:00 +0000
+++ userdb/migrations/20090120-01.sql	2009-01-20 00:37:29 +0000
@@ -0,0 +1,1 @@
+ALTER TABLE login ALTER COLUMN state SET DEFAULT 'no_agreement';

=== modified file 'userdb/users.sql'
--- userdb/users.sql	2008-09-03 00:38:55 +0000
+++ userdb/users.sql	2009-01-20 00:37:29 +0000
@@ -5,7 +5,8 @@
     login       VARCHAR UNIQUE NOT NULL,
     passhash    VARCHAR,
     state	VARCHAR NOT NULL CHECK (state in ('no_agreement', 'pending',
-                                              'enabled', 'disabled')),
+                                              'enabled', 'disabled'))
+                                 DEFAULT 'no_agreement',
     rolenm      VARCHAR NOT NULL CHECK (rolenm in ('anyone', 'student',
                                                    'marker', 'tutor',
                                                    'lecturer', 'admin')),

=== modified file 'www/apps/browser/__init__.py'
--- www/apps/browser/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/browser/__init__.py	2009-02-02 04:00:25 +0000
@@ -31,6 +31,10 @@
 from ivle import (util, studpath)
 import ivle.svn
 
+import genshi
+import genshi.core
+import genshi.template
+
 # url path for this app
 THIS_APP = "files"
 
@@ -44,6 +48,8 @@
         redirectPath = util.make_path(os.path.join(THIS_APP,req.user.login))
         req.throw_redirect(util.make_path(redirectPath))
 
+    ctx = genshi.template.Context()
+
     # Set request attributes
     req.content_type = "text/html"
     req.styles = [
@@ -74,88 +80,65 @@
             "The path specified is invalid.")
 
     # Start writing data
-    req.write("""
-<!-- Top bar section -->
 
-<div id="topbar">
-  <div id="path">
-    """)
     # FIXME: This isn't completely reliable! We're not inside the jail, so we
     # can't know the type for sure. This is now only used for adding a / to the
     # end of displayed paths, so I'm leaving this although it will often break.
     isdir = os.path.isdir(localpath)
-    presentpath(req, browsepath, isdir)
-    req.write("""
-  </div>
-  <div id="actions1">
-""")
-    present_actions1(req)
-    req.write("""  </div>
-  <div id="actions2">
-""")
-    present_actions2(req)
-    req.write("""  </div>
-</div>
-<!-- End topbar -->
-
-<!-- Body. The JavaScript places content here relevant to the path -->
-<div id="filesbody">
-</div>
-<!-- End body -->
-""")
-
-def presentpath(req, path, isdir):
-    """
-    Presents a path list (address bar inside the page) for clicking.
-    Writes to req, expecting to have just written the opening div containing
-    the listing.
-    This will also have a revision indicator on the end, if we are viewing a
-    revision.
-    """
+    ctx['isdir'] = isdir
+    gen_path(req, browsepath, ctx)
+    gen_actions(req, ctx)
+    
+    ctx['fileservice_action'] = util.make_path(os.path.join("fileservice", req.path))
+    ctx['filename'] = cgi.escape(req.path)
+
+    loader = genshi.template.TemplateLoader(".", auto_reload=True)
+    tmpl = loader.load(util.make_local_path("apps/browser/template.html"))
+    
+    req.write(tmpl.generate(ctx).render('html')) #'xhtml', doctype='xhtml'))
+
+#TODO: Move all this logic into the template
+def gen_path(req, path, ctx):
+
     href_path = util.make_path(THIS_APP)
     nav_path = ""
-
     revision = ivle.svn.revision_from_string(
                      req.get_fieldstorage().getfirst('r'))
-
     try: 
         revno = revision.number
     except:
         revno = None
-       
+      
+    ctx['revno'] = revno
+    
     # Create all of the paths
     pathlist = path.split("/")
-    segs_left = len(pathlist)
+    ctx['paths'] = []
     for path_seg in pathlist:
         if path_seg == "":
             continue
-        # Write a slash at the end unless this is the last path seg AND
-        # it's not a directory.
-        segs_left -= 1
-        add_slash = segs_left != 0 or isdir
-        # Make an 'a' element
+        new_seg = {}
+        
+        nav_path = nav_path + path_seg
         href_path = href_path + '/' + path_seg
-        nav_path = nav_path + path_seg
-        if add_slash:
-            nav_path = nav_path + '/'
-        link = '<a href="%s" title="Navigate to %s">%s</a>' % (
-            href_path if revno is None else href_path + '?r=%d' % revno,
-            nav_path, path_seg)
-        req.write(link)
-        if add_slash:
-            req.write('/')
-
-    if revno is not None:
-        req.write(' (revision %d)' % revno)
-
-def present_actions1(req):
+        
+        new_seg['path'] = path_seg        
+        new_seg['nav_path'] = nav_path
+        new_seg['href_path'] = href_path
+        if revno is not None:
+            new_seg['href_path'] += '?r=%d' % revno
+        
+        ctx['paths'].append(new_seg)
+
+
+def gen_actions(req, ctx):
     """
     Presents a set of links/buttons for the "actions1" row of the top bar.
     This is always exactly the same - the JavaScript will customize it later.
     """
     # Set up our actions. The second field of each group is whether to disable
     # the items by default.
-    moreactions = [
+    ctx['moreactions'] = [
       ('Publishing', True, [
         ('publish', ['Publish',          'Make it so this directory can be seen by anyone on the web']),
         ('share',   ['Share this file',  'Get a link to this published file, to give to friends']),
@@ -174,91 +157,15 @@
         ('upload',  ['Upload File',      'Upload a file to the current directory'])
       ]),
       ('Subversion', True, [
-        ('svnadd',      ['Add',          'Schedule the selected temporary files to be added permanently']),
-        ('svnremove',   ['Remove',       'Schedule the selected permanent files to be removed']),
-        ('svndiff',     ['Diff',         'View any changes to the selected file since its last committed state']),
-        ('svnrevert',   ['Revert',       'Restore the selected files back to their last committed state']),
-        ('svnupdate',   ['Update',       'Update your files with changes from the permanent repository']),
-        ('svncommit',   ['Commit',       'Commit any changes to the permanent repository']),
-        ('svnresolved', ['Mark Resolved','Mark a conflicted file as being resolved']),
-        ('svnlog',      ['View Log',     'View the log of commits of the selected file']),
+        ('svncut',      ['Svn Cut',      'Prepare to move the selected files to another directory, maintaining history']),
+        ('svncopy',     ['Svn Copy',     'Prepare to copy the selected files to another directory, maintaining history']),
+        ('svnadd',      ['Add',            'Schedule the selected temporary files to be added permanently']),
+        ('svnremove',   ['Remove',         'Schedule the selected permanent files to be removed']),
+        ('svndiff',     ['Diff',           'View any changes to the selected file since its last committed state']),
+        ('svnrevert',   ['Revert',         'Restore the selected files back to their last committed state']),
+        ('svnupdate',   ['Update',         'Update your files with changes from the permanent repository']),
+        ('svncommit',   ['Commit',         'Commit any changes to the permanent repository']),
+        ('svnresolved', ['Mark Resolved',  'Mark a conflicted file as being resolved']),
+        ('svnlog',      ['View Log',       'View the log of commits of the selected file']),
       ])
     ]
-
-    req.write("""    <a id="act_open" class="disabled">Open</a> :
-    <a id="act_serve"
-        title="View this file on the web, running your code if this is a CGI file"
-        class="disabled" target="_blank">Serve</a> :
-    <a id="act_run" title="Run the selected Python file in the console"
-        class="disabled">Run</a> :
-    <a id="act_download" class="choice">Download</a> :
-    <a title="Refresh the current page" onclick="refresh()"
-        class="choice">Refresh</a><span id="moreactions_area"> :
-    <select id="moreactions" onchange="handle_moreactions()"
-        onblur="handle_moreactions()">
-      <option class="moreactions" value="top"
-        selected="selected">More actions...</option>
-""")
-
-    for (name, disablement, group) in moreactions:
-        if disablement:
-            disable = 'class="disabled" disabled="disabled"'
-        else:
-            disable = ''
-        req.write('<optgroup label="%s">' % name)
-        for (id, bits) in group:
-            req.write('<option id="act_%s" %s title="%s" value="%s">%s</option>'
-                      % (id, disable, bits[1], id, bits[0]))
-        req.write('</optgroup>')
-
-    req.write('</select></span>')
-
-def present_actions2(req):
-    """
-    Presents a set of links/buttons for the "actions2" row of the top bar.
-    This depends on whether it is a directory (listing) or normal file
-    (editor), but we'll let the JavaScript decide which.
-    """
-    req.write("""    <form id="actions2_directory"
-          target="upload_iframe"
-          action="%s"
-          enctype="multipart/form-data" method="post">
-      <p>
-      <span id="selectallpanel">
-      Select:
-      <a onclick="action_selectall(true)"
-          title="Select all files in this directory">All</a> :
-      <a onclick="action_selectall(false)"
-          title="Deselect all files in this directory">None</a>
-      </span>
-
-      <span style="display:none" id="uploadpanel">|
-        <label for="upload_file">Upload file:</label>
-        <input type="hidden" value="putfiles" name="action" />
-        <input type="hidden" value="" name="path" />
-        <input type="file" name="data" id="upload_file" />
-        <input type="checkbox" checked="checked" value="true" name="unpack" id="unpack" />
-        <label title="Extract files and directories from a ZIP file if one is uploaded" for="unpack">Unpack ZIP</label>
-        <input type="button" onclick="show_uploadpanel(false)" value="Cancel" />
-        <input type="submit" value="Upload" />
-      </span>
-      <!-- This iframe is for making a call to upload the file without
-           refreshing the page. (It will refresh the listing). -->
-      <iframe onload="upload_callback()" style="display: none;"
-          name="upload_iframe" id="upload_iframe"></iframe>
-      </p>
-    </form>
-""" % cgi.escape(util.make_path(os.path.join("fileservice", req.path))))
-
-    req.write("""    <p id="actions2_file">
-      <input type="button" id="save_button" value="Save" onclick="save_file('%s')" />
-      <input type="button" id="saveas_button" value="Save As..." onclick="save_file_as('%s')" />
-      <select id="highlighting_select" onchange="highlighting_changed(this)">
-          <option value="text">Text</option>
-          <option value="python">Python</option>
-          <option value="html">HTML</option>
-          <option value="javascript">JavaScript</option>
-          <option value="css">CSS</option>
-      </select>
-    </p>
-""" % ((cgi.escape(req.path),) * 2))

=== added file 'www/apps/browser/template.html'
--- www/apps/browser/template.html	1970-01-01 00:00:00 +0000
+++ www/apps/browser/template.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,102 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+<!-- Top bar section -->
+<body py:strip="">
+  <div id="topbar">
+    <div id="path">
+      <span>
+        <py:for each="segment in paths">/<a href="${segment['href_path']}" title="Navigate to ${segment['nav_path']}">${segment['path']}</a></py:for><py:if test="isdir">/</py:if><py:if test="revno is not None"> (revision ${revno})</py:if>
+      </span>
+    </div>
+    <div id="actions1">
+      <a id="act_open" class="disabled">Open</a> :
+      <a id="act_serve"
+          title="View this file on the web, running your code if this is a CGI file"
+          class="disabled" target="_blank">Serve</a> :
+      <a id="act_run" title="Run the selected Python file in the console"
+          class="disabled">Run</a> :
+      <a id="act_download" class="choice">Download</a> :
+      <a title="Refresh the current page" onclick="refresh()"
+          class="choice">Refresh</a>
+      <span id="moreactions_area"> :
+        <select id="moreactions" onchange="handle_moreactions()"
+            onblur="handle_moreactions()">
+          <option class="moreactions" value="top"
+              selected="selected">More actions...</option>
+          <py:for each="(name, disablement, group) in moreactions">
+
+            <optgroup label="${name}">
+
+              <py:choose test="disablement">
+
+                <py:when test="True">
+                  <py:for each="(id, bits) in group">
+                    <option id="act_${id}" class="disabled" disabled="disabled"
+                        title="${bits[1]}" value="${id}">${bits[0]}</option>
+                  </py:for>
+                </py:when>
+
+                <py:otherwise>
+                  <py:for each="(id, bits) in group">
+                    <option id="act_${id}" title="${bits[1]}" 
+                        value="${id}">${bits[0]}</option>
+                  </py:for>                
+                </py:otherwise>
+
+              </py:choose>
+
+            </optgroup>
+          </py:for>
+        </select>
+      </span>
+    </div>
+    <div id="actions2">
+      <form id="actions2_directory"
+          target="upload_iframe"
+          action="${fileservice_action}"
+          enctype="multipart/form-data" method="post">
+        <p>
+          <span id="selectallpanel">
+          Select:
+            <a onclick="action_selectall(true)"
+                title="Select all files in this directory">All</a> :
+            <a onclick="action_selectall(false)"
+                title="Deselect all files in this directory">None</a>
+          </span>
+
+          <span style="display:none" id="uploadpanel">|
+            <label for="upload_file">Upload file:</label>
+            <input type="hidden" value="putfiles" name="action" />
+            <input type="hidden" value="" name="path" />
+            <input type="file" name="data" id="upload_file" />
+            <input type="checkbox" checked="checked" value="true" name="unpack" id="unpack" />
+            <label title="Extract files and directories from a ZIP file if one is uploaded" for="unpack">Unpack ZIP</label>
+            <input type="button" onclick="show_uploadpanel(false)" value="Cancel" />
+            <input type="submit" value="Upload" />
+          </span>
+          <!-- This iframe is for making a call to upload the file without
+               refreshing the page. (It will refresh the listing). -->
+          <iframe onload="upload_callback()" style="display: none;"
+              name="upload_iframe" id="upload_iframe"></iframe>
+        </p>
+      </form>
+      <p id="actions2_file">
+        <input type="button" id="save_button" value="Save" onclick="save_file('${filename}')" />
+        <input type="button" id="saveas_button" value="Save As..." onclick="save_file_as('${filename}')" />
+        <select id="highlighting_select" onchange="highlighting_changed(this)">
+          <option value="text">Text</option>
+          <option value="python">Python</option>
+          <option value="html">HTML</option>
+          <option value="javascript">JavaScript</option>
+          <option value="css">CSS</option>
+        </select>
+      </p>
+    </div>
+  </div>
+  <!-- Body. The JavaScript places content here relevant to the path -->
+  <div id="filesbody">
+  </div>
+  <!-- End body -->
+</body>
+</html>

=== modified file 'www/apps/debuginfo/__init__.py'
--- www/apps/debuginfo/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/debuginfo/__init__.py	2009-01-14 22:59:55 +0000
@@ -24,6 +24,7 @@
 # Important: This application should be removed from a production system.
 
 import os
+import cgi
 
 from ivle import util
 import ivle.conf
@@ -70,8 +71,6 @@
         ("user", req.user),
         ("hostname", req.hostname),
     ])
-    if req.user is not None:
-        print_table(req, "req.user", dict(req.user).items(), "h4")
 
     # Violate encapsulation here to print out the hidden properties
     print_table(req, "Apache (Hidden) Request Properties", [
@@ -88,9 +87,6 @@
         getfieldvalues(req.get_fieldstorage().items()))
     session = req.get_session()
     print_table(req, "Session Variables", session.items())
-    if 'user' in session:
-        print_table(req, "session['user']", dict(session['user']).items(),
-            "h4")
 
     print_table(req, "HTTP Request Headers",
         req.apache_req.headers_in.items())
@@ -124,6 +120,7 @@
     req.write("<%s>%s</%s>\n" % (head_elem, tablename, head_elem))
     req.write('<table border="1">\n')
     for (k,v) in mapping:
-        req.write("<tr><th>%s</th><td>%s</td></tr>\n" % (str(k), repr(v)))
+        req.write("<tr><th>%s</th><td>%s</td></tr>\n" %
+                  (cgi.escape(str(k)), cgi.escape(repr(v))))
     req.write("</table>\n")
 

=== modified file 'www/apps/diff/__init__.py'
--- www/apps/diff/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/diff/__init__.py	2009-01-20 00:04:15 +0000
@@ -54,6 +54,6 @@
             req.throw_redirect(os.path.join(req.uri,req.user.login));
         interp_object = ivle.interpret.interpreter_objects["cgi-python"]
         user_jail_dir = os.path.join(ivle.conf.jail_base, req.user.login)
-        ivle.interpret.interpret_file(req, req.user.login, user_jail_dir,
+        ivle.interpret.interpret_file(req, req.user, user_jail_dir,
             diffservice_path, interp_object)
 

=== modified file 'www/apps/fileservice/__init__.py'
--- www/apps/fileservice/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/fileservice/__init__.py	2009-01-20 00:04:15 +0000
@@ -61,5 +61,5 @@
     else:
         interp_object = ivle.interpret.interpreter_objects["cgi-python"]
         user_jail_dir = os.path.join(ivle.conf.jail_base, req.user.login)
-        ivle.interpret.interpret_file(req, req.user.login, user_jail_dir,
+        ivle.interpret.interpret_file(req, req.user, user_jail_dir,
             fileservice_path, interp_object, gentle=False)

=== modified file 'www/apps/groups/__init__.py'
--- www/apps/groups/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/groups/__init__.py	2009-02-02 04:00:25 +0000
@@ -25,8 +25,13 @@
 
 import cgi
 
-from ivle import (util, caps)
-import ivle.db
+from ivle import caps
+from ivle.database import Subject
+from ivle import util
+
+import genshi
+import genshi.core
+import genshi.template
 
 def handle(req):
     # Set request attributes
@@ -39,82 +44,69 @@
     ]
     req.write_html_head_foot = True     # Have dispatch print head and foot
 
-    req.write('<div id="ivle_padding">\n')
+    ctx = genshi.template.Context()
+    
+    ctx['enrolments'] = []
     # Show a group panel per enrolment
-    db = ivle.db.DB()
-    try:
-        subjects = db.get_enrolment(req.user.login)
-        # Sort by year,semester,subj_code (newer subjects first)
-        # Leave all fields as strings, just in case (eg. semester='y')
-        subjects.sort(key=lambda(subject):
-                          (subject["year"],subject["semester"],subject["subj_code"]),
-                      reverse=True)
-        if len(subjects) == 0:
-            req.write("<p>Error: You are not currently enrolled in any subjects."
-                      "</p>\n")
-        for subject in subjects:
-            show_subject_panel(req, db, subject['offeringid'],
-                subject['subj_name'])
-        if req.user.hasCap(caps.CAP_MANAGEGROUPS):
-            show_groupadmin_panel(req, db)
+    enrolments = req.user.active_enrolments
+    if enrolments.count() == 0:
+        ctx['no_enrolments'] = True
+    else:
+        ctx['no_enrolments'] = False
+    
+    for enrolment in enrolments:
+        add_subject_panel(req, enrolment.offering, ctx)
         
-        req.write("</div>\n")
-    finally:
-        db.close()
-
-def show_groupadmin_panel(req, db):
-    """
-    Shows the group admin panel
-    """
-    req.write("<hr/>\n")
-    req.write("<h1>Group Administration</h1>")
-    # Choose subject
-    subjects = db.get_subjects()
-    req.write("<label for=\"subject_select\">Subject:</label>\n")
-    req.write("<select id=\"subject_select\">\n")
-    for s in subjects:
-        req.write("    <option value=\"%d\">%s (%s)</option>\n"%
-            (s['subjectid'], s['subj_name'], s['subj_code']))
-    req.write("</select>\n")
-    req.write("<input type=\"button\" value=\"Manage\" \
-        onclick=\"manage_subject()\" />\n")
-    req.write("<div id=\"subject_div\"></div>")
-
-def show_subject_panel(req, db, offeringid, subj_name):
+    if req.user.hasCap(caps.CAP_MANAGEGROUPS):
+        ctx['manage_groups'] = True
+        ctx['manage_subjects'] = []
+        subjects = req.store.find(Subject)
+        for s in subjects:
+            new_s = {}
+            new_s['id'] = s.id
+            new_s['name'] = s.name
+            new_s['code'] = s.code
+            ctx['manage_subjects'].append(new_s)
+    else:
+        ctx['manage_groups'] = False
+  
+    loader = genshi.template.TemplateLoader(".", auto_reload=True)
+    tmpl = loader.load(util.make_local_path("apps/groups/template.html"))
+    
+    req.write(tmpl.generate(ctx).render('html'))
+
+def add_subject_panel(req, offering, ctx):
     """
     Show the group management panel for a particular subject.
     Prints to req.
     """
     # Get the groups this user is in, for this offering
-    groups = db.get_groups_by_user(req.user.login, offeringid=offeringid)
-    if len(groups) == 0:
+    groups = req.user.get_groups(offering)
+    if groups.count() == 0:
         return
-
-    req.write("<div id=\"subject%d\"class=\"subject\">"%offeringid)
-    req.write("<h1>%s</h1>\n" % cgi.escape(subj_name))
-    for groupid, groupnm, group_nick, is_member in groups:
-        if group_nick is None:
-            group_nick = "";
-        req.write("<h2>%s (%s)</h2>\n" %
-            (cgi.escape(group_nick), cgi.escape(groupnm)))
-        if is_member:
-            req.write('<p>You are a member of this group.</p>\n')
-        else:
-            req.write('<p>You have been invited to this group.</p>\n')
-            req.write('<p>'
-                '<input type="button" '
-                'onclick="accept(&quot;%(groupnm)s&quot;)" '
-                'value="Accept" />\n'
-                '<input type="button" '
-                'onclick="decline(&quot;%(groupnm)s&quot;)" '
-                'value="Decline" />\n'
-                '</p>\n' % {"groupnm": cgi.escape(groupnm)})
-        req.write("<h3>Members</h3>\n")
-        req.write("<ul>\n")
-        for user in db.get_projectgroup_members(groupid):
-            req.write("<li>%s (%s)</li>" %
-                      (cgi.escape(user['fullname']),
-                       cgi.escape(user['login'])))
-        req.write("</ul>\n")
-
-    req.write("</div>")
+    
+    offering_groups = {}
+    
+    offering_groups['offering_id'] = offering.id
+    offering_groups['offering_name'] = offering.subject.name
+    offering_groups['groups'] = []
+    
+    #TODO: Use a better way to manage group membership and invitations
+    for group in groups:
+        new_group = {}
+        new_group['nick'] = cgi.escape(group.nick if group.nick else '')
+        new_group['name'] = cgi.escape(group.name)
+        
+        # XXX - This should be set to reflect whether or not a user is invited
+        #     - or if they have accepted the offer
+        new_group['is_member'] = True
+        new_group['members'] = []
+        
+        for user in group.members:
+            member = {}
+            member['fullname'] = cgi.escape(user.fullname)
+            member['login'] = cgi.escape(user.login)
+            new_group['members'].append(member)
+        offering_groups['groups'].append(new_group)
+
+    ctx['enrolments'].append(offering_groups)

=== added file 'www/apps/groups/template.html'
--- www/apps/groups/template.html	1970-01-01 00:00:00 +0000
+++ www/apps/groups/template.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,52 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+
+  <body py:strip="">
+    <div id="ivle_padding">
+      <p py:if="no_enrolments">Error: You are not currently enrolled in any subjects.</p>
+      <py:for each="enrolment in enrolments">
+        <div id="subject${enrolment.offering_id}" class="subject">
+          <h1>${enrolment.offering_name}</h1>
+          <py:for each="group in enrolment['groups']">
+            <h2>${group['nick']} (${group['name']})</h2>
+            <py:choose test="group['is_member']">
+              <py:when test="True">
+                <p>You are a member of this group.</p>
+              </py:when>
+              <py:otherwise>
+                <p>You have been invited to this group.</p>
+                <p>
+                  <input type="button" 
+                      onclick="accept(&quot;${group['name']}&quot;)" 
+                      value="Accept" />
+                  <input type="button"
+                      onclick="decline(&quot;${group['name']}&quot;)"
+                      value="Decline" />
+                </p>
+              </py:otherwise>
+            </py:choose>
+            <h3>Members</h3>
+            <ul>
+              <py:for each="member in group['members']">
+                <li>${member['fullname']} (${member['login']})</li>
+              </py:for>
+            </ul>
+          </py:for>
+        </div>
+      </py:for>
+      <py:if test="manage_groups">
+        <hr />
+        <h1>Group Administration</h1>
+        <label for="subject_select">Subject:</label>
+        <select id="subject_select">
+          <py:for each="subject in manage_subjects">
+            <option value="${subject['id']}">${subject['name']} (${subject['code']})</option>
+          </py:for>
+        </select>
+        <input type="button" value="Manage" onclick="manage_subject()" />
+        <div id="subject_div"></div>
+      </py:if>
+    </div>
+  </body>
+</html>

=== modified file 'www/apps/help/__init__.py'
--- www/apps/help/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/help/__init__.py	2009-02-02 04:00:25 +0000
@@ -28,6 +28,9 @@
 from ivle import util
 import ivle.conf
 
+import genshi
+import genshi.template
+
 # TODO: Nicer 404 errors
 
 def handle(req):
@@ -50,28 +53,28 @@
 
 def show_help_menu(req):
     """Show the help menu."""
-
     # Set request attributes
     req.content_type = "text/html"
     req.write_html_head_foot = True
 
-    # Start writing data
-    req.write('<div id="ivle_padding">\n')
-    req.write("<h1>Help</h1>\n")
+    ctx = genshi.template.Context()
 
     # Write a list of links to all apps with help modules
-    req.write("<ul>\n")
+    ctx['apps'] = []
     # Tab apps, in order of tabs
     for appurl in ivle.conf.apps.apps_in_tabs:
         app = ivle.conf.apps.app_url[appurl]
         if app.hashelp:
-            req.write('  <li><a href="%s">%s</a></li>\n'
-                % (os.path.join(util.make_path("help"), appurl), app.name))
+            new_app = {}
+            new_app['appurl'] = os.path.join(util.make_path("help"), appurl)
+            new_app['name'] = app.name
+            ctx['apps'].append(new_app)
     # Terms of Service
-    req.write('  <li><a href="%s">Terms of Service</a></li>\n'
-        % (util.make_path("tos")))
-    req.write("</ul>\n")
-    req.write('</div>\n')
+    ctx['tos'] = util.make_path("tos")
+
+    loader = genshi.template.TemplateLoader(".", auto_reload=True)
+    tmpl = loader.load(util.make_local_path("apps/help/template.html"))
+    req.write(tmpl.generate(ctx).render('html')) #'xhtml', doctype='xhtml'))
 
 def show_help_app(req, app):
     """Show help for an application."""

=== added file 'www/apps/help/template.html'
--- www/apps/help/template.html	1970-01-01 00:00:00 +0000
+++ www/apps/help/template.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,17 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+
+  <body py:strip="">
+  
+    <div id="ivle_padding">
+      <h1>Help</h1>
+      <ul>
+        <py:for each="app in apps">
+          <li><a href="${app['appurl']}">${app['name']}</a></li>
+        </py:for>
+        <li><a href="${tos}">Terms of Service</a></li>
+      </ul>
+    </div>
+  </body>
+</html>

=== modified file 'www/apps/logout/__init__.py'
--- www/apps/logout/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/logout/__init__.py	2009-02-02 04:00:25 +0000
@@ -19,9 +19,11 @@
 # Author: Nick Chadwick
 # Date: 13/01/2009
 
-import cgi
 from ivle import util
 
+import genshi
+import genshi.template
+
 # url path for this app
 THIS_APP = "logout"
 
@@ -29,10 +31,10 @@
     if req.method == "POST":
         req.logout()
     else:
+        req.content_type = "text/html"
         req.write_html_head_foot = True
-        req.content_type = "text/html"
-        req.write('<div id="ivle_padding">\n'
-                  '<h3>Are you sure you want to logout?</h3><p>'
-                  '<form action="%s" method="POST">\n'
-                  '    <input type="submit" value="Logout" />\n'
-                  '</form>\n</div>\n' % (cgi.escape(util.make_path('logout'))))
+        ctx = genshi.template.Context()
+        ctx['path'] =  util.make_path('logout')
+        loader = genshi.template.TemplateLoader(".", auto_reload=True)
+        tmpl = loader.load(util.make_local_path("apps/logout/template.html"))
+        req.write(tmpl.generate(ctx).render('html')) #'xhtml', doctype='xhtml'))

=== added file 'www/apps/logout/template.html'
--- www/apps/logout/template.html	1970-01-01 00:00:00 +0000
+++ www/apps/logout/template.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,15 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+
+  <body py:strip="">
+    <div id="ivle_padding">
+      <h3>Are you sure you want to logout?</h3>
+      <p>
+        <form action="${path}" method="POST">
+          <input type="submit" value="Logout" />
+        </form>
+      </p>
+    </div>
+  </body>
+</html>

=== modified file 'www/apps/server/__init__.py'
--- www/apps/server/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/server/__init__.py	2009-01-20 00:04:15 +0000
@@ -29,6 +29,7 @@
 
 from ivle import (util, studpath, interpret)
 import ivle.conf
+from ivle.database import User
 
 serveservice_path = os.path.join(ivle.conf.share_path, 'services/serveservice')
 interpretservice_path = os.path.join(ivle.conf.share_path,
@@ -45,16 +46,15 @@
 
     # Get the username of the student whose work we are browsing, and the path
     # on the local machine where the file is stored.
-    (user, path) = studpath.url_to_local(req.path)
+    (login, path) = studpath.url_to_local(req.path)
 
-    try:
-        interpret.get_uid(user)
-    except KeyError:
+    owner = User.get_by_login(req.store, login)
+    if not owner:
         # There is no user.
         req.throw_error(req.HTTP_NOT_FOUND,
             "The path specified is invalid.")
 
-    serve_file(req, user, path)
+    serve_file(req, owner, path)
 
 def authorize(req):
     """Given a request, checks whether req.username is allowed to
@@ -76,9 +76,9 @@
     serving it directly, or denying it and returning a 403 Forbidden error.
     No return value. Writes to req (possibly throwing a server error exception
     using req.throw_error).
-    
+
     req: An IVLE request object.
-    owner: Username of the user who owns the file being served.
+    owner: The user who owns the file being served.
     filename: Filename in the local file system.
     download:  Should the file be viewed in browser or downloaded
     """
@@ -86,12 +86,12 @@
     # We need a no-op trampoline run to ensure that the jail is mounted.
     # Otherwise we won't be able to authorise for public mode!
     noop_object = interpret.interpreter_objects["noop"]
-    user_jail_dir = os.path.join(ivle.conf.jail_base, owner)
+    user_jail_dir = os.path.join(ivle.conf.jail_base, owner.login)
     interpret.interpret_file(req, owner, user_jail_dir, '', noop_object)
 
     # Authorize access. If failure, this throws a HTTP_FORBIDDEN error.
     authorize(req)
-    
+
     # Jump into the jail
     interp_object = interpret.interpreter_objects["cgi-python"]
     if download:

=== modified file 'www/apps/settings/__init__.py'
--- www/apps/settings/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/settings/__init__.py	2009-02-02 04:00:25 +0000
@@ -24,6 +24,9 @@
 
 from ivle import util
 
+import genshi
+import genshi.template
+
 def handle(req):
     """Handler for the Settings application."""
 
@@ -44,30 +47,6 @@
     ]
     req.write_html_head_foot = True     # Have dispatch print head and foot
 
-    # Start writing data
-    req.write("""<div id="ivle_padding">
-
-  <h2>User Profile</h2>
-  <p><span id="login"></span></p>
-  <p id="role"></p>
-  <h3>Change settings</h3>
-  <form action="">
-    <table>
-      <tr><td>Display name:</td><td><input type="text"
-            name="nick" id="nick" size="40" /></td></tr>
-      <tr><td>Email address:</td><td><input type="text"
-            name="email" id="email" size="40" /></td></tr>
-    </table>
-    <div id="changepassword"></div>
-    <p id="result"></p>
-    <p>
-      <input value="Save" onclick="return save_settings()" id="save"
-          type="submit" />
-      <input value="Revert" onclick="return revert_settings()" id="revert"
-          type="reset" />
-    </p>
-  </form>
-
-  <div id="notices"></div>
-</div>
-""")
+    loader = genshi.template.TemplateLoader(".", auto_reload=True)
+    tmpl = loader.load(util.make_local_path("apps/settings/template.html"))
+    req.write(tmpl.generate().render('html')) #'xhtml', doctype='xhtml'))

=== added file 'www/apps/settings/template.html'
--- www/apps/settings/template.html	1970-01-01 00:00:00 +0000
+++ www/apps/settings/template.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,43 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+
+  <body py:strip="">
+    <div id="ivle_padding">
+
+      <h2>User Profile</h2>
+      <p>
+        <span id="login"></span>
+      </p>
+      <p id="role"></p>
+      <h3>Change settings</h3>
+      <form action="">
+        <table>
+          <tr>
+            <td>Display name:</td>
+            <td>
+              <input type="text"
+                  name="nick" id="nick" size="40" />
+            </td>
+          </tr>
+          <tr>
+            <td>Email address:</td>
+            <td>
+              <input type="text"
+                  name="email" id="email" size="40" />
+            </td>
+          </tr>
+        </table>
+        <div id="changepassword"></div>
+        <p id="result"></p>
+        <p>
+          <input value="Save" onclick="return save_settings()" id="save"
+                type="submit" />
+          <input value="Revert" onclick="return revert_settings()" id="revert"
+                type="reset" />
+        </p>
+      </form>
+      <div id="notices"></div>
+    </div>
+  </body>
+</html>

=== modified file 'www/apps/subjects/__init__.py'
--- www/apps/subjects/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/subjects/__init__.py	2009-02-02 04:00:25 +0000
@@ -26,58 +26,62 @@
 import urllib
 import cgi
 
+import ivle.database
 from ivle import util
-import ivle.db
+
+import genshi
+import genshi.template
 
 def handle(req):
     """Handler for the Subjects application. Links to subject home pages."""
 
     req.styles = ["media/subjects/subjects.css"]
+    ctx = genshi.template.Context()
     if req.path == "":
-        handle_toplevel_menu(req)
+        # This is represented as a directory. Redirect and add a slash if it is
+        # missing.
+        if req.uri[-1] != '/':
+            req.throw_redirect(req.uri + '/')
+        ctx['whichpage'] = "toplevel"
+        handle_toplevel_menu(req, ctx)
     else:
-        handle_subject_page(req, req.path)
-
-def handle_toplevel_menu(req):
-    # This is represented as a directory. Redirect and add a slash if it is
-    # missing.
-    if req.uri[-1] != '/':
-        req.throw_redirect(req.uri + '/')
-
-    (enrolled_subjects, unenrolled_subjects) = \
-              ivle.db.DB().get_subjects_status(req.user.login)
-
-    def print_subject(subject):
-        if subject['url'] is None:
-            req.write('  <li>%s (no home page)</li>\n'
-                % cgi.escape(subject['subj_name']))
-        else:
-            req.write('  <li><a href="%s">%s</a></li>\n'
-                % (cgi.escape(subject['url']),
-                   cgi.escape(subject['subj_name'])))
+        ctx['whichpage'] = "subject"
+        handle_subject_page(req, req.path, ctx)
+        
+    loader = genshi.template.TemplateLoader(".", auto_reload=True)
+    tmpl = loader.load(util.make_local_path("apps/subjects/template.html"))
+    req.write(tmpl.generate(ctx).render('html')) #'xhtml', doctype='xhtml'))
+
+def handle_toplevel_menu(req, ctx):
+
+    enrolled_subjects = req.user.subjects
+    unenrolled_subjects = [subject for subject in
+                           req.store.find(ivle.database.Subject)
+                           if subject not in enrolled_subjects]
+
+    ctx['enrolled_subjects'] = []
+    ctx['other_subjects'] = []
 
     req.content_type = "text/html"
     req.write_html_head_foot = True
-    req.write('<div id="ivle_padding">\n')
-    req.write("<h2>IVLE Subject Homepages</h2>\n")
-    req.write("<h2>Subjects</h2>\n<ul>\n")
+
     for subject in enrolled_subjects:
-        print_subject(subject)
-    req.write("</ul>\n")
+        new_subj = {}
+        new_subj['name'] = subject.name
+        new_subj['url'] = subject.url
+        ctx['enrolled_subjects'].append(new_subj)
+
     if len(unenrolled_subjects) > 0:
-        req.write("<h3>Other Subjects</h3>\n")
-        req.write("<p>You are not currently enrolled in these subjects</p>\n")
-        req.write("<ul>\n")
         for subject in unenrolled_subjects:
-            print_subject(subject)
-        req.write("</ul>\n")
-    req.write("</div>\n")
-
-def handle_subject_page(req, path):
+            new_subj = {}
+            new_subj['name'] = subject.name
+            new_subj['url'] = subject.url
+            ctx['other_subjects'].append(new_subj)
+
+
+def handle_subject_page(req, path, ctx):
     req.content_type = "text/html"
     req.write_html_head_foot = True     # Have dispatch print head and foot
 
     # Just make the iframe pointing to media/subjects
-    serve_loc = util.make_path(os.path.join('media', 'subjects', path))
-    req.write('<object class="fullscreen" type="text/html" \
-data="%s"></iframe>'% urllib.quote(serve_loc))
+    ctx['serve_loc'] = urllib.quote(util.make_path(os.path.join('media', 'subjects', path)))

=== added file 'www/apps/subjects/subject_page.html'
--- www/apps/subjects/subject_page.html	1970-01-01 00:00:00 +0000
+++ www/apps/subjects/subject_page.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,8 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+
+  <body py:strip="">
+    <object class="fullscreen" type="text/html" data="${serve_loc}"></iframe>
+  </body>
+</html>

=== added file 'www/apps/subjects/template.html'
--- www/apps/subjects/template.html	1970-01-01 00:00:00 +0000
+++ www/apps/subjects/template.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,9 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+
+  <py:choose test="whichpage">
+    <xi:include py:when="'toplevel'" href="toplevel_menu.html" />
+    <xi:include py:when="'subject'" href="subjectmenu.html" />
+  </py:choose>
+</html>

=== added file 'www/apps/subjects/toplevel_menu.html'
--- www/apps/subjects/toplevel_menu.html	1970-01-01 00:00:00 +0000
+++ www/apps/subjects/toplevel_menu.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,41 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+
+  <body py:strip="">
+    <div id="ivle_padding">
+      <h2>IVLE Subject Homepages</h2>
+      <py:if test="(len(enrolled_subjects) > 0)">
+        <h2>Subjects</h2>
+        <ul>
+          <py:for each="subject in enrolled_subjects">
+            <py:choose test="subject['url']">
+              <py:when test="None">
+                <li>${subject['name']} (no home page)</li>
+              </py:when>
+              <py:otherwise>
+                <li><a href="${subject['url']}">${subject['name']}</a></li>
+              </py:otherwise>
+            </py:choose>
+          </py:for>
+        </ul>
+      </py:if>
+      <py:if test="(len(other_subjects) > 0)">
+        <h3>Other Subjects</h3>
+        <p>You are not currently enrolled in these subjects</p>
+        <ul>
+          <py:for each="subject in other_subjects">
+            <py:choose test="subject['url']">
+             <py:when test="None">
+               <li>${subject['name']} (no home page)</li>
+             </py:when>
+             <py:otherwise>
+               <li><a href="${subject['url']}">${subject['name']}</a></li>
+             </py:otherwise>
+            </py:choose>
+          </py:for>
+        </ul>
+      </py:if>
+    </div>
+  </body>
+</html>

=== modified file 'www/apps/svnlog/__init__.py'
--- www/apps/svnlog/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/svnlog/__init__.py	2009-01-20 00:04:15 +0000
@@ -35,7 +35,7 @@
         req.throw_redirect(os.path.join(req.uri, req.user.login))
     interpreter = ivle.interpret.interpreter_objects["cgi-python"]
     jail_dir = os.path.join(ivle.conf.jail_base, req.user.login)
-    ivle.interpret.interpret_file(req, req.user.login, jail_dir,
+    ivle.interpret.interpret_file(req, req.user, jail_dir,
           os.path.join(ivle.conf.share_path, 'services/svnlogservice'),
           interpreter)
 

=== modified file 'www/apps/tutorial/__init__.py'
--- www/apps/tutorial/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/tutorial/__init__.py	2009-02-03 03:59:37 +0000
@@ -30,7 +30,7 @@
 
 import os
 import os.path
-import time
+from datetime import datetime
 import cgi
 import urllib
 import re
@@ -41,10 +41,15 @@
 
 from ivle import util
 import ivle.conf
-import ivle.db
+import ivle.database
+import ivle.worksheet
 
 from rst import rst
 
+import genshi
+import genshi.core
+import genshi.template
+
 THIS_APP = "tutorial"
 
 # Regex for valid identifiers (subject/worksheet names)
@@ -55,6 +60,11 @@
         self.id = id
         self.name = name
         self.assessable = assessable
+        self.loc = urllib.quote(id)
+        self.complete_class = ''
+        self.optional_message = ''
+        self.total = 0
+        self.mand_done = 0
     def __repr__(self):
         return ("Worksheet(id=%s, name=%s, assessable=%s)"
                 % (repr(self.id), repr(self.name), repr(self.assessable)))
@@ -76,6 +86,9 @@
 def handle(req):
     """Handler for the Tutorial application."""
 
+    # TODO: Take this as an argument instead (refactor dispatch)
+    ctx = genshi.template.Context()
+
     # Set request attributes
     req.content_type = "text/html"
     req.scripts = [
@@ -106,11 +119,20 @@
             worksheet = path_segs[1]
 
     if subject == None:
-        handle_toplevel_menu(req)
+        ctx['whichmenu'] = 'toplevel'
+        handle_toplevel_menu(req, ctx)
     elif worksheet == None:
-        handle_subject_menu(req, subject)
+        ctx['whichmenu'] = 'subjectmenu'
+        handle_subject_menu(req, ctx, subject)
     else:
-        handle_worksheet(req, subject, worksheet)
+        ctx['whichmenu'] = 'worksheet'
+        handle_worksheet(req, ctx, subject, worksheet)
+
+    # Use Genshi to render out the template
+    # TODO: Dispatch should do this instead
+    loader = genshi.template.TemplateLoader(".", auto_reload=True)
+    tmpl = loader.load(util.make_local_path("apps/tutorial/template.html"))
+    req.write(tmpl.generate(ctx).render('html')) #'xhtml', doctype='xhtml'))
 
 def handle_media_path(req):
     """
@@ -138,45 +160,23 @@
     req.content_type = type
     req.sendfile(filename)
 
-def handle_toplevel_menu(req):
+def handle_toplevel_menu(req, ctx):
     # This is represented as a directory. Redirect and add a slash if it is
     # missing.
     if req.uri[-1] != '/':
         req.throw_redirect(make_tutorial_path())
     req.write_html_head_foot = True
-    req.write('<div id="ivle_padding">\n')
-    req.write("<h1>IVLE Tutorials</h1>\n")
-    req.write("""<p>Welcome to the IVLE tutorial system.
-  Please select a subject from the list below to select a worksheet
-  for that subject.</p>\n""")
-
-    (enrolled_subjects, unenrolled_subjects) = \
-              ivle.db.DB().get_subjects_status(req.user.login)
-
-    def print_subject(subject):
-        req.write('  <li><a href="%s">%s</a></li>\n'
-            % (urllib.quote(subject['subj_code']) + '/',
-               cgi.escape(subject['subj_name'])))
-
-    req.write("<h2>Subjects</h2>\n<ul>\n")
-    for subject in enrolled_subjects:
-        print_subject(subject)
-    req.write("</ul>\n")
-    if len(unenrolled_subjects) > 0:
-        req.write("<h3>Other Subjects</h3>\n")
-        req.write("<p>You are not currently enrolled in these subjects.\n"
-                  "   Your marks will not be counted.</p>\n")
-        req.write("<ul>\n")
-        for subject in unenrolled_subjects:
-            print_subject(subject)
-        req.write("</ul>\n")
-    req.write("</div>\n")   # tutorialbody
+
+    ctx['enrolled_subjects'] = req.user.subjects
+    ctx['unenrolled_subjects'] = [subject for subject in
+                           req.store.find(ivle.database.Subject)
+                           if subject not in ctx['enrolled_subjects']]
 
 def is_valid_subjname(subject):
     m = re_ident.match(subject)
     return m is not None and m.end() == len(subject)
 
-def handle_subject_menu(req, subject):
+def handle_subject_menu(req, ctx, subject):
     # This is represented as a directory. Redirect and add a slash if it is
     # missing.
     if req.uri[-1] != '/':
@@ -188,104 +188,102 @@
     # Parse the subject description file
     # The subject directory must have a file "subject.xml" in it,
     # or it does not exist (404 error).
+
+    ctx['subject'] = subject
     try:
         subjectfile = open(os.path.join(ivle.conf.subjects_base, subject,
-            "subject.xml"))
+            "subject.xml")).read()
     except:
         req.throw_error(req.HTTP_NOT_FOUND,
             "Subject %s not found." % repr(subject))
 
-    # Read in data about the subject
-    subjectdom = minidom.parse(subjectfile)
-    subjectfile.close()
-    # TEMP: All of this is for a temporary XML format, which will later
-    # change.
-    worksheetsdom = subjectdom.documentElement
-    worksheets = []     # List of string IDs
-    for worksheetdom in worksheetsdom.childNodes:
-        if worksheetdom.nodeType == worksheetdom.ELEMENT_NODE:
-            # Get the 3 attributes for this node and construct a Worksheet
-            # object.
-            # (Note: assessable will default to False, unless it is explicitly
-            # set to "true").
-            worksheet = Worksheet(worksheetdom.getAttribute("id"),
-                worksheetdom.getAttribute("name"),
-                worksheetdom.getAttribute("assessable") == "true")
-            worksheets.append(worksheet)
-
-    db = ivle.db.DB()
-    try:
-        # Now all the errors are out the way, we can begin writing
-        req.title = "Tutorial - %s" % subject
-        req.write_html_head_foot = True
-        req.write('<div id="ivle_padding">\n')
-        req.write("<h1>IVLE Tutorials - %s</h1>\n" % cgi.escape(subject))
-        req.write('<h2>Worksheets</h2>\n<ul id="tutorial-toc">\n')
-        # As we go, calculate the total score for this subject
-        # (Assessable worksheets only, mandatory problems only)
-        problems_done = 0
-        problems_total = 0
-        for worksheet in worksheets:
-            req.write('  <li><a href="%s">%s</a>'
-                % (urllib.quote(worksheet.id), cgi.escape(worksheet.name)))
-            try:
-                # If the assessable status of this worksheet has changed,
-                # update the DB
-                # (Note: This fails the try block if the worksheet is not yet
-                # in the DB, which is fine. The author should visit the
-                # worksheet page to get it into the DB).
-                if (db.worksheet_is_assessable(subject, worksheet.id) !=
-                    worksheet.assessable):
-                    db.set_worksheet_assessable(subject, worksheet.id,
-                        assessable=worksheet.assessable)
-                if worksheet.assessable:
-                    mand_done, mand_total, opt_done, opt_total = (
-                        db.calculate_score_worksheet(req.user.login, subject,
-                            worksheet.id))
-                    if opt_total > 0:
-                        optional_message = " (excluding optional exercises)"
-                    else:
-                        optional_message = ""
-                    if mand_done >= mand_total:
-                        complete_class = "complete"
-                    elif mand_done > 0:
-                        complete_class = "semicomplete"
-                    else:
-                        complete_class = "incomplete"
-                    problems_done += mand_done
-                    problems_total += mand_total
-                    req.write('\n    <ul><li class="%s">'
-                            'Completed %d/%d%s</li></ul>\n  '
-                            % (complete_class, mand_done, mand_total,
-                                optional_message))
-            except ivle.db.DBException:
-                # Worksheet is probably not in database yet
-                pass
-            req.write('</li>\n')
-        req.write("</ul>\n")
-        if problems_total > 0:
-            if problems_done >= problems_total:
-                complete_class = "complete"
-            elif problems_done > 0:
-                complete_class = "semicomplete"
-            else:
-                complete_class = "incomplete"
-            problems_pct = (100 * problems_done) / problems_total       # int
-            req.write('<ul><li class="%s">Total exercises completed: %d/%d '
-                        '(%d%%)</li></ul>\n'
-                % (complete_class, problems_done, problems_total,
-                    problems_pct))
-            # XXX Marks calculation (should be abstracted out of here!)
-            # percent / 16, rounded down, with a maximum mark of 5
-            max_mark = 5
-            mark = min(problems_pct / 16, max_mark)
-            req.write('<p style="font-weight: bold">Worksheet mark: %d/%d'
-                        '</p>\n' % (mark, max_mark))
-        req.write("</div>\n")   # tutorialbody
-    finally:
-        db.close()
-
-def handle_worksheet(req, subject, worksheet):
+    subjectfile = genshi.Stream(list(genshi.XML(subjectfile)))
+
+    ctx['worksheets'] = get_worksheets(subjectfile)
+    
+    # Now all the errors are out the way, we can begin writing
+
+    req.write_html_head_foot = True
+    # As we go, calculate the total score for this subject
+    # (Assessable worksheets only, mandatory problems only)
+    problems_done = 0
+    problems_total = 0
+    for worksheet in ctx['worksheets']:
+        stored_worksheet = ivle.database.Worksheet.get_by_name(req.store,
+            subject, worksheet.id)
+        # If worksheet is not in database yet, we'll simply not display
+        # data about it yet (it should be added as soon as anyone visits
+        # the worksheet itself).
+        if stored_worksheet is not None:
+            # If the assessable status of this worksheet has changed,
+            # update the DB
+            # (Note: This fails the try block if the worksheet is not yet
+            # in the DB, which is fine. The author should visit the
+            # worksheet page to get it into the DB).
+            if worksheet.assessable != stored_worksheet.assessable:
+                # XXX If statement to avoid unnecessary database writes.
+                # Is this necessary, or will Storm check for us?
+                stored_worksheet.assessable = worksheet.assessable
+                req.store.commit()
+            if worksheet.assessable:
+                # Calculate the user's score for this worksheet
+                mand_done, mand_total, opt_done, opt_total = (
+                    ivle.worksheet.calculate_score(req.store, req.user,
+                        stored_worksheet))
+                if opt_total > 0:
+                    optional_message = " (excluding optional exercises)"
+                else:
+                    optional_message = ""
+                if mand_done >= mand_total:
+                    worksheet.complete_class = "complete"
+                elif mand_done > 0:
+                    worksheet.complete_class = "semicomplete"
+                else:
+                    worksheet.complete_class = "incomplete"
+                problems_done += mand_done
+                problems_total += mand_total
+                worksheet.mand_done = mand_done
+                worksheet.total = mand_total
+                worksheet.optional_message = optional_message
+
+
+    ctx['problems_total'] = problems_total
+    ctx['problems_done'] = problems_done
+    if problems_total > 0:
+        if problems_done >= problems_total:
+            ctx['complete_class'] = "complete"
+        elif problems_done > 0:
+            ctx['complete_class'] = "semicomplete"
+        else:
+            ctx['complete_class'] = "incomplete"
+        ctx['problems_pct'] = (100 * problems_done) / problems_total
+        # TODO: Put this somewhere else! What is this on about? Why 16?
+        # XXX Marks calculation (should be abstracted out of here!)
+        # percent / 16, rounded down, with a maximum mark of 5
+        ctx['max_mark'] = 5
+        ctx['mark'] = min(ctx['problems_pct'] / 16, ctx['max_mark'])
+
+def get_worksheets(subjectfile):
+    '''Given a subject stream, get all the worksheets and put them in ctx'''
+    worksheets = []
+    for kind, data, pos in subjectfile:
+        if kind is genshi.core.START:
+            if data[0] == 'worksheet':
+                worksheetid = ''
+                worksheetname = ''
+                worksheetasses = False
+                for attr in data[1]:
+                    if attr[0] == 'id':
+                        worksheetid = attr[1]
+                    elif attr[0] == 'name':
+                        worksheetname = attr[1]
+                    elif attr[0] == 'assessable':
+                        worksheetasses = attr[1] == 'true'
+                worksheets.append(Worksheet(worksheetid, worksheetname, \
+                                                            worksheetasses))
+    return worksheets
+
+def handle_worksheet(req, ctx, subject, worksheet):
     # Subject and worksheet names must be valid identifiers
     if not is_valid_subjname(subject) or not is_valid_subjname(worksheet):
         req.throw_error(req.HTTP_NOT_FOUND,
@@ -301,141 +299,65 @@
     except:
         req.throw_error(req.HTTP_NOT_FOUND,
             "Worksheet file not found.")
-    worksheetmtime = time.localtime(worksheetmtime)
-
-    worksheetdom = minidom.parse(worksheetfile)
-    worksheetfile.close()
-    # TEMP: All of this is for a temporary XML format, which will later
-    # change.
-    worksheetdom = worksheetdom.documentElement
-    if worksheetdom.tagName != "worksheet":
-        req.throw_error(req.HTTP_INTERNAL_SERVER_ERROR,
-            "The worksheet XML file's top-level element must be <worksheet>.")
-    worksheetname = worksheetdom.getAttribute("name")
-
-    # Now all the errors are out the way, we can begin writing
-    req.title = "Tutorial - %s" % worksheetname
+    worksheetmtime = datetime.fromtimestamp(worksheetmtime)
+    worksheetfile = worksheetfile.read()
+    
+    ctx['worksheetstream'] = genshi.Stream(list(genshi.XML(worksheetfile)))
+
     req.write_html_head_foot = True
-    req.write('<div id="ivle_padding">\n')
-    req.write("<h1>IVLE Tutorials - %s</h1>\n<h2>%s</h2>\n"
-        % (cgi.escape(subject), cgi.escape(worksheetname)))
-    exercise_list = present_table_of_contents(req, worksheetdom, 0)
-    # If the database is missing this worksheet or out of date, update its
-    # details about this worksheet
-    # Note: Do NOT set assessable (this is done at the subject level).
-    update_db_worksheet(subject, worksheet, worksheetmtime, exercise_list)
-
-    # Write each element
-    exerciseid = 0
-    for node in worksheetdom.childNodes:
-        exerciseid = present_worksheet_node(req, node, exerciseid)
-    req.write("</div>\n")   # tutorialbody
-
-def present_table_of_contents(req, node, exerciseid):
-    """Given a node of a worksheet XML document, writes out a table of
-    contents to the request. This recursively searches for "excercise"
-    and heading elements to write out.
-
-    When exercise elements are encountered, the DB is queried for their
-    completion status, and the ball is shown of the appropriate colour.
-
-    exerciseid is the ID to use for the first exercise.
-
-    As a secondary feature, this records the identifier (xml filename) and
-    optionality of each exercise in a list of pairs [(str, bool)], and returns
-    this list. This can be used to cache this information in the database.
-    """
-    exercise_list = []
-    # XXX This means the DB is queried twice for each element.
-    # Consider caching these results for lookup later.
-    req.write("""<div id="tutorial-toc">
-<h2>Worksheet Contents</h2>
-<ul>
-""")
-    db = ivle.db.DB()
-    try:
-        for tag, xml in find_all_nodes(req, node):
-            if tag == "ex":
-                # Exercise node
-                # Fragment ID is an accumulating exerciseid
-                # (The same algorithm is employed when presenting exercises)
-                fragment_id = "exercise%d" % exerciseid
-                exerciseid += 1
-                exercisesrc = xml.getAttribute("src")
-                # Optionality: Defaults to False
-                exerciseoptional = xml.getAttribute("optional") == "true"
-                # Record the name and optionality for returning in the list
-                exercise_list.append((exercisesrc, exerciseoptional))
-                # TODO: Get proper exercise title
-                title = exercisesrc
-                # Get the completion status of this exercise
-                complete, _ = db.get_problem_status(req.user.login,
-                    exercisesrc)
-                req.write('  <li class="%s" id="toc_li_%s"><a href="#%s">%s'
-                    '</a></li>\n'
-                    % ("complete" if complete else "incomplete",
-                        fragment_id, fragment_id, cgi.escape(title)))
+
+    ctx['subject'] = subject
+    
+    #TODO: Replace this with a nice way, possibly a match template
+    generate_worksheet_data(ctx, req)
+    
+    update_db_worksheet(req.store, subject, worksheet, worksheetmtime,
+        ctx['exerciselist'])
+    
+    ctx['worksheetstream'] = add_exercises(ctx['worksheetstream'], ctx, req)
+
+# This generator adds in the exercises as they are required. This is returned    
+def add_exercises(stream, ctx, req):
+    """A filter adds exercises into the stream."""
+    exid = 0
+    for kind, data, pos in stream:
+        if kind is genshi.core.START:
+            if data[0] == 'exercise':
+                new_stream = ctx['exercises'][exid]['stream']
+                exid += 1
+                for item in new_stream:
+                    yield item
             else:
-                # Heading node
-                fragment_id = getID(xml)
-                title = getTextData(xml)
-                req.write('  <li><a href="#%s">%s</a></li>\n'
-                    % (fragment_id, cgi.escape(title)))
-    finally:
-        db.close()
-    req.write('</ul>\n</div>\n')
-    return exercise_list
-
-def find_all_nodes(req, node):
-    """Generator. Searches through a node and yields all headings and
-    exercises. (Recursive).
-    When finding a heading, yields a pair ("hx", headingnode), where "hx" is
-    the element name, such as "h1", "h2", etc.
-    When finding an exercise, yields a pair ("ex", exercisenode), where
-    exercisenode is the DOM node for this exercise.
-    """
-    if node.nodeType == node.ELEMENT_NODE:
-        if node.tagName == "exercise":
-            yield "ex", node
-        elif (node.tagName == "h1" or node.tagName == "h2"
-            or node.tagName == "h3"):
-            yield node.tagName, node
-        else:
-            # Some other element. Recurse.
-            for childnode in node.childNodes:
-                for yieldval in find_all_nodes(req, childnode):
-                    yield yieldval
-
-def present_worksheet_node(req, node, exerciseid):
-    """Given a node of a worksheet XML document, writes it out to the
-    request. This recursively searches for "exercise" elements and handles
-    those specially (presenting their XML exercise spec and input box), and
-    just dumps the other elements as regular HTML.
-
-    exerciseid is the ID to use for the first exercise.
-    Returns the new exerciseid after all the exercises have been written
-    (since we need unique IDs for each exercise).
-    """
-    if node.nodeType == node.ELEMENT_NODE:
-        if node.tagName == "exercise":
-            present_exercise(req, node.getAttribute("src"), exerciseid)
-            exerciseid += 1
-        else:
-            # Some other element. Write out its head and foot, and recurse.
-            req.write("<" + node.tagName)
-            # Attributes
-            attrs = map(lambda (k,v): '%s="%s"'
-                    % (cgi.escape(k), cgi.escape(v)), node.attributes.items())
-            if len(attrs) > 0:
-                req.write(" " + ' '.join(attrs))
-            req.write(">")
-            for childnode in node.childNodes:
-                exerciseid = present_worksheet_node(req, childnode, exerciseid)
-            req.write("</" + node.tagName + ">")
-    else:
-        # No need to recurse, so just print this node's contents
-        req.write(node.toxml())
-    return exerciseid
+                yield kind, data, pos
+        else:
+            yield kind, data, pos
+
+# This function runs through the worksheet, to get data on the exercises to
+# build a Table of Contents, as well as fill in details in ctx
+def generate_worksheet_data(ctx, req):
+    """Runs through the worksheetstream, generating the exericises"""
+    exid = 0
+    ctx['exercises'] = []
+    ctx['exerciselist'] = []
+    for kind, data, pos in ctx['worksheetstream']:
+        if kind is genshi.core.START:
+            if data[0] == 'exercise':
+                exid += 1
+                src = ""
+                optional = False
+                for attr in data[1]:
+                    if attr[0] == 'src':
+                        src = attr[1]
+                    if attr[0] == 'optional':
+                        optional = attr[1] == 'true'
+                # Each item in toc is of type (name, complete, stream)
+                ctx['exercises'].append(present_exercise(req, src, exid))
+                ctx['exerciselist'].append((src, optional))
+            elif data[0] == 'worksheet':
+                ctx['worksheetname'] = 'bob'
+                for attr in data[1]:
+                    if attr[0] == 'name':
+                        ctx['worksheetname'] = attr[1]
 
 def innerXML(elem):
     """Given an element, returns its children as XML strings concatenated
@@ -445,21 +367,6 @@
         s += child.toxml()
     return s
 
-def getID(element):
-    """Get the first ID attribute found when traversing a node and its
-    children. (This is used to make fragment links to a particular element).
-    Returns None if no ID is found.
-    """
-    id = element.getAttribute("id")
-    if id is not None and id != '':
-        return id
-    for child in element.childNodes:
-        if child.nodeType == child.ELEMENT_NODE:
-            id = getID(child)
-            if id is not None:
-                return id
-    return None
-
 def getTextData(element):
     """ Get the text and cdata inside an element
     Leading and trailing whitespace are stripped
@@ -475,147 +382,86 @@
 
     return data.strip()
 
+#TODO: This needs to be re-written, to stop using minidom, and get the data
+# about the worksheet directly from the database
 def present_exercise(req, exercisesrc, exerciseid):
     """Open a exercise file, and write out the exercise to the request in HTML.
     exercisesrc: "src" of the exercise file. A path relative to the top-level
         exercises base directory, as configured in conf.
     """
-    req.write('<div class="exercise" id="exercise%d">\n'
-        % exerciseid)
+    # Exercise-specific context is used here, as we already have all the data
+    # we need
+    curctx = genshi.template.Context()
+    curctx['filename'] = exercisesrc
+    curctx['exerciseid'] = exerciseid
+
+    # Retrieve the exercise details from the database
+    exercise = ivle.database.Exercise.get_by_name(req.store, exercisesrc)
+    #Open the exercise, and double-check that it exists
     exercisefile = util.open_exercise_file(exercisesrc)
     if exercisefile is None:
-        req.write("<p><b>Server Error</b>: "
-            + "Exercise file could not be opened.</p>\n")
-        req.write("</div>\n")
-        return
+        req.throw_error(req.HTTP_EXPECTATION_FAILED, \
+                                        "Exercise file could not be opened")
     
     # Read exercise file and present the exercise
     # Note: We do not use the testing framework because it does a lot more
     # work than we need. We just need to get the exercise name and a few other
     # fields from the XML.
 
+    #TODO: Replace calls to minidom with calls to the database directly
     exercisedom = minidom.parse(exercisefile)
     exercisefile.close()
     exercisedom = exercisedom.documentElement
     if exercisedom.tagName != "exercise":
         req.throw_error(req.HTTP_INTERNAL_SERVER_ERROR,
             "The exercise XML file's top-level element must be <exercise>.")
-    exercisename = exercisedom.getAttribute("name")
-    rows = exercisedom.getAttribute("rows")
-    if not rows:
-        rows = "12"
+    curctx['exercisename'] = exercisedom.getAttribute("name")
+    
+    curctx['rows'] = exercisedom.getAttribute("rows")
+    if not curctx['rows']:
+        curctx['rows'] = "12"
     # Look for some other fields we need, which are elements:
     # - desc
     # - partial
-    exercisedesc = None
-    exercisepartial= ""
+    curctx['exercisedesc'] = None
+    curctx['exercisepartial'] = ""
     for elem in exercisedom.childNodes:
         if elem.nodeType == elem.ELEMENT_NODE:
             if elem.tagName == "desc":
-                exercisedesc = rst(innerXML(elem).strip())
+                curctx['exercisedesc'] = genshi.XML(rst(innerXML(elem).strip()))
             if elem.tagName == "partial":
-                exercisepartial= getTextData(elem) + '\n'
-    exercisepartial_backup = exercisepartial
+                curctx['exercisepartial'] = getTextData(elem) + '\n'
+    curctx['exercisepartial_backup'] = curctx['exercisepartial']
 
     # If the user has already saved some text for this problem, or submitted
     # an attempt, then use that text instead of the supplied "partial".
-    saved_text = None
-    db = ivle.db.DB()
-    try:
-        saved_text = db.get_problem_stored_text(login=req.user.login,
-            exercisename=exercisesrc)
-        # Also get the number of attempts taken and whether this is complete.
-        complete, attempts = db.get_problem_status(login=req.user.login,
-            exercisename=exercisesrc)
-    finally:
-        db.close()
+    saved_text = ivle.worksheet.get_exercise_stored_text(req.store,
+        req.user, exercise)
+    # Also get the number of attempts taken and whether this is complete.
+    complete, curctx['attempts'] = \
+            ivle.worksheet.get_exercise_status(req.store, req.user, exercise)
     if saved_text is not None:
-        # Important: We got the string from the DB encoded in UTF-8
-        # Make it a unicode string.
-        exercisepartial = saved_text.decode('utf-8')
-
-    # Print this exercise out to HTML 
-    req.write("<p><b>Exercise:</b> %s</p>\n" % cgi.escape(exercisename))
-    if exercisedesc is not None:
-        req.write("<div>%s</div>\n" % exercisedesc)
-    filename = cgi.escape(cjson.encode(exercisesrc), quote=True)
-    req.write("""<input id="input_resettext_exercise%d" type="hidden"
-    value="%s" />"""
-        % (exerciseid, urllib.quote(exercisepartial_backup)))
-    req.write("""<textarea id="textarea_exercise%d" class="exercisebox"
-    onkeypress="return catch_textbox_input(&quot;exercise%d&quot;, %s,
-        event.keyCode)"
-    onchange="set_saved_status(&quot;exercise%d&quot;, %s,
-        &quot;Save&quot;)"
-    cols="80" rows="%s">%s</textarea>"""
-        % (exerciseid, exerciseid, filename, exerciseid, filename,
-            rows, cgi.escape(exercisepartial)))
-    req.write("""\n<div class="exercisebuttons">\n""")
-    req.write("""  <input type="button" value="Saved" disabled="disabled"
-    id="savebutton_exercise%d"
-    onclick="saveexercise(&quot;exercise%d&quot;, %s)"
-    title="Save your solution to this exercise" />\n"""
-        % (exerciseid, exerciseid, filename))
-    req.write("""  <input type="button" value="Reset"
-    id="resetbutton_exercise%d"
-    onclick="resetexercise(&quot;exercise%d&quot;, %s)"
-    title="Reload the original partial solution for this exercise" />\n"""
-        % (exerciseid, exerciseid, filename))
-    req.write("""  <input type="button" value="Run"
-    onclick="runexercise(&quot;exercise%d&quot;, %s)"
-    title="Run this program in the console" />\n"""
-        % (exerciseid, filename))
-    req.write("""  <input type="button" value="Submit"
-    id="submitbutton_exercise%d"
-    onclick="submitexercise(&quot;exercise%d&quot;, %s)"
-    title="Submit this solution for evaluation" />\n"""
-        % (exerciseid, exerciseid, filename))
-    req.write("""</div>
-<div class="testoutput">
-</div>
-""")
-    # Write the "summary" - whether this problem is complete and how many
-    # attempts it has taken.
-    req.write("""<div class="problem_summary">
-  <ul><li id="summaryli_exercise%d" class="%s">
-    <b><span id="summarycomplete_exercise%d">%s</span>.</b>
-    Attempts: <span id="summaryattempts_exercise%d">%d</span>.
-  </li></ul>
-</div>
-""" % (exerciseid, "complete" if complete else "incomplete",
-        exerciseid, "Complete" if complete else "Incomplete",
-        exerciseid, attempts))
-    # Write the attempt history infrastructure
-    req.write("""<div class="attempthistory">
-  <p><a title="Click to view previous submissions you have made for this \
-exercise" onclick="open_previous(&quot;exercise%d&quot;, %s)">View previous \
-attempts</a></p>
-  <div style="display: none">
-    <h3>Previous attempts</h3>
-    <p><a title="Close the previous attempts" \
-onclick="close_previous(&quot;exercise%d&quot;)">Close attempts</a></p>
-    <p>
-      <select title="Select an attempt's time stamp from the list">
-        <option></option>
-      </select>
-      <input type="button" value="View"
-        onclick="select_attempt(&quot;exercise%d&quot;, %s)" />
-    </p>
-    <p><textarea readonly="readonly" class="exercisebox" cols="80" rows="%s"
-        title="You submitted this code on a previous attempt">
-       </textarea>
-    </p>
-  </div>
-</div>
-""" % (exerciseid, filename, exerciseid, exerciseid, filename, rows))
-    req.write("</div>\n")
-
-def update_db_worksheet(subject, worksheet, file_mtime,
+        curctx['exercisepartial'] = saved_text.text
+    if complete:
+        curctx['complete'] = 'complete'
+    else:
+        curctx['complete'] = 'incomplete'
+
+    #Save the exercise details to the Table of Contents
+
+    loader = genshi.template.TemplateLoader(".", auto_reload=True)
+    tmpl = loader.load(util.make_local_path("apps/tutorial/exercise.html"))
+    ex_stream = tmpl.generate(curctx)
+    return {'name': curctx['exercisename'], 'complete': curctx['complete'], \
+              'stream': ex_stream, 'exid': exerciseid}
+
+
+def update_db_worksheet(store, subject, worksheetname, file_mtime,
     exercise_list=None, assessable=None):
     """
     Determines if the database is missing this worksheet or out of date,
     and inserts or updates its details about the worksheet.
-    file_mtime is a time.struct_time with the modification time of the XML
+    file_mtime is a datetime.datetime with the modification time of the XML
     file. The database will not be updated unless worksheetmtime is newer than
     the mtime in the database.
     exercise_list is a list of (filename, optional) pairs as returned by
@@ -625,10 +471,38 @@
     the existing data. If the worksheet does not yet exist, and assessable
     is omitted, it defaults to False.
     """
-    db = ivle.db.DB()
-    try:
-        db_mtime = db.get_worksheet_mtime(subject, worksheet)
-        if db_mtime is None or file_mtime > db_mtime:
-            db.create_worksheet(subject, worksheet, exercise_list, assessable)
-    finally:
-        db.close()
+    worksheet = ivle.database.Worksheet.get_by_name(store, subject,
+                                                    worksheetname)
+
+    updated_database = False
+    if worksheet is None:
+        # If assessable is not supplied, default to False.
+        if assessable is None:
+            assessable = False
+        # Create a new Worksheet
+        worksheet = ivle.database.Worksheet(subject=unicode(subject),
+            name=unicode(worksheetname), assessable=assessable,
+            mtime=datetime.now())
+        store.add(worksheet)
+        updated_database = True
+    else:
+        if file_mtime > worksheet.mtime:
+            # File on disk is newer than database. Need to update.
+            worksheet.mtime = datetime.now()
+            if exercise_list is not None:
+                # exercise_list is supplied, so delete any existing problems
+                worksheet.remove_all_exercises(store)
+            if assessable is not None:
+                worksheet.assessable = assessable
+            updated_database = True
+
+    if updated_database and exercise_list is not None:
+        # Insert each exercise into the worksheet
+        for exercise_name, optional in exercise_list:
+            # Get the Exercise from the DB
+            exercise = ivle.database.Exercise.get_by_name(store,exercise_name)
+            # Create a new binding between the worksheet and the exercise
+            worksheetexercise = ivle.database.WorksheetExercise(
+                    worksheet=worksheet, exercise=exercise, optional=optional)
+
+    store.commit()

=== added file 'www/apps/tutorial/exercise.html'
--- www/apps/tutorial/exercise.html	1970-01-01 00:00:00 +0000
+++ www/apps/tutorial/exercise.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,73 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+  <div class="exercise" id="exercise${exerciseid}">
+    <p><b>Exercise:</b> ${exercisename}</p>
+    <py:if test="exercisedesc is not None">
+      <div>${exercisedesc}</div>
+    </py:if>
+    <input id="input_resettext_exercise${exerciseid}" type="hidden" value="${exercisepartial_backup}" />
+    <textarea id="textarea_exercise${exerciseid}" class="exercisebox"
+        onkeypress="return catch_textbox_input(&quot;exercise${exerciseid}&quot;, '${filename}',
+        event.keyCode)"
+        onchange="set_saved_status(&quot;exercise${exerciseid}&quot;, '${filename}',
+        &quot;Save&quot;)"
+        cols="80" rows="${rows}">
+        ${exercisepartial}
+    </textarea>
+    <div class="exercisebuttons">
+      <input type="button" value="Saved" disabled="disabled" id="savebutton_exercise${exerciseid}"
+          onclick="saveexercise(&quot;exercise${exerciseid}&quot;, '${filename}')"
+          title="Save your solution to this exercise" />
+      <input type="button" value="Reset" id="resetbutton_exercise${exerciseid}"
+          onclick="resetexercise(&quot;exercise${exerciseid}&quot;, '${filename}')"
+          title="Reload the original partial solution for this exercise" />
+      <input type="button" value="Run"
+          onclick="runexercise(&quot;exercise${exerciseid}&quot;, '${filename}')"
+          title="Run this program in the console" />
+      <input type="button" value="Submit"
+              id="submitbutton_exercise${exerciseid}"
+              onclick="submitexercise(&quot;exercise${exerciseid}&quot;, '${filename}')"
+              title="Submit this solution for evaluation" />
+    </div>
+    <div class="testoutput">
+    </div>
+    <div class="problem_summary">
+      <ul>
+        <li id="summaryli_exercise${exerciseid}" class="${complete}">
+          <b><span id="summarycomplete_exercise${exerciseid}">${complete}</span>.</b>
+            Attempts: <span id="summaryattempts_exercise${exerciseid}">${attempts}</span>.
+        </li>
+      </ul>
+    </div>
+    <!-- This is filled in by javascript when the page loads -->
+    <div class="attempthistory">
+      <p>
+        <a title="Click to view previous submissions you have made for this exercise" 
+              onclick="open_previous(&quot;exercise${exerciseid}&quot;, '${filename}')">
+          View previous attempts
+        </a>
+      </p>
+    <div style="display: none">
+      <h3>Previous attempts</h3>
+        <p>
+          <a title="Close the previous attempts" onclick="close_previous(&quot;exercise${exerciseid}&quot;)">
+            Close attempts
+          </a>
+        </p>
+        <p>
+          <select title="Select an attempt's time stamp from the list">
+            <option></option>
+          </select>
+          <input type="button" value="View"
+              onclick="select_attempt(&quot;exercise${exerciseid}&quot;, '${filename})'" />
+        </p>
+        <p>
+          <textarea readonly="readonly" class="exercisebox" cols="80" rows="${rows}"
+              title="You submitted this code on a previous attempt">
+          </textarea>
+        </p>
+      </div>
+    </div>
+  </div>
+</html>

=== added file 'www/apps/tutorial/subjectmenu.html'
--- www/apps/tutorial/subjectmenu.html	1970-01-01 00:00:00 +0000
+++ www/apps/tutorial/subjectmenu.html	2009-02-03 03:53:54 +0000
@@ -0,0 +1,38 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+  <head foo="bar" py:strip="">
+    <title>Tutorial - Subject Menu</title>
+  </head>
+  <body class="innerbody" py:strip="">
+  
+    <div id="ivle_padding">
+      <h1>IVLE Tutorials - ${subject}</h1>
+      <h2>Worksheets</h2>
+      <ul id="tutorial-toc">
+        <py:for each="worksheet in worksheets">
+          <li>
+            <a href="${worksheet.id}">${worksheet.name}</a>
+            <py:if test="worksheet.assessable">
+              <ul>
+                <li class="${worksheet.complete_class}">
+                  Completed ${worksheet.mand_done}/${worksheet.total}${worksheet.optional_message}
+                </li>
+              </ul>
+            </py:if>
+          </li>
+        </py:for>
+      </ul>
+      <py:if test="problems_total > 0">
+        <ul>
+          <li class="${complete_class}">
+            Total exercises completed: ${problems_done}/${problems_total} (${problems_pct}%)
+          </li>
+        </ul> 
+        <p style="font-weight: bold">
+          Worksheet mark: ${mark}/${max_mark}
+        </p>
+      </py:if>
+    </div>            
+  </body>
+</html>

=== added file 'www/apps/tutorial/template.html'
--- www/apps/tutorial/template.html	1970-01-01 00:00:00 +0000
+++ www/apps/tutorial/template.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,9 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+  <py:choose test="whichmenu">
+    <xi:include py:when="'toplevel'" href="toplevel.html" />
+    <xi:include py:when="'subjectmenu'" href="subjectmenu.html" />
+    <xi:include py:when="'worksheet'" href="worksheet.html" />
+  </py:choose>
+</html>

=== added file 'www/apps/tutorial/toplevel.html'
--- www/apps/tutorial/toplevel.html	1970-01-01 00:00:00 +0000
+++ www/apps/tutorial/toplevel.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,39 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+  <head foo="bar" py:strip="">
+    <title>Tutorial - Top Level</title>
+  </head>
+  <body class="innerbody" py:strip="">
+    <div id="ivle_padding">
+      <h1>IVLE Tutorials</h1>
+      <p>
+        Welcome to the IVLE tutorial system.
+  Please select a subject from the list below to select a worksheet
+  for that subject.
+      </p>
+      <h2>Subjects</h2>
+      <ul>
+        <py:for each="subject in enrolled_subjects">
+          <li>
+            <a href="${subject.code}">${subject.name}</a>
+          </li>
+        </py:for>
+      </ul>
+      <py:if test="len(unenrolled_subjects) > 0">
+        <h3>Other Subjects</h3>
+        <p>
+          You are not currently enrolled in these subjects.
+           Your marks will not be counted.
+        </p>
+        <ul>
+          <py:for each="subject in unenrolled_subjects">
+            <li>
+              <a href="${subject.code}">${subject.name}</a>
+            </li>
+          </py:for>
+        </ul>
+      </py:if>
+    </div>
+  </body>
+</html>

=== added file 'www/apps/tutorial/worksheet.html'
--- www/apps/tutorial/worksheet.html	1970-01-01 00:00:00 +0000
+++ www/apps/tutorial/worksheet.html	2009-02-03 03:53:54 +0000
@@ -0,0 +1,28 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+  <head foo="bar" py:if="False">
+    <title>Tutorial - Worksheet</title>
+  </head>
+  <body class="innerbody" py:strip="">
+    <div id="ivle_padding">
+      <h1>IVLE Tutorials - ${subject}</h1>
+      <!-- Display the Table Of Contents -->
+      <div id="tutorial-toc">
+        <h2>Worksheet Contents</h2>
+        <ul>
+          <py:for each="exercise in exercises">
+            <li class="${exercise['complete']}" id="toc_li_exercise${exercise['exid']}">
+              <a href="#exercise${exercise['exid']}">
+                ${exercise.name}
+              </a>
+            </li>
+          </py:for>
+        </ul>
+      </div>
+      <!-- Display The Worksheet -->
+      <h2>${worksheetname}</h2>
+      ${worksheetstream}
+    </div>
+  </body>
+</html>

=== modified file 'www/apps/tutorialservice/__init__.py'
--- www/apps/tutorialservice/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/tutorialservice/__init__.py	2009-01-20 05:03:55 +0000
@@ -43,10 +43,14 @@
 
 import os
 import time
+import datetime
 
 import cjson
 
-from ivle import (db, util, console)
+from ivle import util
+from ivle import console
+import ivle.database
+import ivle.worksheet
 import ivle.conf
 import test # XXX: Really .test, not real test.
 
@@ -54,6 +58,9 @@
 # attempts. If False, will not allow this.
 HISTORY_ALLOW_INACTIVE = False
 
+TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M:%S'
+
+
 def handle(req):
     """Handler for Ajax backend TutorialService app."""
     # Set request attributes
@@ -94,7 +101,7 @@
         # The time *should* be in the same format as the DB (since it should
         # be bounced back to us from the getattempts output). Assume this.
         try:
-            date = time.strptime(date, db.TIMESTAMP_FORMAT)
+            date = datetime.datetime.strptime(date, TIMESTAMP_FORMAT)
         except ValueError:
             # Date was not in correct format
             req.throw_error(req.HTTP_BAD_REQUEST)
@@ -102,34 +109,29 @@
     else:
         req.throw_error(req.HTTP_BAD_REQUEST)
 
-def handle_save(req, exercise, code, fields):
+def handle_save(req, exercisename, code, fields):
     """Handles a save action. This saves the user's code without executing it.
     """
     # Need to open JUST so we know this is a real exercise.
     # (This avoids users submitting code for bogus exercises).
-    exercisefile = util.open_exercise_file(exercise)
+    exercisefile = util.open_exercise_file(exercisename)
     if exercisefile is None:
         req.throw_error(req.HTTP_NOT_FOUND,
             "The exercise was not found.")
     exercisefile.close()
 
+    exercise = ivle.database.Exercise.get_by_name(req.store, exercisename)
+    ivle.worksheet.save_exercise(req.store, req.user, exercise,
+                                 unicode(code), datetime.datetime.now())
+    req.store.commit()
+
     req.write('{"result": "ok"}')
 
-    conn = db.DB()
-
-    try:
-        conn.write_problem_save(
-            login = req.user.login,
-            exercisename = exercise,
-            date = time.localtime(),
-            text = code)
-    finally:
-        conn.close()
-
-def handle_test(req, exercise, code, fields):
+
+def handle_test(req, exercisesrc, code, fields):
     """Handles a test action."""
 
-    exercisefile = util.open_exercise_file(exercise)
+    exercisefile = util.open_exercise_file(exercisesrc)
     if exercisefile is None:
         req.throw_error(req.HTTP_NOT_FOUND,
             "The exercise was not found.")
@@ -150,49 +152,44 @@
     # Close the console
     cons.close()
 
-    conn = db.DB()
-    try:
-        conn.insert_problem_attempt(
-            login = req.user.login,
-            exercisename = exercise,
-            date = time.localtime(),
-            complete = test_results['passed'],
-            attempt = code)
-
-        # Query the DB to get an updated score on whether or not this problem
-        # has EVER been completed (may be different from "passed", if it has
-        # been completed before), and the total number of attempts.
-        completed, attempts = conn.get_problem_status(req.user.login,
-            exercise)
-        test_results["completed"] = completed
-        test_results["attempts"] = attempts
-
-        req.write(cjson.encode(test_results))
-    finally:
-        conn.close()
-
-def handle_getattempts(req, exercise):
+    # Get the Exercise from the database
+    exercise = ivle.database.Exercise.get_by_name(req.store, exercisesrc)
+
+    attempt = ivle.database.ExerciseAttempt(user=req.user,
+                                            exercise=exercise,
+                                            date=datetime.datetime.now(),
+                                            complete=test_results['passed'],
+                                            text=unicode(code)) # XXX
+
+    req.store.add(attempt)
+    req.store.commit()
+    # Query the DB to get an updated score on whether or not this problem
+    # has EVER been completed (may be different from "passed", if it has
+    # been completed before), and the total number of attempts.
+    completed, attempts = ivle.worksheet.get_exercise_status(req.store,
+        req.user, exercise)
+    test_results["completed"] = completed
+    test_results["attempts"] = attempts
+
+    req.write(cjson.encode(test_results))
+
+def handle_getattempts(req, exercisename):
     """Handles a getattempts action."""
-    conn = db.DB()
-    try:
-        attempts = conn.get_problem_attempts(
-            login=req.user.login,
-            exercisename=exercise,
-            allow_inactive=HISTORY_ALLOW_INACTIVE)
-        req.write(cjson.encode(attempts))
-    finally:
-        conn.close()
+    exercise = ivle.database.Exercise.get_by_name(req.store, exercisename)
+    attempts = ivle.worksheet.get_exercise_attempts(req.store, req.user,
+        exercise, allow_inactive=HISTORY_ALLOW_INACTIVE)
+    # attempts is a list of ExerciseAttempt objects. Convert to dictionaries.
+    time_fmt = lambda dt: datetime.datetime.strftime(dt, TIMESTAMP_FORMAT)
+    attempts = [{'date': time_fmt(a.date), 'complete': a.complete}
+                for a in attempts]
+    req.write(cjson.encode(attempts))
 
-def handle_getattempt(req, exercise, date):
-    """Handles a getattempts action. Date is a struct_time."""
-    conn = db.DB()
-    try:
-        attempt = conn.get_problem_attempt(
-            login=req.user.login,
-            exercisename=exercise,
-            as_of=date,
-            allow_inactive=HISTORY_ALLOW_INACTIVE)
-        # attempt may be None; will write "null"
-        req.write(cjson.encode({'code': attempt}))
-    finally:
-        conn.close()
+def handle_getattempt(req, exercisename, date):
+    """Handles a getattempts action. Date is a datetime.datetime."""
+    exercise = ivle.database.Exercise.get_by_name(req.store, exercisename)
+    attempt = ivle.worksheet.get_exercise_attempt(req.store, req.user,
+        exercise, as_of=date, allow_inactive=HISTORY_ALLOW_INACTIVE)
+    if attempt is not None:
+        attempt = attempt.text
+    # attempt may be None; will write "null"
+    req.write(cjson.encode({'code': attempt}))

=== modified file 'www/apps/userservice/__init__.py'
--- www/apps/userservice/__init__.py	2009-01-13 01:36:15 +0000
+++ www/apps/userservice/__init__.py	2009-02-02 04:42:36 +0000
@@ -139,24 +139,33 @@
 
 import os
 import sys
-import time
+import datetime
 
 import cjson
 import pg
 
-import ivle.db
-import ivle.makeuser
+import ivle.database
 from ivle import (util, chat, caps)
 from ivle.conf import (usrmgt_host, usrmgt_port, usrmgt_magic)
-
-from ivle.auth import authenticate
-from ivle.auth.autherror import AuthError
+import ivle.pulldown_subj
+
+from ivle.rpc.decorators import require_method, require_cap
+
+from ivle.auth import AuthError, authenticate
 import urllib
 
 # The user must send this declaration message to ensure they acknowledge the
 # TOS
 USER_DECLARATION = "I accept the IVLE Terms of Service"
 
+# List of fields returned as part of the user JSON dictionary
+# (as returned by the get_user action)
+user_fields_list = (
+    "login", "state", "unixid", "email", "nick", "fullname",
+    "rolenm", "studentid", "acct_exp", "pass_exp", "last_login",
+    "svn_pass"
+)
+
 def handle(req):
     """Handler for the Console Service AJAX backend application."""
     if req.user is None:
@@ -176,6 +185,7 @@
         "%s is not a valid userservice action." % repr(req.path))
     func(req, fields)
 
+@require_method('POST')
 def handle_activate_me(req, fields):
     """Create the jail, svn, etc, for the currently logged in user (this is
     put in the queue for usermgt to do).
@@ -194,11 +204,7 @@
     "accepting" the terms - at least this way requires them to acknowledge
     their acceptance). It must only be called through a POST request.
     """
-    db = ivle.db.DB()
     try:
-        if req.method != "POST":
-            req.throw_error(req.HTTP_METHOD_NOT_ALLOWED,
-            "Only POST requests are valid methods to activate_me.")
         try:
             declaration = fields.getfirst('declaration')
         except AttributeError:
@@ -211,22 +217,19 @@
         # Make sure the user's status is "no_agreement", and set status to
         # pending, within the one transaction. This ensures we only do this
         # one time.
-        db.start_transaction()
         try:
-
-            user_details = db.get_user(req.user.login)
             # Check that the user's status is "no_agreement".
             # (Both to avoid redundant calls, and to stop disabled users from
             # re-enabling their accounts).
-            if user_details.state != "no_agreement":
+            if req.user.state != "no_agreement":
                 req.throw_error(req.HTTP_BAD_REQUEST,
                 "You have already agreed to the terms.")
             # Write state "pending" to ensure we don't try this again
-            db.update_user(req.user.login, state="pending")
+            req.user.state = u"pending"
         except:
-            db.rollback()
+            req.store.rollback()
             raise
-        db.commit()
+        req.store.commit()
 
         # Get the arguments for usermgt.activate_user from the session
         # (The user must have already logged in to use this app)
@@ -235,6 +238,9 @@
         }
         msg = {'activate_user': args}
 
+        # Release our lock on the db so usrmgt can write
+        req.store.rollback()
+
         # Try and contact the usrmgt server
         try:
             response = chat.chat(usrmgt_host, usrmgt_port, msg, usrmgt_magic)
@@ -249,23 +255,18 @@
             status = 'failure'
         
         if status == 'okay':
-            req.user.state = "enabled"
+            req.user.state = u"enabled"
         else:
             # Reset the user back to no agreement
-            req.user.state = "no_agreement"
-            db.update_user(req.user.login, state="no_agreement")
-
-
-        # Update the users state
-        session = req.get_session()
-        session['user'] = req.user
-        session.save()
+            req.user.state = u"no_agreement"
+            req.store.commit()
 
         # Write the response
         req.content_type = "text/plain"
         req.write(cjson.encode(response))
-    finally:
-        db.close()
+    except:
+        req.store.rollback()
+        raise
 
 create_user_fields_required = [
     'login', 'fullname', 'rolenm'
@@ -273,6 +274,9 @@
 create_user_fields_optional = [
     'password', 'nick', 'email', 'studentid'
 ]
+
+@require_method('POST')
+@require_cap(caps.CAP_UPDATEUSER)
 def handle_create_user(req, fields):
     """Create a new user, whose state is no_agreement.
     This does not create the user's jail, just an entry in the database which
@@ -299,14 +303,6 @@
                       STRING OPTIONAL
        Return Value: the uid associated with the user. INT
     """
-    if req.method != "POST":
-        req.throw_error(req.HTTP_METHOD_NOT_ALLOWED,
-            "Only POST requests are valid methods to create_user.")
-    # Check if this user has CAP_UPDATEUSER
-    if not req.user.hasCap(caps.CAP_UPDATEUSER):
-        req.throw_error(req.HTTP_FORBIDDEN,
-        "You do not have permission to create users.")
-
     # Make a dict of fields to create
     create = {}
     for f in create_user_fields_required:
@@ -323,8 +319,11 @@
         else:
             pass
 
-    ivle.makeuser.make_user_db(**create)
-    user = ivle.db.DB().get_user(create["login"])
+    user = ivle.database.User(**create)
+    req.store.add(user)
+    ivle.pulldown_subj.enrol_user(req.store, user)
+    req.store.commit()
+
     req.content_type = "text/plain"
     req.write(str(user.unixid))
 
@@ -335,15 +334,13 @@
     'password', 'nick', 'email', 'rolenm', 'unixid', 'fullname',
     'studentid'
 ]
+
+@require_method('POST')
 def handle_update_user(req, fields):
     """Update a user's account details.
     This can be done in a limited form by any user, on their own account,
     or with full powers by a user with CAP_UPDATEUSER on any account.
     """
-    if req.method != "POST":
-        req.throw_error(req.HTTP_METHOD_NOT_ALLOWED,
-        "Only POST requests are valid methods to update_user.")
-
     # Only give full powers if this user has CAP_UPDATEUSER
     fullpowers = req.user.hasCap(caps.CAP_UPDATEUSER)
     # List of fields that may be changed
@@ -362,42 +359,35 @@
         # If login not specified, update yourself
         login = req.user.login
 
-    # Make a dict of fields to update
-    update = {}
+    user = ivle.database.User.get_by_login(req.store, login)
+
     oldpassword = fields.getfirst('oldpass')
-    
-    for f in fieldlist:
-        val = fields.getfirst(f)
-        if val is not None:
-            update[f] = val
-        else:
-            pass
+    if oldpassword is not None: # It was specified.
+        oldpassword = oldpassword.value
 
-    if 'password' in update:
+    # If the user is trying to set a new password, check that they have
+    # entered old password and it authenticates.
+    if fields.getfirst('password') is not None:
         try:
-            authenticate.authenticate(login, oldpassword)
+            authenticate.authenticate(req.store, login, oldpassword)
         except AuthError:
             req.headers_out['X-IVLE-Action-Error'] = \
                 urllib.quote("Old password incorrect.")
             req.status = req.HTTP_BAD_REQUEST
+            # Cancel all the changes made to user (including setting new pass)
+            req.store.rollback()
             return
 
-    update['login'] = login
-
-    db = ivle.db.DB()
-    db.update_user(**update)
-
-    # Re-read the user's details from the DB so we can update their session
-    # XXX potentially-unsafe session write
-    if login == req.user.login:
-        user = db.get_user(login)
-        session = req.get_session()
-        session.lock()
-        session['user'] = user
-        session.save()
-        session.unlock()
-
-    db.close()
+    # Make a dict of fields to update
+    for f in fieldlist:
+        val = fields.getfirst(f)
+        if val is not None:
+            # Note: May be rolled back if auth check below fails
+            setattr(user, f, val.value.decode('utf-8'))
+        else:
+            pass
+
+    req.store.commit()
 
     req.content_type = "text/plain"
     req.write('')
@@ -424,33 +414,15 @@
         login = req.user.login
 
     # Just talk direct to the DB
-    db = ivle.db.DB()
-    user = db.get_user(login)
-    db.close()
-    user = dict(user)
-    if 'role' in user:
-        user['rolenm'] = str(user['role'])
-        del user['role']
-    try:
-        del user['svn_pass']
-    except KeyError:
-        pass
+    userobj = ivle.database.User.get_by_login(req.store, login)
+    user = ivle.util.object_to_dict(user_fields_list, userobj)
     # Convert time stamps to nice strings
-    try:
-        if user['pass_exp'] is not None:
-            user['pass_exp'] = str(user['pass_exp'])
-    except KeyError:
-        pass
-    try:
-        if user['acct_exp'] is not None:
-            user['acct_exp'] = str(user['acct_exp'])
-    except KeyError:
-        pass
-    try:
-        if user['last_login'] is not None:
-            user['last_login'] = str(user['last_login'])
-    except KeyError:
-        pass
+    for k in 'pass_exp', 'acct_exp', 'last_login':
+        if user[k] is not None:
+            user[k] = unicode(user[k])
+
+    user['local_password'] = userobj.passhash is not None
+
     response = cjson.encode(user)
     req.content_type = "text/plain"
     req.write(response)
@@ -466,24 +438,32 @@
     ##fullpowers = req.user.hasCap(caps.CAP_GETUSER)
 
     try:
-        login = fields.getfirst('login')
-        if login is None:
+        user = ivle.database.User.get_by_login(req.store,
+                    fields.getfirst('login'))
+        if user is None:
             raise AttributeError()
-        if not fullpowers and login != req.user.login:
+        if not fullpowers and user != req.user:
             # Not allowed to edit other users
             req.throw_error(req.HTTP_FORBIDDEN,
             "You do not have permission to see another user's subjects.")
     except AttributeError:
         # If login not specified, update yourself
-        login = req.user.login
+        user = req.user
 
-    # Just talk direct to the DB
-    db = ivle.db.DB()
-    enrolments = db.get_enrolment(login)
-    for e in enrolments:
-        e['groups'] = db.get_enrolment_groups(login, e['offeringid'])
-    db.close()
-    response = cjson.encode(enrolments)
+    dict_enrolments = []
+    for e in user.active_enrolments:
+        dict_enrolments.append({
+            'offeringid':      e.offering.id,
+            'subj_code':       e.offering.subject.code,
+            'subj_name':       e.offering.subject.name,
+            'subj_short_name': e.offering.subject.short_name,
+            'url':             e.offering.subject.url,
+            'year':            e.offering.semester.year,
+            'semester':        e.offering.semester.semester,
+            'groups':          [{'name': group.name,
+                                 'nick': group.nick} for group in e.groups]
+        })
+    response = cjson.encode(dict_enrolments)
     req.content_type = "text/plain"
     req.write(response)
 
@@ -503,14 +483,16 @@
     except:
         req.throw_error(req.HTTP_BAD_REQUEST,
             "subjectid must be a integer")
-    
-    db = ivle.db.DB()
-    try:
-        offerings = db.get_offering_semesters(subjectid)
-    finally:
-        db.close()
-
-    response = cjson.encode([o for o in offerings if o['active']])
+
+    subject = req.store.get(ivle.database.Subject, subjectid)
+
+    response = cjson.encode([{'offeringid': offering.id,
+                              'subj_name': offering.subject.name,
+                              'year': offering.semester.year,
+                              'semester': offering.semester.semester,
+                              'active': offering.semester.active
+                             } for offering in subject.offerings
+                                    if offering.semester.active])
     req.content_type = "text/plain"
     req.write(response)
 
@@ -530,116 +512,27 @@
     except:
         req.throw_error(req.HTTP_BAD_REQUEST,
             "offeringid must be a integer")
-    
-    db = ivle.db.DB()
-    try:
-        projectsets = db.get_projectsets_by_offering(offeringid)
-        for p in projectsets:
-            p['groups'] = db.get_groups_by_projectset(p['projectsetid'])
-    except Exception, e:
-        req.throw_error(req.HTTP_INTERNAL_SERVER_ERROR, repr(e))
-    finally:
-        db.close()
-
-    response = cjson.encode(projectsets)
-    req.write(response)
-
-def handle_create_project_set(req, fields):
-    """Required cap: CAP_MANAGEPROJECTS
-    Creates a project set for a offering - returns the projectsetid
-    Required:
-        offeringid, max_students_per_group
-    """
-    
-    if req.method != "POST":
-        req.throw_error(req.HTTP_METHOD_NOT_ALLOWED,
-            "Only POST requests are valid methods to create_user.")
-    # Check if this user has CAP_MANAGEPROJECTS
-    if not req.user.hasCap(caps.CAP_MANAGEPROJECTS):
-        req.throw_error(req.HTTP_FORBIDDEN,
-        "You do not have permission to manage projects.")
-    # Get required fields
-    offeringid = fields.getfirst('offeringid')
-    max_students_per_group = fields.getfirst('max_students_per_group')
-    if offeringid is None or max_students_per_group is None:
-        req.throw_error(req.HTTP_BAD_REQUEST,
-            "Required: offeringid, max_students_per_group")
-
-    # Talk to the DB
-    db = ivle.db.DB()
-    dbquery = db.return_insert(
-        {
-            'offeringid': offeringid,
-            'max_students_per_group': max_students_per_group,
-        },
-        "project_set",
-        frozenset(["offeringid", "max_students_per_group"]),
-        ["projectsetid"],
-    )
-    db.close()
-    
-    response = cjson.encode(dbquery.dictresult()[0])
-
-    req.content_type = "text/plain"
-    req.write(response)
-
-def handle_create_project(req, fields):
-    """Required cap: CAP_MANAGEPROJECTS
-    Creates a project in a specific project set
-    Required:
-        projectsetid
-    Optional:
-        synopsis, url, deadline
-    Returns:
-        projectid
-    """
-    
-    if req.method != "POST":
-        req.throw_error(req.HTTP_METHOD_NOT_ALLOWED,
-            "Only POST requests are valid methods to create_user.")
-    # Check if this user has CAP_MANAGEPROJECTS
-    if not req.user.hasCap(caps.CAP_MANAGEPROJECTS):
-        req.throw_error(req.HTTP_FORBIDDEN,
-        "You do not have permission to manage projects.")
-    # Get required fields
-    projectsetid = fields.getfirst('projectsetid')
-    if projectsetid is None:
-        req.throw_error(req.HTTP_BAD_REQUEST,
-            "Required: projectsetid")
-    # Get optional fields
-    synopsis = fields.getfirst('synopsis')
-    url = fields.getfirst('url')
-    deadline = fields.getfirst('deadline')
-    if deadline is not None:
-        try:
-            deadline = util.parse_iso8601(deadline).timetuple()
-        except ValueError, e:
-            req.throw_error(req.HTTP_BAD_REQUEST, e.message)
-
-    # Talk to the DB
-    db = ivle.db.DB()
-    try:
-        dbquery = db.return_insert(
-            {
-                'projectsetid': projectsetid,
-                'synopsis': synopsis,
-                'url': url,
-                'deadline': deadline,
-            },
-            "project", # table
-            frozenset(["projectsetid", "synopsis", "url", "deadline"]),
-            ["projectid"], # returns
-        )
-    except Exception, e:
-        req.throw_error(req.HTTP_INTERNAL_SERVER_ERROR, repr(e))
-    finally:
-        db.close()
-    
-    response = cjson.encode(dbquery.dictresult()[0])
-
-    req.content_type = "text/plain"
-    req.write(response)
-
+
+    offering = req.store.get(ivle.database.Offering, offeringid)
+
+    dict_projectsets = []
+    try:
+        for p in offering.project_sets:
+            dict_projectsets.append({
+                'projectsetid': p.id,
+                'max_students_per_group': p.max_students_per_group,
+                'groups': [{'groupid': g.id,
+                            'groupnm': g.name,
+                            'nick': g.nick} for g in p.project_groups]
+            })
+    except Exception, e:
+        req.throw_error(req.HTTP_INTERNAL_SERVER_ERROR, repr(e))
+
+    response = cjson.encode(dict_projectsets)
+    req.write(response)
+
+@require_method('POST')
+@require_cap(caps.CAP_MANAGEGROUPS)
 def handle_create_group(req, fields):
     """Required cap: CAP_MANAGEGROUPS
     Creates a project group in a specific project set
@@ -650,69 +543,44 @@
     Returns:
         groupid
     """
-    if req.method != "POST":
-        req.throw_error(req.HTTP_METHOD_NOT_ALLOWED,
-            "Only POST requests are valid methods to create_user.")
-    # Check if this is allowed to manage groups
-    if not req.user.hasCap(caps.CAP_MANAGEGROUPS):
-        req.throw_error(req.HTTP_FORBIDDEN,
-        "You do not have permission to manage groups.")
     # Get required fields
-    projectsetid = fields.getfirst('projectsetid')
-    groupnm = fields.getfirst('groupnm')
+    projectsetid = fields.getfirst('projectsetid').value
+    groupnm = fields.getfirst('groupnm').value
     if projectsetid is None or groupnm is None:
         req.throw_error(req.HTTP_BAD_REQUEST,
             "Required: projectsetid, groupnm")
+    groupnm = unicode(groupnm)
     try:
         projectsetid = int(projectsetid)
     except:
         req.throw_error(req.HTTP_BAD_REQUEST,
             "projectsetid must be an int")
     # Get optional fields
-    nick = fields.getfirst('nick')
-
-    # Talk to the DB
-    db = ivle.db.DB()
-    # Other fields
-    createdby = db.get_user_loginid(req.user.login)
-    epoch = time.localtime()
+    nick = fields.getfirst('nick').value
+    if nick is not None:
+        nick = unicode(nick)
 
     # Begin transaction since things can go wrong
-    db.start_transaction()
     try:
-        dbquery = db.return_insert(
-            {
-                'groupnm': groupnm,
-                'projectsetid': projectsetid,
-                'nick': nick,
-                'createdby': createdby,
-                'epoch': epoch,
-            },
-            "project_group", # table
-            frozenset(["groupnm", "projectsetid", "nick", "createdby",
-                "epoch"]), # fields
-            ["groupid"], # returns
-        )
- 
-        singlerow = dbquery.dictresult()[0]
-        groupid = singlerow['groupid']
+        group = ivle.database.ProjectGroup(name=groupnm,
+                                           project_set_id=projectsetid,
+                                           nick=nick,
+                                           created_by=req.user,
+                                           epoch=datetime.datetime.now())
+        req.store.add(group)
 
-        # Create the groups repository
-        # Get the arguments for usermgt.activate_user from the session
-        # (The user must have already logged in to use this app)
-    
-        # Find the rest of the parameters we need
-        offeringinfo = db.get_offering_info(projectsetid)
-                
-        subj_short_name = offeringinfo['subj_short_name']
-        year = offeringinfo['year']
-        semester = offeringinfo['semester']
+        # Create the group repository
+        # Yes, this is ugly, and it would be nice to just pass in the groupid,
+        # but the object isn't visible to the extra transaction in
+        # usrmgt-server until we commit, which we only do once the repo is
+        # created.
+        offering = group.project_set.offering
 
         args = {
-            "subj_short_name": subj_short_name,
-            "year": year,
-            "semester": semester,
-            "groupnm": groupnm,
+            "subj_short_name": offering.subject.short_name,
+            "year": offering.semester.year,
+            "semester": offering.semester.semester,
+            "groupnm": group.name,
         }
         msg = {'create_group_repository': args}
 
@@ -728,18 +596,13 @@
                 "Failure creating repository: %s"%str(usrmgt))
     
         # Everything went OK. Lock it in
-        db.commit()
+        req.store.commit()
 
     except Exception, e:
-        db.rollback()
         req.throw_error(req.HTTP_INTERNAL_SERVER_ERROR, repr(e))
-    finally:
-        db.close()
-
-    response = cjson.encode(singlerow)
 
     req.content_type = "text/plain"
-    req.write(response)
+    req.write('')
 
 def handle_get_group_membership(req, fields):
     """ Required cap: None
@@ -759,19 +622,27 @@
     except:
         req.throw_error(req.HTTP_BAD_REQUEST,
             "groupid must be an int")
+    group = req.store.get(ivle.database.ProjectGroup, groupid)
+
     try:
         offeringid = int(offeringid)
     except:
         req.throw_error(req.HTTP_BAD_REQUEST,
             "offeringid must be an int")
-
-    db = ivle.db.DB()
-    try:
-        offeringmembers = db.get_offering_members(offeringid)
-        groupmembers = db.get_projectgroup_members(groupid)
-    finally:
-        db.close()
-    
+    offering = req.store.get(ivle.database.Offering, offeringid)
+
+
+    offeringmembers = [{'login': user.login,
+                        'fullname': user.fullname
+                       } for user in offering.members.order_by(
+                            ivle.database.User.login)
+                      ]
+    groupmembers = [{'login': user.login,
+                        'fullname': user.fullname
+                       } for user in group.members.order_by(
+                            ivle.database.User.login)
+                      ]
+
     # Make sure we don't include members in both lists
     for member in groupmembers:
         if member in offeringmembers:
@@ -783,46 +654,30 @@
     req.content_type = "text/plain"
     req.write(response)
 
+@require_method('POST')
+@require_cap(caps.CAP_MANAGEGROUPS)
 def handle_assign_group(req, fields):
     """ Required cap: CAP_MANAGEGROUPS
     Assigns a user to a project group
     Required:
         login, groupid
     """
-    if req.method != "POST":
-        req.throw_error(req.HTTP_METHOD_NOT_ALLOWED,
-            "Only POST requests are valid methods to create_user.")
-    # Check if this user is allowed to manage groups
-    if not req.user.hasCap(caps.CAP_MANAGEGROUPS):
-        req.throw_error(req.HTTP_FORBIDDEN,
-        "You do not have permission to manage groups.")
     # Get required fields
     login = fields.getfirst('login')
     groupid = fields.getfirst('groupid')
     if login is None or groupid is None:
         req.throw_error(req.HTTP_BAD_REQUEST,
             "Required: login, groupid")
-    groupid = int(groupid)
-
-    # Talk to the DB
-    db = ivle.db.DB()
-    try:
-        loginid = db.get_user_loginid(login)
-    except ivle.db.DBException, e:
-        req.throw_error(req.HTTP_BAD_REQUEST, repr(e))
-
-    # Add assignment to database
-    # We can't use a transaction here, as usrmgt-server needs to see the
-    # changes!
-    try:
-        dbquery = db.insert(
-            {
-                'loginid': loginid,
-                'groupid': groupid,
-            },
-            "group_member", # table
-            frozenset(["loginid", "groupid"]), # fields
-        )
+
+    group = req.store.get(ivle.database.ProjectGroup, int(groupid))
+    user = ivle.database.User.get_by_login(req.store, login)
+
+    # Add membership to database
+    # We can't keep a transaction open until the end here, as usrmgt-server
+    # needs to see the changes!
+    try:
+        group.members.add(user)
+        req.store.commit()
 
         # Rebuild the svn config file
         # Contact the usrmgt server
@@ -838,8 +693,6 @@
                     "Failure creating repository: %s"%str(usrmgt))
     except Exception, e:
         req.throw_error(req.HTTP_INTERNAL_SERVER_ERROR, repr(e))
-    finally:
-        db.close()
 
     return(cjson.encode({'response': 'okay'}))
 
@@ -854,8 +707,6 @@
     "get_active_offerings": handle_get_active_offerings,
     "get_project_groups": handle_get_project_groups,
     "get_group_membership": handle_get_group_membership,
-    "create_project_set": handle_create_project_set,
-    "create_project": handle_create_project,
     "create_group": handle_create_group,
     "assign_group": handle_assign_group,
 }

=== modified file 'www/media/browser/browser.js'
--- www/media/browser/browser.js	2008-08-05 02:06:49 +0000
+++ www/media/browser/browser.js	2009-01-22 02:14:14 +0000
@@ -377,6 +377,11 @@
     upload_callback_count++;
     if (upload_callback_count >= 2)
     {
+        myFrame = frames['upload_iframe'].document;
+        data = myFrame.firstChild.childNodes[1].firstChild.firstChild.nodeValue;
+        data = JSON.parse(data);
+        if ('Error' in data)
+            alert("Error: " + decodeURIComponent(data['Error']));
         document.getElementsByName('data')[0].value = '';
         refresh();
     }
@@ -545,10 +550,24 @@
     }
 }
 
+/* Updates the list of available actions based on files selected */
 function update_actions()
 {
     var file;
     var numsel = selected_files.length;
+    var svn_selection = false;
+    
+    if (numsel > 0)
+    {
+        svn_selection = true;
+        for (var i = 0; i < selected_files.length; i++){
+            if (file_listing[selected_files[i]]["svnstatus"] == "unversioned")
+            {
+                svn_selection = false;        
+            }
+        }
+    }
+    
     if (numsel <= 1)
     {
         if (numsel == 0)
@@ -619,8 +638,8 @@
      */
     var run = document.getElementById("act_run");
      
-    if (!file.isdir && file.type == "text/x-python" && numsel <= 1
-        && current_file.svnstatus != 'revision')
+    if (numsel <= 1 && !file.isdir && file.type == "text/x-python" 
+            && current_file.svnstatus != 'revision')
     {
         if (numsel == 0)
         {
@@ -703,7 +722,6 @@
     var pubcond = numsel <= 1 && file.isdir;
     if (pubcond)
     {
-        /* TODO: Work out of file is svn'd */
         /* If this dir is already published, call it "Unpublish" */
         if (file.published)
         {
@@ -742,8 +760,11 @@
     /* Subversion actions */
     /* These are only useful if we are in a versioned directory and have some
      * files selected. */
-    set_action_state(["svnadd", "svnremove", "svnrevert", "svncommit"], numsel >= 1 && current_file.svnstatus);
-
+    set_action_state(["svnadd",], numsel >= 1 && current_file.svnstatus);
+    /* And these are only usefull is ALL the selected files are versioned */
+    set_action_state(["svnremove", "svnrevert", "svncommit", "svncopy", 
+            "svncut"], numsel >= 1 && current_file.svnstatus && svn_selection);
+    
     /* Diff, log and update only support one path at the moment, so we must
      * have 0 or 1 versioned files selected. If 0, the directory must be
      * versioned. */
@@ -869,6 +890,12 @@
     case "svnlog":
         window.location = path_join(app_path('svnlog'), current_path, selected_files[0] || '');
         break;
+    case "svncopy":
+        action_svncopy(selected_files);
+        break;
+    case "svncut":
+        action_svncut(selected_files);
+        break;
     }
 }
 

=== modified file 'www/media/browser/listing.js'
--- www/media/browser/listing.js	2008-12-20 03:46:56 +0000
+++ www/media/browser/listing.js	2009-01-19 22:56:46 +0000
@@ -145,11 +145,26 @@
 {
     /* Get the "clipboard" object from the browser cookie */
     var clip_obj = read_cookie("clipboard");
+    var under_subversion;
+    
     if (clip_obj == null)
     {
         alert("No files have been cut or copied.");
         return false;
     }
+
+    if (clip_obj.mode == "svnmove" || clip_obj.mode == "svncopy")
+    {
+        under_subversion = ("svnstatus" in current_file) && 
+                                (current_file.svnstatus != "unversioned");
+        if (!under_subversion)
+        {
+            alert("Cannot perform an Subversion Move outside of"
+                                                + " Permanent directories!");
+            return false;
+        }        
+    }
+    
     /* The clip_obj is exactly what we want to pass, plus the current path
      * as destination. */
     clip_obj.dst = ".";
@@ -157,6 +172,18 @@
     return false;
 }
 
+function action_svncopy(files)
+{
+    action_copy_or_cut(files, "svncopy");
+    return false;
+}
+
+function action_svncut(files)
+{
+    action_copy_or_cut(files, "svnmove");
+    return false;
+}
+
 function action_add(files)
 {
     do_action("svnadd", current_path, {"path":files});
@@ -689,13 +716,10 @@
         row.appendChild(td);
         td = document.createElement("td");
         td.setAttribute("class", "thincol");
-        if (under_subversion)
-        {
-            var icon = svnstatus_to_icon(file.svnstatus);
-            if (icon)
-                td.appendChild(dom_make_img(icon, icon_size, icon_size,
-                    svnstatus_to_string(file.svnstatus)));
-        }
+        var icon = svnstatus_to_icon(file.svnstatus);
+        if (icon)
+            td.appendChild(dom_make_img(icon, icon_size, icon_size,
+                                        svnstatus_to_string(file.svnstatus)));
         row.appendChild(td);
 
         /* Column 3: Filename */

=== modified file 'www/plugins/console/__init__.py'
--- www/plugins/console/__init__.py	2009-01-13 01:36:15 +0000
+++ www/plugins/console/__init__.py	2009-02-02 04:00:25 +0000
@@ -30,6 +30,10 @@
 
 from ivle import util
 
+import genshi
+import genshi.core
+import genshi.template
+
 def insert_scripts_styles(scripts, styles, scripts_init):
     """Given 2 lists of strings: scripts and styles. These lists are lists of
     pathnames, as expected on the "scripts" and "styles" attributes of a
@@ -53,45 +57,14 @@
     windowpane: If True, starts the console in "window pane" mode, where it
     will float over the page and have a "minimize" button.
     """
-    req.write("""<div id="console_body">
-  <div id="console_heading">Python Console
-""")
-    if windowpane:
-        req.write("""<span class="console_button minimize">
-      <a onclick="console_minimize()"
-        title="Minimize the Python console">
-        <img src="%s" /></a>
-    </span>
-""" % cgi.escape(util.make_path("media/images/interface/minimize.png")))
-    req.write("""</div>
-  <div id="console_body2">
-  <pre id="console_output" class='outputMsg'></pre>
-  <div id="console_input">
-    <div id="console_inputArea">
-    </div>
-    <label id="console_prompt">&gt;&gt;&gt; </label>
-    <input id="console_inputText"
-      type="text" size="80" onkeypress="return catch_input(event.keyCode)" />
-    <input type='button' value='Interrupt' onclick='set_interrupt();'/>
-    <input type='button' value='Clear Output' onclick='return clear_output();'/>
-    <input type='button' value='Reset' onclick='console_reset();'/>
-""")
-    if windowpane:
-        req.write("""<span class="console_button maximize">
-      <a onclick="console_maximize()"
-        title="Open up the Python console">
-        <img src="%s" /></a>
-    </span>
-""" % cgi.escape(util.make_path("media/images/interface/maximize.png")))
-    req.write("""</div>
-</div></div>
-""")
-    if windowpane:
-        req.write("""
-<!-- Console filler, provides extra vertical space to stop the console
-     covering over the bottom content -->
-<div id="console_filler"></div>
-""")
+    ctx = genshi.template.Context()
+    ctx['windowpane'] = windowpane
+    ctx['minimize_path'] = util.make_path("media/images/interface/minimize.png") 
+    ctx['maximize_path'] = util.make_path("media/images/interface/maximize.png")
+    loader = genshi.template.TemplateLoader(".", auto_reload=True)
+    tmpl = loader.load(util.make_local_path("plugins/console/template.html"))
+    #TODO: Make the dispatch render this
+    req.write(tmpl.generate(ctx).render('html'))
 
 def _append_if_absent(list, *values):
     """Appends an arbitray number of values to a list, but omits values that

=== added file 'www/plugins/console/template.html'
--- www/plugins/console/template.html	1970-01-01 00:00:00 +0000
+++ www/plugins/console/template.html	2009-02-02 04:00:25 +0000
@@ -0,0 +1,40 @@
+<html xmlns="http://www.w3.org/1999/xhtml"
+      xmlns:xi="http://www.w3.org/2001/XInclude"
+      xmlns:py="http://genshi.edgewall.org/" py:strip="">
+      
+  <body py:strip="">
+    <div id="console_body">
+      <div id="console_heading">Python Console<span py:if="windowpane" 
+          class="console_button minimize">
+          <a onclick="console_minimize()"
+              title="Minimize the Python console">
+            <img src="${minimize_path}" />
+          </a>
+        </span>
+      </div>
+      <div id="console_body2">
+        <pre id="console_output" class='outputMsg'></pre>
+        <div id="console_input">
+          <div id="console_inputArea">
+          </div>
+          <label id="console_prompt">&gt;&gt;&gt; </label>
+          <input id="console_inputText" type="text" size="80" 
+              onkeypress="return catch_input(event.keyCode)" />
+          <input type='button' value='Interrupt' onclick='set_interrupt();'/>
+          <input type='button' value='Clear Output' onclick='return clear_output();'/>
+          <input type='button' value='Reset' onclick='console_reset();'/>
+          <span py:if="windowpane" class="console_button maximize">
+            <a onclick="console_maximize()"
+                title="Open up the Python console">
+              <img src="${maximize_path}" />
+            </a>
+          </span>
+        </div>
+      </div>
+    </div>
+    <div py:if="windowpane" id="console_filler">
+    <!-- Console filler, provides extra vertical space to stop the console
+                                            covering over the bottom content -->
+    </div>
+  </body>
+</html>

